/*************************************************************************** 
 *
 *                       uhfa.c Version 2.02
 *
 *            Control Panel for DSP-10 2-meter Transciever
 *
 *
 * NOTICE: Copyright (C) 1996, 1997, 1998, 1999, 2000, 2001
 *         by Robert S. Larkin
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * The program author can be contacted by email at boblark@proaxis.com.
 * Alternatively contact the author at
 *     Robert S. (Bob) Larkin
 *     2982 N. W. Acacia Place
 *     Corvallis, OR  97330  USA
 *
 * For more details on the background or use of the program, pay a visit to
 * http://www.proaxis.com/~boblark/dsp10.htm  and,
 * http://www.ao.com/~gnome/
 *
 * Compatable with EZKIT-Lite program DSP3 version 0X13 and later.
 *
 * RSL = Bob Larkin  W7PUA   Initial write 17 May 96 
 * 29 Oct 96  Added audio FIR, warm boot.  RSL
 * 16 Nov 96  Added audio input and 8/16 bit data option. RSL
 * 18 Nov 96  Added normalization and recording to disk. RSL
 * 10 Dec 96  Reorganized keyboard to allow CW. RSL
 * 22 Dec 96  Added screen-save recording of screen image.  RSL
 *  2 Jan 97  Added transverters and configuration file.  RSL
 * 20 Jan 97  Added CW transmit functions.  RSL
 * 27 June 97 Revised to updated hardware: RF gains, T/R sequencing RSL
 * 28 Sept 97 Added control for DAC for carrier cancellation
 * 11 Mar 98  Revised to crystal filter hdw. Removed i-q corrections. RSL
 * Ver1.3:
 * 21 Sept 98 Added experimental versions of LHL and PUA43 Code estimators
 * 25 Nov  98 Revised to a 5 kHz ref freq for 1st LO (v1.31)
 * Ver1.40
 * 17 Jan  99 Full review.  Fixed error 112 problems. Ver 1.4 RSL
 * 17 Apr  99 Update of audio processor unique elements.
 * 29 Apr  99 Corrections & added binaural. Ver 1.41
 *  2 Aug  99 Startup detail, Fd corrected.  v1.5  RSL
 * 12 Sept 99 Put under GNU GPL, corrected startup error.
 *            Removed GIF for current distribution.  V1.6, RSL
 * 12 Jan  00 Added LHL, PUA codes, 2 sec clock timing for codes.
 *
 * The ASYNC routines are used to provide buffered serial communications
 * with the DSP.  These are freeware from
 *	David Kessner
 *	david@kessner.denver.co.us
 *	1135 Fairfax
 *	Denver CO  80220
 * Many thanks for use of these routines.
 *
 ************************************************************************/

/* 1.62 Rearranged LHL Code. 1.63 Added CW key, 1.64 Space in LHL single tone
 * 1.67 Repeater Offsets, Programmable colors, CW text moved up.
 * 1.70 Added EME-2 mode
 * 1.71 Fixed clock x2 problem and Error 201 underflow problem.
 * 1.72 Fix lockup in PUA43
 * 1.80 Corrected timing in LHL7
 * 1.90 Added transmit, data box & tidyed up PUA43. Needs dsp v1.9
 *      to have correct DSP timing for all triggered modes.
 * 1.91 Update EME data in transmit for EME2, add &=1sec in CW, 1 Hz RIT
 * 1.92 Changed to dR/dt Doppler calculation
 * 1.93D Data Dump for EME2
 * 1.94 Less clock checking for slow computers 16 Aug 00 RSL
 * 1.95 Tukey25 window, Variable BW FFT, AutoNotch. ... 18 Oct 00 RSL
 * 1.96 Integrated var BW into EME2. Added eme doppler to SSB, CW. 30 Oct00 RSL
 * 1.97 Fixed PUA43 for narrower fft BW; Added audio equalizers,
 *      filter design 12 Nov 00 RSL
 * 1.98 Added long term integration to USB 9 Dec 00 RSL
 * 1.99 Converted to LTI, Added Filter 8 design. 22 Jan 00. RSL
 * 2.00 Bugs fixed, error messages.  9 Feb 01  RSL
 * 2.01 NB to PUA43, Top5 List on PUA43 6 May 01  RSL
 * 2.02 Worked over comments. Removed extra variables.  8 May 01 RSL
 */

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <math.h>
#include <float.h>
#include <graphics.h>
#include <string.h>
#include <dos.h>
#include <bios.h>
#include <time.h>
#include <alloc.h>
#include <signal.h>
#include "util.h"
#include "uhf_kbd.h"
#include "lbox.h"
#include "u_code.h"
#include "async.h"
#include "moonsun.h"
#include "umatrix.h"
#include "ufilter.h"
#include "u_cmd.h"
#include "uhfa.h"

/* KGAIN makes the S-Meter for non-FM correct and reflects the nominal gain
 * of the transceiver. Detailed corrections come from the >CFG variable
 * gain_adj and for external preamps/transverters from tvtr[ ].gain_db.
 */
#define KGAIN -158.8

/* LOG_OFFSET is sent to dsp to offset log data before
 * conversion to 8 bits. Every 4096 is 10 dB.
 */
#define LOG_OFFSET 12288
#define LOG_OFFSET_D 12288.0

#define UDUMP 0

void update_data(void);
void prep_display(void);
void print_stars(void);
void draw_grid(void);
void sa_graph(void);
void mark_frequency(int);
void t_r_toggle(void);
void new_freq(void);
void set_afgain(int);
void set_rfgain(void);
void set_mike_gain(void);
void set_fm_squelch(void);
void set_cwoffset(void);
void set_eme_doppler(void);
void show_er(void);
void set_fsk(int);
void set_xmit_power(void);
void set_contrast(void);
void set_brightness(void);
void set_window(void);
void set_affilt_sw(void);
void set_affilt_select(void);
void set_delay_right(int);
void filter_design(void);
void select_denoise(void);
void set_denoise(void);
void annotate_af_filt(void);
void set_dot(void);
void set_display_width(void);
void set_disp_offset(int);
void set_db_div(int);
void set_noncohave(int);
void set_mode(int);
void set_code_nx2(void);
void set_input(void);
void set_8_16(int);		/* 0=set to 8 bit, 1=toggle 8/16 */
void set_data_save(void);
void disp_toggle(void);
void normalize_plot(int);	/* 0=clear, 1=15 second, 2=10 min */
void normalize(int);		/* Do the normalization calcs, 256 or 512 */
void lt_normalize(void);
void dofreqplot(int);
void fsk2sec(void);		/* Sends 2 sec fsk tone */
void set_clock(int);		/* Manual set of DSP clock (NOT PC clock) */
long int  update_clock(int);	/* Set/get start of next second */
void diagnostics(int);
void freq_data(int, long int, long int, long int, long int, long int,
               unsigned int);
void initialize(void);
void presets(void);
void do_transmit(void);
void eme_timing(void);		/* Poles eme_state after times all */
void lhl7_timing(void);
void do_auto_disp(int k);
int  read_config_file(void);
void save_state(void);
void save_lt_pwr(void);
void new_band(int);
void rit_toggle(void);
void display_summary(char*, char*);
void print_time(void);
void send_shift_data(unsigned long);  /* Load shift regs and synth from dsp */
void tiny_time(int, int, int);
void tiny_char(int, int, int);
void micro_char(int, int, int);
void point1(int, int);
void point5(int, int);
void beep(int);
int  near_int(double);
void bar_graph(int, int, unsigned int);
int  get_kbd_int(void);
int  keyident(void);
void get_mem(void);
void set_agc(int);
void set_cw_speed(void);
void draw_panel(void);
double p2db(double);
double upow10(double);
void get_dsp_ptrs(void);
void get_dsp_mem(int what, int* tbl, int item, int offset);
void set_dsp_mem(int what, int adr, long int data);
int  get_status_bytes(int type, double timeout);
double update_moon(struct tm *ptm, int do_now);
void update_sun(struct tm *ptm, int do_now);
void set_cur_loc(int);
void set_bins(void);
void do_t_error(int, char*, int);

/******************   Global Variables ******************/
char		*big_area_ptr1;
char		*big_area_ptr2;
char		cc[256];		/* General purpose - use in any fcn */
int		contrast=18;
double		dis_gain=0.0525;
int		bright=64;
double		dis_offset=0.0;
int		power_exp=10;	/* Power of 2 that dsp data is divided by
				   before integration; must be (0, 255).  */
int		scroll_line=0;
/* T/R is controlled either by PTT on rf board or by 'Home'. The following
   three variables allow either but not both at any time.   */
unsigned int	transmit=0;		/* 1=transmit, 0=receive */
unsigned int	ptt=0;			/* PTT: 1=transmit, 0=receive */
unsigned int	home_tr=0;		/* 'Home' key: 1=transmit, 0=receive */
unsigned int	cw_on=0;		/* 0=cw off, 1=cw on */
int	rfgain=0;			/* (0, 6) for 0, -6, ... -36 dB */
int	xmit_power=100;
int	mike_gain=95;
int	afgain=25;
int	fm_afgain=25;
int	cwoffset=600;
int     fsk=0;				/* Xmtr tone LHL7, PUA43, EME2, LTI */
int	sidetone_lev=80;
int	fm_squelch=66;
int	noncohave=2;
int	d_noncoh=0;			/* Communicate change -1 or +1 */
double	fft_per[4]={1.5, 3.0, 6.0, 8.0};  /* Equiv FFT per data set from DSP */

/* Non-windowed bandwidth for FFT at 2400, 4800, 9600 & 48000 samp/s */
double  bin_width[4]={2.34375, 4.6875, 9.375, 23.4375};

/* Relative noise bandwidths of the windowing functions */
double  noise_bw[4]={1.0, 1.10, 1.36, 2.0};

int     win_offset=0;			/* 3 dB per offset step, DSP spectrm*/
int	do_window=1;
int	do_window_save=1;		/* When changing modes */
int	fir_save=0;			/* When changing modes */
int	fft_bw=2;			/* 0=1200, 1=2400, 3=4800 */
int	fft_bw_save=2;
/* Bit 10 of the fft_fcn is a software trigger of the FFT data collection.
 * If it is set to 1 and the "F" command is sent to the DSP it will trigger
 * the data collection. It is reset by the DSP and doesn't need to be set
 * explicitly to 0 by the control program. Software trigger is in effect
 * when bit 9 is 1.
 */
unsigned int    fft_fcn=0X0044;		/* Free running, 4800 BW, 1024 point */

double  win_adj_loss=0.0;		/* Correction for window loss */
double  win_adj_residual=0.0;		/* After adjusing pwr_exp for win fcn*/
double  win_adj=0.0;			/* Applied to display and S-meter */
double	knoise=1.0;			/* Spectral correction for BW */
double  kfilt=1.0;			/* Gain correction in FIR's */
int     do_af_filt=FALSE;		/* 0=no filt, 1=af filt on */
int	af_coef=0;			/* 0=filt[0],... */
int     do_LMS=0;			/* 0=no LMS, 1=denoise, -1=auto-notch*/
int	denoise_level=40;
int	denoise_decay=70;
int	denoise_del_len=1;		/* 61 long */

int	autonotch_level=50;
int	delay_state=0;			/* 0=No binaural delay */
int	delay_right=9;			/* Power of 2 x 20.8usec on right ch */
int	agc_decay=5;
int	last_agc=5;			/* Used at init & by set_agc() */
int	dot_type=2;
int	display_width=FULL_FREQ;
double	freq_low=200.0;			/* Lowest frequency for marker */
int	bin_low;			/* Derived from freq_low */
int	db_div_set=0;			/* 0=10db/div, 1=5db, 2=2db, 3=1db */
double	kdb_div=3.2;			/* 80 div/256 values */
double	koffset=0.0;			/* Display offset in dB */
int	mode=USB;
int     mode_mask=91;			/* Sums of bits:
					 * 1=CW, 2=USB, 4=LSB, 8=FM, 16=LHL7
					 * 32=PUA43 64=EME2 128=LTI */
int	narrow=1;
int	binaural=0;
int	iq=1;
int	bits_8_16=0;			/* Start 8 bit */
int	save_8_16=0;			/* Used when switching modes */
int	input_if_af=0;			/* Start as rf in */
					/* Next var allows for operating with
					   non-standard hardware. Set TRUE only
					   if std hardware available */
int	hardware=TRUE;			/* RF Hardware present */
int	hardware_ptt=FALSE;		/* Use RI of Serial Port for PTT */
int	normalize_on=FALSE;		/* 0=off, 1=on, >1 is countdown */
int	auto_disp_on=FALSE;		/* 0=off, 1=on  */
int	rit=0;
int	rit_save=0;
int	rit_on=FALSE;
int	band=1;				/* Band 1 = 144 MHz */
long	tfreqL   =  144200000;

/* The following are personality variables that are set through either the
 * default value, or if present, by the value in the .CFG file.
 * They cannot be set by direct keyboard commands */

/* COLORS:   */
struct  palettetype  savpalette;
struct	palettetype  upalette=
			{16,0,16,34,3,25,41,21,29,61,60,52,46,38,54,55,63};
int c_background=4;
int c_outline=11;
int c_outline_bold=15;
int c_outline_bright=8;
int c_text=11;
int c_text_bold=15;
int c_text_bright=13;
int c_text_note=9;
int c_plot_bold=15;
int c_plot_bright=13;
int c_plot_marker=9;
int c_bar_dark=1;
int c_bar_low=15;
int c_bar_medium=2;
int c_bar_high=9;

int kbd_alt1=FALSE;			/* 10 Hz/100Hz Steps */
int kbd_alt2=FALSE;			/* Increased Scrl locks */
int kbd_alt3=FALSE;			/* RF Gain to Scrl F9, F10 */

int	startup_detail=0;		/* 0=No detail, 1=Show diagnostics
					   Set =1 in case .cfg missing */
int	waterfall=1;			/* 0=Fancy freq dial, 1=Waterfall */
int	time_type=0;			/* 0=GMT, 1=local */
char	time_str[10]="Z";		/* Change to "PST" etc from .CFG file*/
int	show_secs=FALSE;		/* Seconds on waterfall time */
char	call_str[10]="   ";		/* Enter from .CFG file */
unsigned char code_id[10]="DE ?";       /* ID when sending fsk code */
int	code_nx2=1;			/* Number of 2 sec per tone */
int	d_nx2=0;			/* Communications for code_nx2 change*/
int	ms_available=TRUE;
double  millisec;			/* Corrected count of millisec */
double  millisec_offset=0.0;		/* Used in clock setting to line up
					   ticks */
int	tick_time=10;			/* Audable tick for time fcns */
double  clock_cor_factor=1.000000000;	/* Relative speed of PC clock */
long int  dsp_clock_base=0L;		/* Offset to make millisec valid */
long int  cl_ms;			/* Results from last clock() */
int	clock_been_set=FALSE;		/* Error for codes, if not set */
int	quit_save_state=TRUE;		/* TRUE=save state on quit() */
int	beep_on_error=TRUE;		/* When error codes are displayed */
int	error_on_tr=2;			/* Prevent error after T/R change */
unsigned int errcode;
struct	transverter_info tvtr[21];

/* Following 3 variables control the synthesizer, dsp 2nd lo and hardware
   setting of the shift registers.  This needs to be collected into
   a structure for neatness, some day!   */
unsigned int	dsp_deltap_lo=0XA6;	/* Phase incr dsp, low 8 bits */
unsigned int	dsp_deltap_hi=0X4A;	/* Phase incr dsp, high 8 bits */
/* shift_reg:  Bit 0=CW on, 1=trans, 2=recv, 3=PA enable, 4=Ant relay transmit,
   High 8 bits come from rfg[i].gainbyte    */
unsigned int	shift_reg=0X04;		/* Low 8 bit hardware control */

int	dsp_status[18];
int	dsp_xmit_status[10];
double	ave_spectral_pwr=0.0;		/* Ave if color agc on */
double	ref_offset=1.0;			/* Correction factor for freq ref */
int	integrate_count=0;		/* Varies from 0 to noncohave  */
int	contrast_line_count=0;		/* Display contrast line if >0 */
int	cw_speed=10;			/* Index of speed (0, 25) */
int	display_error_count=0;
int	diag_mode=2;			/* 0=none, 1=text, 2=bargraphs
					   3=Moon, 4=Sun data */
int	do_data_save=0;			/* 0=no save to file, 1=save */
int	text_area=FALSE;		/* TRUE if y=(0,32) for text */
unsigned int	com_port;
int	lpt_port=0;			/* Zero means 'not used' */
unsigned int	ntau=200;		/* Count for 20 usec delay */
int	incount;
int bar_color[]={15, 15, 15, 15, 15, 2, 2, 2, 2, 9, 9, 9, 9, 9, 9};
double	fgain[101];			/* 'Exact' gain value */
int	fgainl[101];			/* Low 8 bits of gain value */
int	fgainh[101];			/*   and high 8 bits */
int	raw_data[515];
double*	dnorm;				/* [515] calloc() */
double* lt_dnorm;			/* Used in EME, etc [256] calloc() */

double min_lt_dd=1.6;			/* Yellow trace scale limit */

/*  VVVVV No MORE */
int    lterm_int=FALSE;			/* LTI available */
double lterm_int_flow=200.0;		/* Start freq for yellow trace */
double lterm_int_fhigh=700.0;		/* End freq, for endless integratn */
double lterm_int_center=500.0;		/* Ctr of 21 freq in "Alt A" box */
double lti_frandom=0.0;			/* Current random offset for LTI */
double lti_freq_range=0.0;		/* Total spread of the random freq */

double	normalize_num;
double  eme2_te=290;			/* Effective Noise Temp Kelvins */
double  lhl7_state=0;			/* 0=waiting for trig, 1=triggered */

double*	pwrdat;				/* An array of 515 for power from
						dsp fft.  Uses calloc() */
int	pua14_28=1;			/* 0=14 char, 1=28 char */
int	code_min=0;			/* Minute of day (0, 1439) */
int*	pua43_rnd;			/* Array of 1440 rand num (key) */
double	pua43_noisedecay=0.999;
int	pua43_dorandom=TRUE;
char	pua43_xstring[35]="                            "; /* For PUA43 (28) */
/* The following two parameters specify the PUA43 frequency arrangement
 * as a function of the fft_bw parameter.     */
double	df[3]={2.34375, 4.6875, 9.375};	/* Bin spacing in Hz */
int	b0[3]={192, 96, 48};		/* Bins of offset to first freq, 450 */

double  pave;
double	paveave=0.0;			/* Rng ave of pave in update_disp() */
double*	lt_pwrdat;			/* 'Perpetual' power accumulation
					   [256] from calloc() */
double *pd16;				/* Power for 16 bit data */
double	lt_max, lt_min;
double* power10;			/* 256 entry table for converting log
						to power. Uses calloc() */
int	dly_ant2amp=50;			/* T/R Sequencing delays */
int	dly_amp2xmit=75;
int	dly_xmit2amp=25;
int	dly_amp2ant=25;

FILE*	data_file;
FILE*	config_file;
FILE*   license_file;
FILE*	lt_pave_file;

/* A table of pointers for important locations in DSP memory.
 * Table is downloaded from DSP at initialization by
 * void get_dsp_ptrs();
 * For the ADSP2181 without external memory, these are (0, 16383).
 * The relevant dsp symbolic addresses are:
 *   dsp_ptr[0]   ham_win
 *   dsp_ptr[1]   bh4_win
 *   dsp_ptr[2]   tukey_win
 *   dsp_ptr[3]   fir1_coefs
 *   dsp_ptr[4]   fir1a_coefs
 *   dsp_ptr[5]   wfir2_coefs
 *   dsp_ptr[6]   fir2_coefs
 *   dsp_ptr[7]   fir3_coefs0
 *   dsp_ptr[8]   fir3_coefs1
 *   dsp_ptr[9]   fir3_coefs1
 *   dsp_ptr[10]   fir3_coefs1
 *   dsp_ptr[11]   fir3_coefs1
 *   dsp_ptr[12]   fir3_coefs1
 *   dsp_ptr[13]   fir3_coefs1
 *   dsp_ptr[14]   fir3_coefs1
 *   dsp_ptr[15]   fir4_coefs
 *   dsp_ptr[16]  fir5_coefs
 *   dsp_ptr[17]  fir6_coefs
 *   dsp_ptr[18]  fir7_coefs
 *   dsp_ptr[19]  fir8_coefs
 *   dsp_ptr[20]  NU
 *   dsp_ptr[21]  NU
 *   dsp_ptr[22]  fir1200_coefs
 *   dsp_ptr[23]  fir2400_coefs
 *   dsp_ptr[24]  hilbert1_coeff
 *   dsp_ptr[25]  hilbert3_coeff
 *   dsp_ptr[26]  NU
 *   dsp_ptr[27]  NU
 *   dsp_ptr[28]  NU
 *   dsp_ptr[29]  NU
 *   dsp_ptr[30]  xmit_status_buf
 *   dsp_ptr[31]  cmd_buf
 *   dsp_ptr[32]  esc_a
 *   dsp_ptr[33]  filter_data
 *   dsp_ptr[34]  buf_len_table
 *   dsp_ptr[35]  NU
 *   dsp_ptr[36]  filter_id
 *   dsp_ptr[37]  0X123456 Check number
 */
int dsp_ptr[38];
int dsp_data_index=0;			/* Keeps track of data from DSP */
int table_adr=0;			/* DSP address of pointer table, or
					 * =0 if error in getting table */

/* The following variables and the function pe() are debugging tools */
double dt1=1.0;
double dt2=2.0;
double dt3=3.0;
double dt4=4.0;
double dt5=5.0;
void pe(int);

struct  rf_gn  rfg[] = {0.00,  0x88, 100,
			-6.0,  0x48, 94,
			-12.0, 0x28, 88,
			-18.0, 0x18, 82,
			-24.0, 0x14, 76,
			-30.0, 0x12, 70,
			-36.0, 0x11, 64,
			-99.0, 0x00, 00};
double	gain_adj=0.0;			/* dB factor to correct pwr disp */
/* Values for s-meter bar graph. First is for -140 dBm. Last is -27dBm */
int smbar[114]={1, 1, 1, 1, 1, 1,			/* S1 */
		2, 2, 2, 2, 2, 2,			/* S2 */
		4, 4, 4, 4, 4, 4,			/* S3 */
		8, 8, 8, 8, 8, 8,
		16, 16, 16, 16, 16, 16,
		32, 32, 32, 32, 32, 32,
		64, 64, 64, 64, 64, 64,
		128, 128, 128, 128, 128, 128,
		256, 256, 256, 256, 256, 256,		/* S9 */
		512, 512, 512, 512, 512, 512, 512, 512, 512, 512,
		1024, 1024, 1024, 1024, 1024, 1024, 1024, 1024, 1024, 1024, 
		2048, 2048, 2048, 2048, 2048, 2048, 2048, 2048, 2048, 2048,
		4096, 4096, 4096, 4096, 4096, 4096, 4096, 4096, 4096, 4096,
		8192, 8192, 8192, 8192, 8192, 8192, 8192, 8192, 8192, 8192,
		16384, 16384, 16384, 16384, 16384,
		16384, 16384, 16384, 16384, 16384};

int	screen_save=TRUE;   /* FALSE; */

/* For moonsun we need the data on the home location and dx.  To allow multiple
 * dx stations there is provision for loc[0] which is always used
 * and up to 9 'dx' locations, one of which could be loc[0] for self echoes.
 * These are normally defined in the .cfg file.
 */
struct	location loc[10];
int	cur_loc=0;
struct	eenv	 cur_env;		/* Cur enviornment:day, nutation, etc*/
int	ut_available=0;			/* 0 if not avail, else 1 */
double	ut=0.0;				/* UT for moonsun calculations */
double  dut1=0.0;			/* UT1-UTC if known, else 0.0 */
struct  tm       tdt;			/* TDT for ephereris calculations */
double	f_eme=144200000.0;		/* Dopplr Fr (audio processing only) */

struct  lhl  lc0;			/* Info on LHL codes from u_code */
int	lhl_bins=1;			/* 1, 3, 5, 7 or 9 */
int	lhl_peak_ave=1;			/* Default to ave */
int	lhl7_2line=0;			/* 0 is 1 big line, 1 is two small */
int	sync_flag=FALSE;		/* For time aligning displays */

/* For EME2 and LTI modes we need:     */
int     lt_i1=75;
int     lt_i2=115;
int     lt_i3=125;
int     lt_i4=135;
int     lt_i5=175;
long int    eme_pulse_count=0;		/* The number of 5 second periods */
double      eme_freq_range=200.0;	/* Total random range, Hz */
double      eme_pwr_sum[21];		/* Long term pwr +/- 10 bins */
int         eme_freq_ctrl=1;		/* 0=fixed, 1=random */
long int    eme_millisec_base=0;	/* Starts at 0, 5000, 1000, etc */
int         eme_state=-1;		/* -1=standby
					    0=transmit, still processing data
					    1=transmit, display complete
					    2=switch to rcv, waiting for 2.5sec
					    3=trigger rcv in dsp, get data
					    4=special case- first xmit */
int         eme_record=FALSE;		/* FALSE=off, TRUE=21 freq / 5 sec */
double      code_nb_ratio=1.2589;	/* To throw out impulse noise */
int         dly_emet2r=400;		/* Millisec */
int         eme2_format=7;		/* Bit  0: 1=doppler correct
					 *      1: 1=use random freq
					 *      2: 1=use auto noise blanker */
/* Note: in the interest of  confusion, eme_doppler is used for all modes
 * except EME2.  */
int         eme_doppler=FALSE;		/* FALSE=no correct, TRUE=offset rcvr*/
double	    eme_dop1, eme_dop2;		/* Due to E rotation, E-M move */
double	    eme_dop3, eme_dop4;		/* Due to RA, Dec move */
long int    random_freqL=0L;		/* Used in codes to suppress birdies */

unsigned int mat_err;			/* Details on matrix errors go here */

double  flow=300.0;			/* Parameters for filter design */
double  fhigh=2600.0;
double  kaiser_db=50.0;
int     filter_mask=255;		/* Bit-by-bit for 8 filters */
int     fil8_use=TRUE;			/* Apply the filter */
char    fil8_name[7]="FIL8";		/* Name for screen */
/* Parameters for receive equalizer that is part of filter design: */
#define	NFRB	15			/* Break frequencies (not progrmabl) */
double feqr[NFRB]={0.0,			/* Equalizer break frequencies */
              0.02083,			/* Fractions of 9600 Hz Fs */
              0.04167,			/* 0, 200, 400 Hz */
	      0.06250,			/* 600 Hz */
	      0.08333,			/* 800 Hz */
	      0.10417,			/* 1000 Hz */
	      0.13542,			/* 1300 Hz */
	      0.16667,			/* 1600 Hz */
	      0.20833,			/* 2000 Hz */
	      0.25000,			/* 2400 Hz */
	      0.29167,			/* 2800 Hz */
	      0.33333,			/* 3200 Hz */
	      0.37500,			/* 3600 Hz */
	      0.41667,			/* 4000 Hz */
              0.45833};			/* 4400 Hz */
/* Relative equalizer gains in dB (program from .cfg file): */
double aeqrdb[NFRB-1]={-100.0,		/*    0 to 200 Hz */
	       0.0,			/*  200 to 400 */
	       0.0,			/*  400 to 600 */
               0.0,			/*  600 to 800 */
	       0.0,			/*  800 to 1000 */
               0.0,			/* 1000 to 1300 */
	       0.0,			/* 1300 to 1600 */
	       0.0,			/* 1600 to 2000 */
	       0.0,			/* 2000 to 2400 */
	       0.0,			/* 2400 to 2800 */
	       0.0,			/* 2800 to 3200 */
	       0.0,			/* 3200 to 3600 */
	       0.0,			/* 3600 to 4000 */
	       0.0}; 			/* 4000 to 4400 */

struct lbox_info lb_freq; 		/* Freq info box */
struct lbox_info lb_eme2;		/* EME2 data summary box */
struct lbox_info lb_lhl7;		/* and for LHL7 */
struct lbox_info lb_pua43;		/* and for PUA43 */
struct lbox_info lb_lterm_int;		/* and long term integration of USB */
void* current_info_box=0;		/* 0 if no box, otherwise ptr */

/* Associated with the audio filters:   */
struct filter_info  filt[8];

/* ------------------------------------------------------------------------ */

void main(int argc, char *argv[])
	{
	int grdrv, grmode, colormax, i;
	int cx, new_char, last_char;
	int stop_now;
	int esc_count;
	int y1, y2;
	int check_sum_dsp;		/* Received from dsp */
	int check_sum_pc;		/* Calculated by pc frm data*/ 
	int xmax, ymax;
	char  *cptr;
	unsigned int is;
	time_t t0;
	struct tm *tp;
	char *ptc;
	char version[10];
	int ser_cnt;

	strcpy(version, "2.01");	

	cptr = &cc[0];
	stop_now=FALSE;
	/* Get command line arguments for com port */
	if (argc==1)			/* Should be 2 arguments */
	    {
	    display_summary(argv[0], version);
	    exit(0);
	    }
	sscanf(argv[1], "%d", &com_port);
	if (com_port<1 || com_port>4)
	    {
	    beep(150);
	    printf("COM port must be 1, 2, 3 or 4\n");
	    exit(0);
	    }
	if (com_port==1) com_port=COM1;
	if (com_port==2) com_port=COM2;
	if (com_port==3) com_port=COM3;
	if (com_port==4) com_port=COM4;

	/* NOTE: Read the license before deletion or modification to any
	 * of the following license related code:
	 */
        if (NULL == ( license_file=fopen("GNU_GPL.txt", "r") ))
	    {
	    printf("\nNo GNU_GPL.txt license file was found. This must be in");
	    printf(" the same directory\nas UHFA.EXE file. Exiting program.");
	    exit(1);
	    }
	fclose(license_file);
	license_info();
	for(;;)
	    {
	    if ( kbhit() )
		{		
		cx=getch() & 0X7F;
		if (cx==13)
		    {clrscr(); break;}
		else if (cx=='1')
		    {warranty(); license_info();}
		else if (cx=='2')
		    {license(); license_info();}
		else license_info();
	        }
	    }

	if ((ptc=getenv("TZ")) != 0)
	    {
	    ut_available=1;
	    tzset();
	    printf("\nGMT Conversion from 'SET TZ=%s'", ptc);
	    time(&t0);
	    tp=localtime(&t0);
	    strcpy(cc, asctime(tp));   /* Get copy to rearrange */
	    cc[24]=0;
	    printf("\nComputer time =  %s (local)", cc);
	    tp=gmtime(&t0);
	    strcpy(cc, asctime(tp));
	    cc[24]=0;
	    printf("\n   translates to %s GMT\n\n", cc);
	    }
	else
	    {
	    ut_available=0;
	    printf("\nMoon & Sun data unavailable, needs 'SET TZ=PST8', etc.");
	    time(&t0);
	    tp=localtime(&t0);		/* time_type==0 --> use GMT */
	    strcpy(cc, asctime(tp));   /* Get copy to rearrange */
	    cc[24]=0;
	    printf("\nComputer time=%s (local)", cc);
	    }
	if (clock()==(clock_t) -1) 	/* This may not be there */
	    {
	    startup_detail=TRUE;	/* They may want to see this */
	    printf("Milli-second clock is NOT AVAILABLE,\n");
	    printf("   therefore, modes LHL7, PUA43 unavailable.\n");
	    if (mode>=4)  mode=0;		/* i.e., unavalailable */
	    ms_available=FALSE;			/* Communicate it */
	    }
	presets();			/* These may be overwritten */
	if (read_config_file() == 0)
	    {
	    printf("\nNo uhfa.cfg present, using default values");
	    }

	get_mem();			/* Memory for spectral disp scroll */
	if (mode_mask & 0X20)
	    pua43_update(0, -3);	/* Allocate memory for PUA43 *
	if (input_if_af==1) tfreqL=0;	/* Prevent inconsistency when in af */

	/* Download the DSP pointer table */
	printf("\nDownloading DSP pointers...");
	ASYNCINIT(com_port);
	ASYNCSET(BAUD_RATE, COM_SETTINGS);
	ASYNCCLEAR();
	get_dsp_ptrs();
	ASYNCSTOP();

	if(startup_detail)
	   {
	   grdrv=DETECT;	
	   detectgraph(&grdrv, &grmode);

	   printf("Graphics driver= %d, Mode %d\n", grdrv, grmode);
	   if (grdrv != 9)
	      {
	      beep(1000);
	      printf("If driver is not 9 (VGA) & Mode is not 2 (640x480) ");
	      printf("you may encounter\ngraphics driver incompatibilities. ");
	      printf("Hit ENTER to continue\n");
	      }
	   printf("Graphresult= %d,  Hit 'Enter' to proceed", graphresult() );
	   getch();
	   }

	grdrv=9;			/* Only allow VGA 640x480 */
	grmode=2;
	initgraph(&grdrv, &grmode, "");
	strcpy(cptr, getdrivername() );
	xmax=getmaxx() + 1;  ymax = getmaxy() + 1;  /* Global */
	colormax = getmaxcolor();
	is = imagesize(1, 10,  511, 179);
	if (signal(SIGINT, ctrlc_handler)==SIG_ERR)
	    {
	    printf("\nInvalid signal number for Ctrl-C handler\n");
	    exit(0);
	    }
	if (signal(SIGFPE, fpe_handler)==SIG_ERR)
	    {
	    printf("\nInvalid signal number for Floating Point ");
	    printf("error handler\n");
	    exit(0);
	    }
	if(startup_detail)
	    {
	    closegraph();
	    printf("Graphics is %s, Mode %d, %d x %d\n",
		cptr, grmode, xmax, ymax);
	    printf("Bottom image size (170 lines worth) is %u bytes\n",is);
	    printf("Max color is %d\n", colormax);
	    printf("Graphresult= %d\n\n", graphresult() );
	    printf("\nHit 'Enter' (with Scroll Lock off) to proceed.\n");
	    printf("Any other key will give 'key ident' for uhfa_kbd.h.\n");
	    printf("Exit program with Scroll Lock Alt F4. See Readme\n");
	    printf("file for list of keyboard commands.  No mouse is used.\n");
	    for(;;)
		{
		if ( kbhit() )
		    if ( keyident() )  break;
		}
	    initgraph(&grdrv, &grmode, "");
	    }
	getpalette(&savpalette);	/* Palette when we arrived */
	setallpalette(&upalette);
	if ( graphresult()==grError )
	     do_t_error(40, "E40 Graphics", 200);
	setbkcolor(c_background);
	draw_grid();
	initialize();			/* This needs to be after get_mem() */
	scroll_line = 0;

	/* The following is the main control loop for the transceiver */
 pp:	/* If contrast or brightness was recently changed, do color bar: */
	if (--contrast_line_count>0)
		{setviewport(X1L+1, Y1T+1, X1R-1, Y1B-1, DO_CLIP);
		 for (i=1; i<16; i++)
		     {setfillstyle(1, i);
		      /* y1=bottom of bar */
		      y1=128-(int)((dis_offset+((double)i)/dis_gain)/2.0);
		      if (y1>127)  y1=127;
		      if (i==1)  y1=127;
		      y2=128-(int)((dis_offset+((double)(i+1))/dis_gain)/2.0);
		      if (y2<1) y2=1;
		      if (i==15)  y2=1;
		      bar(390, y2, 415, y1);
		     }
		 setcolor(c_text_bold);
		}
	else
		contrast_line_count=0;

	esc_count = 0;
	for(;;)				/* Wait for 1st character */
		{
		if ( ASYNCINSTAT() > 0 )	/* Have ch, break if ESC_CHAR*/
			{new_char = ASYNCIN();
			 if (new_char==ESC_CHAR)  esc_count++;
			 break;			/* Break the for(;;) */
			}
		while ( kbhit() )		/* Check for commands */
		    {
		    stop_now=do_cmd();
		    if (stop_now)  quit();
		    if (transmit && mode!=EME2) do_transmit();
		    }
		/* Check for PTT via fast serial RI line */
		if ( ((0X4000 & ASYNCSTAT())==0) && hardware_ptt
			 && hardware && mode!=EME2)
		    {
		    ptt=1;		/* Set to override 'Home' key */
		    t_r_toggle();		/* Go to transmit */
		    do_transmit();		/* And wait here */
		    }
		if(mode==EME2)
		    eme_timing();	/* Transmit, etc. */
		if(mode==LHL7 || mode==PUA43)
		    lhl7_timing();	/* Receive triggering */
		}
	for(;;)			/* Wait here for ESC_CHAR A to come */
		{last_char=new_char;
		 if(ASYNCINSTAT() > 0)
			{new_char = ASYNCIN();
			 if (new_char==ESC_CHAR) esc_count++;	
			 if (esc_count==2)	/* Waiting for ESC_CHAR, 'A' */
				{esc_count=0;	/*    but got 2xESC_CHAR     */
				 goto pp;	/* So try again              */
				}
			 if ((last_char==ESC_CHAR) && (new_char=='A'))
				break;	/* The beginning of a new data set */
			}
		 while ( kbhit() )
		    {
		    stop_now=do_cmd();
		    if (stop_now)  quit();
		    if (transmit && mode!=EME2) do_transmit();
		    }
		/* Check for PTT via fast serial RI line */
		if ( ((0X4000 & ASYNCSTAT())==0) && hardware_ptt
			 && hardware && mode!=EME2)
		    {
		    ptt=1;		/* Set to override 'Home' key */
		    t_r_toggle();		/* Go to transmit */
		    do_transmit();		/* And wait here */
		    }
		if(mode==EME2)
		    eme_timing();	/* Transmit, etc. */
		if(mode==LHL7 || mode==PUA43)
		    lhl7_timing();	/* Receive triggering */
		}
	for(;;)				/* Wait for a buf full */
		{if ( ASYNCINSTAT() > 490 )	/* Almost full (was 520) */
			 break;
		 while ( kbhit() )
		    {
		    stop_now=do_cmd();
		    if (stop_now)  quit();
		    if (transmit && mode!=EME2) do_transmit();
		    }
		/* Check for PTT via fast serial RI line */
		if ( ((0X4000 & ASYNCSTAT())==0) && hardware_ptt
			 && hardware && mode!=EME2)
		    {
		    ptt=1;		/* Set to override 'Home' key */
		    t_r_toggle();		/* Go to transmit */
		    do_transmit();		/* And wait here */
		    }
		if(mode==EME2)
		    eme_timing();	/* Transmit, etc. */
		if(mode==LHL7 || mode==PUA43)
		    lhl7_timing();	/* Receive triggering */
		}
	if (--display_error_count == 1)
	    {
	    do_t_error(9999,"", 0);	/* Clears any errors */
	    display_error_count=0;	/* Don't do it until another error */
	    }
	incount=-1;			/* incount is for diagnostic display */
	i=0;  check_sum_pc=0;  esc_count=0;
	for(;;)	/* Stay here until 512 data points are received */
		{if( (ser_cnt=ASYNCINSTAT()) > 0 )
			{new_char = ASYNCIN();
			 ++incount;
			 if (new_char==ESC_CHAR) esc_count++;
			 if (esc_count==2)
				{esc_count=0;
				 goto skip_char;
				}
			 raw_data[i]=new_char;
			 check_sum_pc += new_char & 0XFF;
			 /* If buffer has a lot of stuff, just get it */
			 if (ser_cnt<1000)
			   {
			   if(mode==EME2) eme_timing(); /* Time eme */
			   if(mode==LHL7 || mode==PUA43)
			     lhl7_timing();	/* Receive trig */
			   }
			 if (++i > 511)
				{if (esc_count==1) { delay(2); ASYNCIN(); }
				 break;
				}
 skip_char:		}

		}
	for(;;)				/* Get check sum */
		{if ( ASYNCINSTAT() > 0 )
			{check_sum_dsp=ASYNCIN();
			 break;
			}
		}
	check_sum_pc &= 0XFF;
	if (error_on_tr > 0)		/* Prevent error after T/R change */
	    error_on_tr--;

	/* Now get dsp_status bytes */
	i=0; esc_count=0;
	for(;;)
		{if( ASYNCINSTAT() > 0 )
			{new_char = ASYNCIN();
			 /* First esc_char goes into buff, but not second */
			 if (new_char==ESC_CHAR) esc_count++;
			 if (esc_count==2)
				{esc_count=0;
				 goto dstat1;
				}
			 dsp_status[i] = new_char;
			 if (++i > 17) 
				 break;
dstat1:			}
		}

		/* Hold up screen update until after the next xmit pulse
		 * has started   */
		if(mode==EME2 && transmit)
		    {
		    while (eme_state==3)
			eme_timing();	/* Transmit, etc. */
		    }

	if (check_sum_pc==check_sum_dsp)
	    {
	    /* error_on_tr prevents update right after reception starts,
	     * but we don't want it to stop any EME2 data */
	    if (error_on_tr==0 || mode==EME2)  update_data();
	    }
	else
	    {
	    if (error_on_tr==0)  do_t_error(1, "E1 DSP Data Lst", 20);
	    goto pp;
	    }

	if(dsp_status[17]==2)		/* Valid status data? */
	    /* See if PTT because of ring detect in ASYNCSTAT() or because
	     * sent in dsp_status(). Not valid w/o hardware. In EME2, the
	     * T/R conttrol is automatic by eme2_timing().
	     */
	    {
	    if (( (((0X4000 & ASYNCSTAT())==0) && hardware_ptt)
			 || (dsp_status[16] & 0X10))
			 && hardware && mode!=EME2)
		{ptt=1;			/* Set to override 'Home' key */
		 t_r_toggle();		/* Go to transmit */
		 do_transmit();		/* And wait here */
		}
	    diagnostics(1);		/* Update diagnostics */
	    }
	else
	    do_t_error(2, "E2 DSP Status", 20);
	print_time();
	goto pp;	
	}				/* End main() */

/*  UPDATE_DATA()  -  Called after 512 data points if check sum is OK.
 *  Converts log to power, adds to power sums and displays on both upper fft
 *  screen as well as lower spectral map.
 */
void update_data(void)
	{
	static time_line_count=0;	/* Decides when to do time print */
	static double auto_disp_base=0;	/* Base value for setting color agc */
	static double auto_disp_correct=0.0;
	double correction;
	int i, j;
	int ix;
	int pclr;			/* Color for spectral map, 0 to 7 */
	int biggest;			/* Highest during sweep */
	int biggest_freq;
	int display_data;
	static int lt_display_data=0;	/* All lt_... are long term for eme */
	static double spandb=1.0;	/* Intermediate in scaling lt data */
	static double lt_dd=1.0;	/* db/div for lt data */
	static double lt_mid=0.0;	/* dB level for mid graph */
	/* Next is FFT gain, rel to 9600 samp/sec: */
	static double db_fft[4]={6.02, 3.01, 0.0, -6.99};
	int newy;
	static int oldx=0;
	static int oldy=128;
	int lt_newy;
	static int lt_oldx=0;
	static int lt_oldy=48;
	double scale;
	double bar_pwr, big_bp;
	int    noise_blanked;
	time_t tnow;
	struct tm *tmnow;
	double lpave;
	static double pave_print=0.0;	/* Accumulated over noncohave periods*/
	double dbm_level;		/* in i-f, non-FM mode, fft power */
	double fm_log_pwr;		/* Measured log power in FM mode */
	int pwr5n6;			/* Power from dsp_status[5] and [6] */
	double pwr5n6_dbm;		/* Converted to dbm */
	double tmp, tmp1, lt_tmp;	/* Temporary storage */
	int base;			/* For plotting */
	int lt_trace;			/* Yellow trace being used */

	/* Look for serial receive errors at DSP  */
	if (dsp_status[1])  do_t_error(3, "E3 DSP Cmd", 20);

	/* Look for PLL "Not-locked" errors in DSP */
	if (((0X08^dsp_status[16]) & 0X08) && hardware)
	     do_t_error(30, "E30 PLL UnLock", 20);

	/* Catch freq changes here, they happen once per minute.
	 * Also update EME data.
	 */
	if (mode==LTI || eme_doppler)
	     new_freq();

	lt_trace = (mode==EME2 || mode==LTI);  /* Shorthand */

	if (text_area) base=94;
	else           base=126;

	if (lt_trace)  eme_pulse_count++;

	/* Convert log raw data to power and add powers for non-coherent
	   integration.  For 8 bit data we use a lookup table, power10[],
	   to do a fast conversion.  Table is too big for 16 bit data.  */
	big_bp=1.0e-200;
	pave=0.0;			/* Spectral power average for 1 set */
	if (bits_8_16 == 0)		/* If using 8 bit log data */
	    {
	    for (i=0; i<512; i++)
		{
		bar_pwr=power10[ raw_data[i] ];
	 	pwrdat[i] += bar_pwr;	/* Non-coherent integration */
		if (bar_pwr > big_bp)   big_bp=bar_pwr;
		if (i>30) pave += bar_pwr;
		}
	    if (normalize_on > 1) normalize(512);
	    if (eme_pulse_count == 1)
		paveave=pave;
	    else
		paveave = 0.95*paveave + 0.05*pave;	/* Running ave */
	    if (mode==PUA43)
		noise_blanked=(pave>code_nb_ratio*paveave);
	    else
		noise_blanked=FALSE;
	    }				/* End 8 bit non-co ave */
	else				/* using 16 bit log data */
	    {
	    big_bp=1.0e-20;
	    for (i=0; i<256; i++)	/* Pass 1 for 16 bit data */
		{
		/* The following makes the value of 16 bit data the
		    same as that of 8 bit */
		/* Note: LOG_OFFSET_D 12288.0    0.0078125=1/128
		 * and upow10(0.03125*x) is the same as the lookup power10[x].
		 * The 2 constants, 0.0078125*0.03125=0.0002441406 
		 */
		pd16[i]=bar_pwr=upow10(0.0002441406*
		   ((double)(raw_data[2*i]+256*raw_data[2*i+1])+LOG_OFFSET_D));
		if (bar_pwr > big_bp)   big_bp=bar_pwr;
	 	pwrdat[i] += bar_pwr;	/* Non-coherent integration */
		if (i>30)		/* Dodge hum at left */
		   pave += bar_pwr;	/* Noise ave over spectrum */
		pwrdat[i+256]=0.00001;	/* Dummy out unused spectrum */
		}			/* End first pass of 16 bit data */

	    if (eme_pulse_count == 1)
		paveave=pave;
	    else
		paveave = 0.95*paveave + 0.05*pave;	/* Running ave */
	    if ((eme2_format & 4) || mode==PUA43)
		noise_blanked=(pave>code_nb_ratio*paveave);
	    else
		noise_blanked=FALSE;

	    lt_max=0.0;  lt_min=1.0E300;

	    /* For either EME2 or LTI, get data */
	    if (lt_trace && !noise_blanked)
		{
		for (i=lt_i1; i<=lt_i5; i++)
		    {
		    lt_pwrdat[i] += pd16[i]; /* Very long term integration */
		    if (i>=lt_i2 && i<lt_i4)
			{
			if(lt_pwrdat[i] < lt_min)  lt_min=lt_pwrdat[i];
			if(lt_pwrdat[i] > lt_max)  lt_max=lt_pwrdat[i];
			}
		    }
		}
	    if (normalize_on > 1) normalize(256);
	    }				/* End 16 bit non-co ave */
	if(mode==LHL7)
	    {
	    lhl7_timing();	/* Receive trig */
	    lhl7_update(dsp_status[14], 2);
	    }
	if(mode==PUA43)
	    {
	    lhl7_timing();	/* Receive trig */
	    setviewport(600, 5, 638, 15, DO_CLIP);
	    clearviewport();
	    settextjustify(LEFT_TEXT, TOP_TEXT);
	    moveto(0, 0);
	    if (!noise_blanked)
		pua43_update(dsp_status[14], 2);
	    else
		outtext("*NB*");
	    eme_pulse_count++;
	    }

	/* Adjust integrate count to line up with the top of the minute
	 * for LHL7 and PUA43, to stop leakage in bins for display
	 * when noncohave>1
	 */
	if (sync_flag && dsp_status[14]==2)
	    {
	    integrate_count=-1;		/* It will soon be ++ed to 0 */
	    sync_flag=FALSE;		/* Only once */
	    }

	pave_print += pave;

	/* If time for display...  EME2 and PUA43 update everytime integrate
	 * count gets to noncohave for fft_bw==2 (3 groups of 12) but for
	 * the narrower fft_bw the data is sent only once per 2 seconds
	 * and the display is updated each time.
	 */
	if ((++integrate_count >= noncohave)
	       || ((mode==PUA43 || mode==EME2) && (fft_bw==0 || fft_bw==1)))
	    {
	    if (auto_disp_on && (auto_disp_correct<-10.0 |
			auto_disp_correct>10.0) )
		do_t_error(10, "E10 AutoDisp>10", 0);
	    /* Switch the auto-disp stuff to pave or paveave ?? */

	    if (auto_disp_on == -1)	/* Establish base average level */
		{
		auto_disp_on=TRUE;
		auto_disp_base=0.0;
		for (i=0; i<(512-display_width*256); i++)
		    auto_disp_base += pwrdat[i];
		}
	    if ( (auto_disp_on) && (auto_disp_base>1.0e-100) )
		{
		for (i=FLOW; i<(512-display_width*256); i++) /* Cur ave lev */
		    auto_disp_correct += pwrdat[i];
		/* For display, remove dc term: */
		ave_spectral_pwr=auto_disp_correct - pwrdat[0] - pwrdat[1];
		auto_disp_correct /= auto_disp_base;  /* Make rel to base */
		auto_disp_correct = p2db(auto_disp_correct);
		}
	    integrate_count=0;
	    biggest=0; biggest_freq=0;
	    prep_display();
	    /* scale accounts for non coherent averaging */
	    scale=p2db( (double)noncohave );

	    if (lt_trace)	    
		oldx=2*(lt_i1-1);
	    else
		oldx=0;
            oldy=128;
	    lt_oldx=2*(lt_i1-1);   /* <<< Is this right?? */
	    /* -3*power_exp is a correction for data scaling in DSP */
	    correction = 32.0 - koffset + scale + auto_disp_correct
		+ knoise*db_fft[fft_bw]
		- win_adj - 3.0*(double)power_exp;

	    if(display_width==FULL_FREQ)
		{
		setcolor(c_plot_bold);
		for (i=0; i<512; i++)
		    {
		    display_data = (int)
			( kdb_div*(p2db(pwrdat[i]) - dnorm[i]
			     - correction) );
		    if (text_area)
			setviewport(X1L+1, Y1T+33, X1R-1, Y1B-1, DO_CLIP);
		    else
			setviewport(X1L+1, Y1T+1, X1R-1, Y1B-1, DO_CLIP);
		    if (dot_type==0)
			putpixel(i, base-(display_data >> 1), c_plot_bold);
		    if (dot_type==1)
			point1(i, base-(display_data >> 1) );
		    if (dot_type==2)
			{
			newy = base-(display_data >> 1);
			line(oldx, oldy, i, newy );
			oldx=i;
			oldy=newy;
			}
		    if (waterfall)
			{
			setviewport(X2L+1, Y2T+1, X2R-1, Y2B-1, DO_CLIP);
			pclr=(int)
			    (dis_gain*((double)display_data - dis_offset));
			if (pclr>15) pclr=15;
			if (pclr<1) pclr=1;
			putpixel(i, scroll_line, pclr);
			}
		    /* bin_low*9.375 is lowest freq considered for marker*/
		    if (display_data > biggest  &&  i>bin_low )
			{
			biggest=display_data;
			biggest_freq=i;
			}
		    }		/* End 512 times */
		}		/* End FULL_FREQ display */
	    else 		/* if (display_width==HALF_FREQ) */
		{
		/* Set the long term integration plot so that it occupies about
		   half of  the display area. lt_dd is dB/div.  There are 16
		   pixels, vertically, per major grid division.  */
		if (lt_trace)
		   {spandb=p2db(lt_max/lt_min);
		    if     (spandb<0.2)  lt_dd=320.0;	/* .05dB/16 pixels */
		    else if(spandb<0.4)  lt_dd=160.0;	/* .1dB/16  */
		    else if(spandb<0.8)  lt_dd=80.0;	/* .2dB/16  */
		    else if(spandb<2.0)  lt_dd=32.0;	/* .5dB/16  */
		    else if(spandb<4.0)  lt_dd=16.0;	/*  1dB/16  */
		    else if(spandb<8.0)  lt_dd=8.0;	/*  2dB/16  */
		    else if(spandb<20.0) lt_dd=3.2;	/*  5dB/16  */
		    else                 lt_dd=1.6;	/* 10dB/16  */
		    /* See if allowed to overrange display */
		    if (lt_dd<min_lt_dd)  lt_dd=min_lt_dd;

		    /* Set the mid-point of the display for long term
		     * integration at the average of noise in lt_i3+/-2.
		     * lt_mid, p2db() and lt_dnorm all in dB.
		     * lt_i3 is the index of the center bin for EME2, LTI.
		     */
		    lt_mid=0.5*( p2db(lt_pwrdat[lt_i3-2])-lt_dnorm[lt_i3-2]
			        +p2db(lt_pwrdat[lt_i3+2])-lt_dnorm[lt_i3+2] );
		    setviewport(150, 52, 230, 63, DO_CLIP);
		    if (noise_blanked)
			sprintf(cc, "*NB*");
		    else
			sprintf(cc, "%4.2lfdB/", 16.0/lt_dd);
		    setcolor(c_plot_bright);
		    moveto(1,1);
		    settextjustify(LEFT_TEXT, TOP_TEXT);
		    outtext(cc);
		    setcolor(c_text_bold);
		   }
		setcolor(c_plot_bold);
		for (i=0; i<256; i++)
		    {
		    /* EME2 & LTI only displays a subset of the data */
		    if (lt_trace && (i<lt_i1 || i>=lt_i5))
			goto end256;

		    if (!lt_trace)
		      {
		      display_data = (int)
			   (kdb_div*(p2db(pwrdat[i])-dnorm[i]-correction));
		      }
		    else  /* If lt_trace (EME2 or LTI)  */
		      {
		      /* Kinda a kluge (lt_i3-2);
		       * can we do better?? If random<eps otherwise don't
                       */
                      display_data = (int)
			  ( kdb_div*(p2db(pwrdat[i])-correction
                          -lt_dnorm[i]+lt_dnorm[lt_i3-2]) );
		      lt_display_data = (int)
			  ( lt_dd*(p2db(lt_pwrdat[i])-lt_dnorm[i]-lt_mid) );
		      }
		    if (text_area)
			setviewport(X1L+1, Y1T+33, X1R-1, Y1B-1, DO_CLIP);
		    else
			setviewport(X1L+1, Y1T+1, X1R-1, Y1B-1, DO_CLIP);

		    if (dot_type==0)
		       putpixel(2*i,base-(display_data >> 1), c_plot_bold);
		       putpixel(2*i+1,base-(display_data>>1), c_plot_bold);
		    if (dot_type==1)
			point1(2*i, base-(display_data >> 1) );
		    if (dot_type==2)
			    {
			    newy = base-(display_data >> 1);
			    line(oldx, oldy, 2*i, newy );
			    oldx=2*i;
			    oldy=newy;
			    }
		    if (lt_trace && !noise_blanked)
			{
			/* Display the perpetual average in yellow */
			lt_newy = 48-lt_display_data;
			setcolor(c_plot_bright);
			if (i==lt_i1)
			    {
			    line(2*i, lt_newy, 2*i, lt_newy );
			    lt_oldy=lt_newy;
			    }
			else
			    {
			    line(lt_oldx, lt_oldy, 2*i, lt_newy );
			    lt_oldy=lt_newy;
			    }
			lt_oldx=2*i;
			setcolor(c_plot_bold);
			}
		    if (waterfall)
			{
			setviewport(X2L+1, Y2T+1, X2R-1, Y2B-1, DO_CLIP);
			pclr=(int)
			    (dis_gain*((double)display_data - dis_offset));
			if (pclr>15) pclr=15;
			if (pclr<1) pclr=1;
			putpixel(2*i,   scroll_line, pclr);
			putpixel(2*i+1, scroll_line, pclr);
			}
			/* bin_low*9.375 is lowest freq for marker*/
			if (display_data>biggest  &&  i>bin_low )
			    {
			    biggest=display_data;
			    biggest_freq=i;
			    }
end256:
		    }			/* End of 256 times */
	        }			/* End of  else if (half width)

	    if(mode==LHL7 || mode==PUA43)
		 lhl7_timing();	/* Receive trig */
	    /* Add marker at highest point in upper spectral display */
	    setcolor(c_plot_marker);

	    if (text_area)
		setviewport(X1L+1, Y1T+33, X1R-1, Y1B-1, DO_CLIP);
	    else
		setviewport(X1L+1, Y1T+1, X1R-1, Y1B-1, DO_CLIP);

	    if (!text_area && biggest > 250) biggest=250;
	    else if (text_area && biggest > 186) biggest=186;
	    point5((1+display_width)*biggest_freq, base-(biggest >> 1) );

	    if(mode==CW)
		{/* Add red line at cw offset */
		 i=(int)((double)cwoffset/bin_width[fft_bw]);
		 line(i, base-6, i, base+2);
		}
	    else if(mode==LHL7)
		{i=16*(1+display_width);
		 line(4*i-1,  78, 4*i-1,  96);
		 line(5*i-1,  88, 5*i-1,  96);
		 line(6*i-1,  88, 6*i-1,  96);
		 line(7*i-1,  88, 7*i-1,  96);
		 line(8*i-1,  88, 8*i-1,  96);
		 line(9*i-1,  88, 9*i-1,  96);
		 line(10*i-1, 78, 10*i-1, 96);
		}
	    else if(mode==PUA43)	/* Red lines showing band */
		{
		/* Always FULL_WIDTH and band is 42*4=168 bins wide */
		i=b0[fft_bw];
		line(i, 87, i, 94);
		line(i+168, 87, i+168, 94);
		line(i, 94, i+168, 94);
		}
	    else if(lt_trace)
		{
		 i=(1+display_width)*lt_i3;	/* Center freq */
		 line(i, 48, i, 128);
		 i=(1+display_width)*lt_i2;
		 j=(1+display_width)*lt_i4;
		 line(i, 120, i, 128);
		 line(j, 120, j, 128);
		 line(i, 126, j, 126);
		 if (current_info_box == &lb_eme2)
		    {eme2_update(1);
		     eme2_update(2);
		    }
		 if (eme_state > -1)  eme_state=1; /* Screen update finished */
		 if (current_info_box == &lb_lterm_int)
		    {lterm_int_update(1);
		     lterm_int_update(2);
		    }
		}			/* End lt_trace */
	    else if(mode==LTI)
		{

#if 0
		line(128+lt_i2, 119, 128+lt_i2, 126);
		line(128+lt_i4, 119, 128+lt_i4, 126);
		line(128+lt_i2, 126, 128+lt_i4, 126);

		/* Always FULL_WIDTH and band is 42*4=168 bins wide */
		i=b0[fft_bw];
		line(i, 119, i, 126);
		line(i+168, 119, i+168, 126);
		line(i, 126, i+168, 126);
			/* Needs LT bounds  */
#endif
		}

	    setcolor(c_text_bold);

	    /* Several things need tmnow, so do it here: */
	    if (ms_available)
		tnow=update_clock(1)+dsp_clock_base;
	    else
		(void)time(&tnow);
	    if (time_type)
		tmnow = localtime(&tnow);	/* time_type==1 for local */
	    else
		tmnow = gmtime(&tnow);		/*          ==0 for gmt */

	    if ( waterfall && (time_line_count++ >= 20) )  /* Mark the time: */
                {time_line_count = 0;
		 if (scroll_line > 9)	/* Not above box 2 */
		    {
		    setviewport(X2L+1, Y2T+scroll_line-7,
			X2L+33, Y2T+scroll_line+1, DO_CLIP);
		    clearviewport();
		    if (show_secs)
			{
		        tiny_time( (100*tmnow->tm_min+tmnow->tm_sec), 1, 8);
			}
		    else
			{
		        tiny_time( (100*tmnow->tm_hour+tmnow->tm_min), 1, 8);
			}
		    }
		}
	    pwr5n6 = 256*dsp_status[6]+dsp_status[5];
	    if (mode==FM)
		{
		/* In following, the value in status[5] and [6] is
		  1000*log10(pwr) and the bargraph routine wants
		  abs(max volts).  Variable gain_adj comes
		  from uhfa.cfg file and allows compensation for gain variation
		  between units. It is in dB and can be + or -.
		  The 4.0* factor moves the display over 2 bars so that a max
		  signal of about -54 dBm (rf gain full up) lights all s-meter
		  bars  */
 		ix = pwr5n6 + (int)(100.0*gain_adj)
		    -(int)(100.0*(rfg[rfgain].dbgain+tvtr[band].gain_db));
				  /* This last was i-f pwr in db*100 */
		if (dsp_status[17]==2)		/* If status data good */
		    {
		    /* S-Meter bar graph: */
		    setviewport(0,S_METER_POS-19,127,S_METER_POS-9,DO_CLIP);
		    tmp1=4.0*sqrt(upow10(((double)(3703+ix))/1000.0));
		    if (tmp1>32767.0) tmp1=32767.0;
		    bar_graph(20, 0,(unsigned int)tmp1);
 		    }
		/* 93.00 relates to the gain of the radio */
		fm_log_pwr = -94.2+((double)ix)/100.0;
		}			/* End mode==FM  */
	    else if (input_if_af==0 && mode!=FM) /* RF, but not FM */
		{
		/* For the S-meter display */
		dbm_level = KGAIN-rfg[rfgain].dbgain
		    +p2db(pwrdat[biggest_freq])+gain_adj+win_adj
		    -db_fft[fft_bw]+3.0*(double)(power_exp-10)
		    -tvtr[band].gain_db - scale;

		/* Mark the audio power calculation with a 'P' */
		setviewport(113,S_METER_POS+12,122,S_METER_POS+31,NO_CLIP);
		clearviewport();
		/* Strong signals go out-of-range in the FFT output. To
		 * make the S-meter work for these signals, we switch
		 * to the rms audiio power calculation.
		 */
		if (dbm_level >-100.0)
		    {
		    pwr5n6_dbm = -86.95 + 0.01*((double)pwr5n6) + gain_adj
			-rfg[rfgain].dbgain - tvtr[band].gain_db;
		    dbm_level=pwr5n6_dbm;
		    settextjustify(RIGHT_TEXT, TOP_TEXT);
		    settextstyle(0, 0, 1);
		    setcolor(c_text_bold);
	            moveto(8, 0);
		    outtext("P");
		    }		    

		/* S-Meter bar graph: */
		setviewport(0, S_METER_POS-19, 127, S_METER_POS-9, DO_CLIP);

		tmp1=(int)(144.0+dbm_level);
		if (tmp1 < 0)
		    tmp1=0;
		else if (tmp1>113)
		    tmp1=32768;
		else
		    tmp1=smbar[tmp1];
		bar_graph( 20, 0, tmp1);
		} 			/*End RF, not FM */
	    else			/* If af input, not FM */
		{
		/* S-Meter bar graph: */
		setviewport(0, S_METER_POS-19, 127, S_METER_POS-9, DO_CLIP);
		bar_graph(20, 0, (unsigned int)(3.0*sqrt(big_bp)) );
		}
	    /* Display power below bar graph: */
	    setviewport(1,S_METER_POS+12, 112, S_METER_POS+31, NO_CLIP);
	    clearviewport();
	    settextjustify(1,2);
	    settextstyle(0, 0, 1);
	    setcolor(c_text_bold);
	    if (input_if_af==0)
		sprintf(cc, "%.1lf dBm", dbm_level);	/* RF mode */
	    else
		sprintf(cc, "%.1lf dB", ((double)biggest)/kdb_div);
	    if (mode==FM)		/* If FM, overwrite pwr, rf or af */
		sprintf(cc, "%.2lf dBm", fm_log_pwr);
            moveto(64, 0);
	    outtext(cc);

	    sprintf(cc, "%d Hz", (int)(bin_width[fft_bw]*(double)biggest_freq));
            moveto(64, 10);
	    outtext(cc);

	    /* After noncohave times, ready to update all spectral displays
	     * Unless diagnostics is blank, print the relative average power
	     * over freq range (281, 2400 Hz).
	     */
	    setviewport(579, 470, 638, 479, DO_CLIP);
	    clearviewport();
	    if (diag_mode>0)
		{
		/* 3*power_exp+30 corrects for scaling in dsp for 8 bit data */
		if (bits_8_16==0)		/* If using 8 bit log data */
		    lpave = p2db(pave_print)-23.692+3.0*(double)power_exp-30.0;
		else
		    lpave = p2db(pave_print)-23.692;
		settextjustify(RIGHT_TEXT, TOP_TEXT);
		settextstyle(0, 0, 1);
		setcolor(c_text_bold);
		sprintf(cc, "%.2lfdB", lpave);
		moveto(58, 0);
		outtext(cc);
		pave_print=0.0;		/* Accumulate over noncohave */
		}


	    if ((mode!=EME2 && do_data_save) || (mode==EME2 && eme_record))
	        {
	        if (NULL == ( data_file=fopen("C:\UHFA_1.DAT", "a+t") ))
		  do_t_error(47, "E47 DAT File", 200);
		else
		  {
		  setviewport(10,80,101,90,DO_CLIP);
		  clearviewport();
		  setcolor(c_text_note);
		  settextstyle(0, 0, 1);
		  settextjustify(1,2);
		  moveto(50,1);  outtext("RECORDING");
		  setcolor(c_text_bold);
		  if (mode==EME2)
		    {
		    if((eme_state > -1) && eme_record)
		      {
		      fprintf(data_file,
			"HA %d %d %d %02d%02d %d\n",
			tmnow->tm_year+1900, tmnow->tm_mon+1, tmnow->tm_mday,
			tmnow->tm_hour, tmnow->tm_min, tmnow->tm_sec);
	              fprintf(data_file, "HE %d %d %d 0 0 %E %E ",
		        rfgain, do_window, fft_bw, paveave, pave);
		      for(i=lt_i2; i<=lt_i4; i++)
			 fprintf(data_file, "%E ", pd16[i]);
		      fprintf(data_file, "\n");
		      }
		    }
		else		/* if mode not EME2 */
		    {
		    fprintf(data_file,
		      "HA %d %d %d %02d%02d %d\n",
		      tmnow->tm_year+1900, tmnow->tm_mon+1, tmnow->tm_mday,
		      tmnow->tm_hour, tmnow->tm_min, tmnow->tm_sec);
	            fprintf(data_file,
		      "HB %d %d %d %d %d %d %d %d %d %d\n",
		      fft_bw, rfgain, noncohave, do_window,
		      256*dsp_deltap_hi+dsp_deltap_lo,
		      4*mode+16*bits_8_16+64*input_if_af, 
		      0, 0, 0, 16);		/* Spares */
	    	    for (i=0; i<32; i++)
		      fprintf(data_file,
		        "%E %E %E %E %E %E %E %E\n",
		        pwrdat[8*i],pwrdat[8*i+1],pwrdat[8*i+2],pwrdat[8*i+3],
		        pwrdat[8*i+4],pwrdat[8*i+5],pwrdat[8*i+6],
		        pwrdat[8*i+7]);
		    }

		  fclose(data_file);
	          }
	        }			/* End data save to disk file */
	    else
		{
		/* Clear 'Recording' Note */
		setviewport(10,80,101,90,DO_CLIP);
		clearviewport();
		}

	    if(mode==LHL7 || mode==PUA43)
		 lhl7_timing();	/* Receive trig */
	    /* Dump non-coherent integrator to start a new line */
	    for  (i=0; i<512; i++)
		pwrdat[i]=0;		/* Dump integrator */
	    }				/* End display update */
	}

void prep_display(void)
	{int i, x, y;

	/* If waterfall, put data at top, else below frequency dial */
	if (text_area)
	    {
	    setviewport(X1L+1, Y1T+33, X1R-1, Y1B-1, DO_CLIP);
	    clearviewport();
	    setcolor(c_outline);
	    for(i=0; i<=6; i++)
		{
		y = i*16;
		line(1, y, 639, y);	/* Horizontal grid lines */
		}
	    for(i=1; i<=9; i++)
		{
		x = (int)(53.229 * (double)i);
		line(x, 1, x, 96);		/* Vertical lines */
		}
	    }
	else
	    {
	    setviewport(X1L+1, Y1T+1, X1R-1, Y1B-1, DO_CLIP);
	    clearviewport();
	    setcolor(c_outline);
	    for(i=0; i<=8; i++)
		{
		y = i*16;
		line(0, y, 639, y);	/* Horizontal grid lines */
		}
	    for(i=1; i<=9; i++)
		{
		x = (int)(53.229 * (double)i);
		line(x, 0, x, 128);		/* Vertical lines */
		}
	    }

	/* For information, put the noise bandwidth in the upper right */
	if (waterfall)			/* Don't use for simple display */
	    {
	    sprintf(cc, "%.1f Hz", bin_width[fft_bw]*noise_bw[do_window]);
	    if (text_area)
		setviewport(X1R-62, Y1T+53, X1R-3, Y1T+62, DO_CLIP);
	    else
		setviewport(X1R-62, Y1T+20, X1R-3, Y1T+29, DO_CLIP);
	    clearviewport();
	    settextjustify(RIGHT_TEXT, TOP_TEXT);
	    setcolor(c_text);
	    moveto(58, 1);	outtext(cc);
	    setcolor(c_text_bold);

	    }

	set_disp_offset(0);		/* Update display, no dsp command */
	set_db_div(0);			/* Update display only */

	setcolor(c_text_bold);		/* Default */
	setviewport(X2L+1, Y2T+1, X2R-1, Y2B-1, DO_CLIP);
	if(waterfall && ++scroll_line > 321) /* Scroll Waterfall */
		{
		scroll_line=312;
		getimage(0,  10, 510, 170, big_area_ptr1);
		getimage(0, 171, 510, 321, big_area_ptr2);
		setviewport(X2L+1, Y2B-30, X2R-1, Y2B-10, DO_CLIP);
		clearviewport();
		setviewport(X2L+1, Y2T+1, X2R-1, Y2B-1, DO_CLIP);
		putimage(0,   0, big_area_ptr1, COPY_PUT);
		putimage(0, 161, big_area_ptr2, COPY_PUT);
		}
	}

void normalize(int npoints)
	{
	int i;
	double dnorm_big;

	/* Accumulate noise power to find frequency response */
	if (npoints==512)
	    /* For 512 points of 8 bit data from dsp: */
	    for (i=0; i<npoints; i++)
	    	dnorm[i] += (double)raw_data[i];
	else
	    /* For 256 points of 16 bit data from dsp: */
	    for (i=0; i<npoints; i++)
		dnorm[i] += (double)
		    ( (raw_data[2*i]+(raw_data[2*i+1]<<8)) >>7);

	if (--normalize_on == 1)	/* Used as flag and as a counter */
	    {
	    setviewport(0,90,101,100,DO_CLIP);
	    clearviewport();
	    settextstyle(0, 0, 1);
	    settextjustify(1,2);
	    moveto(50,1);
	    if (normalize_num<100.0) outtext("Fast Norm");
	    else outtext("Prec Norm");

	    dnorm_big=-32768.0;
	    for (i=0; i<npoints; i++)
		{
		dnorm[i]=dnorm[i]/(3.2*normalize_num);
		if (dnorm[i]>dnorm_big)  dnorm_big=dnorm[i];
		}
	    for (i=0; i<npoints; i++)
		{
		dnorm[i] -= dnorm_big;
		}
	    }
	}

/* Curve fit to noise response, y(x) where x is in steps of 1 over isize points
 * and y is dB response. Curve fit is done twice. First with all weights
 * set to 1. Then, a second fit is done where the weights are 1/resid^2
 * where resid is the difference between the measured point and the first
 * curve fit. This prevents a signal from pulling the curve fit out of line.
 * The resulting 4th order polynomial is curve_fit[i], i over (0,3); the
 * fitted curve to be used in normalization is global lt_dnorm[].
 * Revised to (lt_i1, lt_i5) 15 Dec 00 RSL
 */
void lt_normalize(void)
    {
    #define POLY_ORD 4

    double curve_fit[10];
    struct MAT      *At, *B, *At_Ai_At, *Z;
    int             i, j, l;
    int		    isize;		/* Number of points to fit */
    double          *x, *y, *w, *yf;
    double          **at, **b, **z;
    double          yfit, xpow;
    double          resid, rsum, rsqsum, sigma, weight;
    char            pline[100];
    char            *px, *py;

    isize=lt_i5-lt_i1+1;

    /* Allocate memory for arrays */
    w = (double*)calloc(isize, sizeof(double));
    if(!w)
	{
	do_t_error(41, "E41 Mem Alloc", 200);
	return;
	}
    x = (double*)calloc(isize, sizeof(double));
    if(!x)
	{
	do_t_error(41, "E41 Mem Alloc", 200);
	return;
	}
    y = (double*)calloc(isize, sizeof(double));
    if(!y)
	{
	do_t_error(41, "E41 Mem Alloc", 200);
	return;
	}
    yf = (double*)calloc(isize, sizeof(double));
    if(!yf)
	{
	do_t_error(41, "E41 Mem Alloc", 200);
	return;
	}
    for(i=0; i<isize; i++)
       {x[i]=(double)(i+lt_i1);            /* Integer, like 49, 50, ... */
        y[i]=p2db(lt_pwrdat[i+lt_i1]);     /* Measured power, in dB */
        w[i]=1.0;
       }

    /* Fourth order curve fit:    */
    if(fit_curve(curve_fit, x, y, w, isize, POLY_ORD) == 0)
	do_t_error(11, "E11 Curve Fit", 200);

    /* Calculate the least squares fit y values and the squared error */
    rsum=0.0;
    rsqsum=0.0;
    for(i=0;  i<isize;  i++)
        {
        yfit = curve_fit[0];
        xpow = x[i];
        for(j=1;  j<=POLY_ORD;  j++)
            {
            yfit += curve_fit[j]*xpow;
            xpow = xpow * x[i];
            }
        yf[i]=yfit;
        resid=y[i]-yfit;
        rsum += resid;
        rsqsum += resid*resid;
        }
    sigma=sqrt(rsqsum/(double)isize);
    weight=1.0/(sigma*sigma);		/* Use in loop */
    for(i=0;  i<isize;  i++)
        {
        resid=y[i]-yf[i];
        if (fabs(resid) > sigma)    /* Cut the weight on potential sigs */
            w[i]=1.0/(resid*resid);
        else
            w[i]=weight;
        }

    /* Compute again with weighted y[i]     */
    if(fit_curve(curve_fit, x, y, w, isize, POLY_ORD) == 0)
	do_t_error(11, "E11 Curve Fit", 200);

    /* Calculate the least squares y values and fill in normalization values */
    rsqsum=0.0;
    for(i=0;  i<isize;  i++)
        {
        yfit = curve_fit[0];
        xpow = x[i];
        for(j=1;  j<=POLY_ORD;  j++)
            {
            yfit += curve_fit[j]*xpow;
            xpow = xpow * x[i];
            }
	lt_dnorm[i+lt_i1]=yfit;
        }
    if (mat_err)  do_t_error(11, "E11 Curve Fit", 200);
    free(w);
    free(x);
    free(y);
    free(yf);
    }

/* pe(int ci)  -  A debug function to dump the current program execution
 * location onto the screen. Also dumps some variables if needed.
 * Generally call with pe('A'); etc to print a character 'A'. Set
 * dt1,dt2...dt5 to variables in memory to dump them when pe(1000);
 * is called.
 */
void pe(int ci)
    {
    static int ii=0;
    static int jj=0;

    if (ci==1000)
      {
      setviewport(140, 250, 630, 262, DO_CLIP);
      clearviewport();
      settextstyle(0, 0, 1);
      settextjustify(LEFT_TEXT, TOP_TEXT);
      moveto(1,1);
      sprintf(cc, "%9.3g %9.3g %9.3g %9.3g %9.3g", dt1, dt2, dt3, dt4, dt5);
      outtext(cc);
      return;
      }

    setviewport(140+10*ii, 350+jj, 150+10*ii, 362+jj, DO_CLIP);
    clearviewport();
    settextstyle(0, 0, 1);
    settextjustify(LEFT_TEXT, TOP_TEXT);
    moveto(1,1);
    sprintf(cc, "%c", (char)ci);
    outtext(cc);
    if (++ii > 45) 
	{ii=0;
	jj += 7;
	if (++jj > 100) jj=0;
	}
    }

/* do_transmit()  -  Here because the transmit/receive toggle was put into
 * the transmit position.  We will stay here until receive is again commanded.
 */
void do_transmit(void)
	{
	int i, alternate;
	int stop_now;
	int new_char, last_char;
	int esc_count;
	int check_sum_dsp;		/* Received from dsp */
	int check_sum_pc;		/* Calculated by pc frm data*/ 
	int nt;				/* Tone repeats for FSK tones */
	int code_mode;
	long int li;

	code_mode=(mode==CW || mode==LHL7 || mode==PUA43 || mode==LTI);
	stop_now=FALSE;
	alternate=0;
	esc_count=0;			/* Start fresh R1.7 */

	if ((mode==LHL7 || mode==PUA43) && clock_been_set==FALSE)
	    do_t_error(12, "W12 Clock Set", 0);  /* Note, but continue */

 ppt:	/* First clear excess status input */
	while (ASYNCINSTAT() >60 )
		ASYNCIN();

	for(;;)				/* Wait for ESC_CHAR */
		{if ( ASYNCINSTAT() > 0 )
			{new_char = ASYNCIN();
			 if (new_char==ESC_CHAR)  esc_count++;
			 break;
			}
		 while ( kbhit() )
		    {
		    stop_now=do_cmd();
		    if (stop_now)  quit();

		    /* Check for PTT off via fast serial RI line */
		    if ((0X4000 & ASYNCSTAT()) && hardware_ptt && ptt)
			{
			ptt=FALSE;
			if (transmit) t_r_toggle();	/* Go to receive */
			return;
			}

		    if (transmit==0) 	/* Back to Receive */
			{
			if (mode==LHL7)
			    set_code_nx2();
			return;
			}
		    }
	    	 if (code_mode)
		      morse(SEND_CW, 0); /* Send ch, if available */
		}
	if(alternate++ <= 5)   goto ppt;
	alternate=0;
	for(;;)				/* Wait for 'T' */
		{last_char =new_char;
		 if(ASYNCINSTAT() > 0)
			{new_char = ASYNCIN();
			 if (new_char==ESC_CHAR) esc_count++;	
			 if (esc_count==2)	/* Waiting for ESC_CHAR, 'T' */
				{esc_count=0;	/*    but got 2xESC_CHAR     */
				 goto ppt;	/* So try again              */
				}
			 if ((last_char==ESC_CHAR) && (new_char=='T'))
				break;
			}
		 while ( kbhit() )
		    {
		    stop_now=do_cmd();
		    if (stop_now)  quit();
		    if (transmit==0)	/* Back to Receive */
			{
			if (mode==LHL7)
			    set_code_nx2();
			return;
			}
		    }
	    	 if (code_mode)
		      morse(SEND_CW, 0); /* Send ch, if available */



		/* Check for PTT off via fast serial RI line */
		if ((0X4000 & ASYNCSTAT()) && hardware_ptt && ptt)
		    {
		    ptt=FALSE;
		    if (transmit) t_r_toggle();	/* Go to receive */
		    return;
		    }



		}
	for(;;)				/* Wait for a buf full */
		{
		if ( ASYNCINSTAT() > 29 )	/* Almost full */
			 break;
		while ( kbhit() )
		    {
		    stop_now=do_cmd();
		    if (stop_now)  quit();
		    if (transmit==0)	/* Back to Receive */
			{
			if (mode==LHL7)
			    set_code_nx2();
			return;
			}
		    }
	    	if (code_mode)
		      morse(SEND_CW, 0); /* Send ch, if available */




		/* Check for PTT off via fast serial RI line */
		if ((0X4000 & ASYNCSTAT()) && hardware_ptt && ptt)
		    {
		    ptt=FALSE;
		    if (transmit) t_r_toggle();	/* Go to receive */
		    return;
		    }




		}
	incount=-1;
	i=0;  check_sum_pc=0;  esc_count=0;
	for(;;)				/* Get xmit_status */
		{if( ASYNCINSTAT() > 0 )
			{new_char = ASYNCIN();
			 ++incount;
			 if (new_char==ESC_CHAR) esc_count++;
			 if (esc_count==2)
				{esc_count=0;
				 goto t_skip_char;
				}
			 dsp_xmit_status[i]=new_char;
			 check_sum_pc += new_char & 0XFF;
			 if (++i > 8)
				{if (esc_count==1) ASYNCIN();
				 break;
				}
 t_skip_char:		}



		/* Check for PTT off via fast serial RI line */
		if ((0X4000 & ASYNCSTAT()) && hardware_ptt && ptt)
		    {
		    ptt=FALSE;
		    if (transmit) t_r_toggle();	/* Go to receive */
		    return;
		    }





		}
	for(;;)				/* Get check sum */
		{if ( ASYNCINSTAT() > 0 )
			{check_sum_dsp=ASYNCIN();
			 if (check_sum_dsp==ESC_CHAR) ASYNCIN();
			 break;
			}


		/* Check for PTT off via fast serial RI line */
		if ((0X4000 & ASYNCSTAT()) && hardware_ptt && ptt)
		    {
		    ptt=FALSE;
		    if (transmit) t_r_toggle();	/* Go to receive */
		    return;
		    }


		}
	check_sum_pc &= 0XFF;
	if (check_sum_pc==check_sum_dsp)  
		{
		}
	else /*   do_t_error( ) <ADD SOMETHING<<<<<<<<<<<<<<<  */ ;
	
	i=0; esc_count=0;
	for(;;)				/* Wait for dsp_status[] */
		{if( ASYNCINSTAT() > 0 )
			{new_char = ASYNCIN();
			 if (new_char==ESC_CHAR) esc_count++;
			 if (esc_count==2)
				{esc_count=0;
				 goto t_dstat1;
				}
			 dsp_status[i] = new_char;
			 if (++i > 17) 
				{if (esc_count==1) ASYNCIN();
				 break;
				}
 t_dstat1:		}



		/* Check for PTT off via fast serial RI line */
		if ((0X4000 & ASYNCSTAT()) && hardware_ptt && ptt)
		    {
		    ptt=FALSE;
		    if (transmit) t_r_toggle();	/* Go to receive */
		    return;
		    }




		}
	if(dsp_status[17]==2)		/* Valid status data? */
	    {
	    if  ( ((dsp_status[16] & 0X10)==0) && ptt)  /* PTT to receive */
		{
		ptt=0;
		t_r_toggle();		/* Back to receive */
		if (mode==LHL7)
		    set_code_nx2();
		return;
		}
	    diagnostics(1);		/* Update diagnostics */
	    }
	else
	    	do_t_error(12, "E12 DSP XStatus", 20);
	print_time();
	goto ppt;	
	}

/* eme_timing()  can be called anytime.
 * For transmit initiation, checks to see if millisec has just
 * passed a multiple of 5000. If so, transmit is started
 * The eme_state is set to 0 indicating transmit. If 10 minutes
 * has passed, an ID is transmitted.
 * Ends transmit pulse after 2 seconds.
 * For receive, waits until 2.56 sec has elapsed and triggers a fft
 * measurement.
 *
 * eme_state: 0  Transmit pulse started, receive continuing of last pulse
 *            1  Transmit pulse on, receive and screen update completed
 *            3  After 2.56 sec, Receive FFT started, waiting for transmit
 *            4  Special case: Initial wait for transmit pulse
 *           -1  Special case: EME2 selected, but in idle, no transmitting
 */
void eme_timing(void)
	{
	int i;
	unsigned char c, *idptr;	/* For sending CW ID */

	if (eme_state == -1) return;

	i=(int)update_clock(1);		/* integer seconds */

	if ( (i%540)==0 )
	    {
	    if (!transmit) t_r_toggle(); /* To transmit */
	    idptr=&code_id[0];
	    while ((c=*idptr++) != 0)
		morse_ch(c,23);		/* Send ID at 25 wpm */
	    if (transmit) t_r_toggle(); /* To receive */
	    while( ((int)update_clock(1))%5 ) ;	/* Wait out 5 sec interval */
	    return;
	    }
	i=i%5;				/* Remainder after div by 5 */

	/* The following may get retriggered 1 second into the 5 sec
	 * sequence. This happens if a screen update finished before the
	 * end of the 5th sec. This is OK as it will still be caught
	 * at the end of the 2nd second to turn the pulse off. The only
	 * problem is when processing is too slow to finish before
	 * the end of the 2nd second.
	 */
	if (i==0 && eme_state>0)	/* Every 5 seconds */
	    {
	    /* Put dsp into transmit, offset by doppler */
	    eme_state=0;		/* Mark the activity */
	    if (!transmit) t_r_toggle();  /* To transmit */
	    dsp_cmd('T', 0X0011, 2, fgainl[xmit_power], fgainh[xmit_power]);
	    }
	/* Normally, here with eme_state==1, waiting for i=2.
	   If eme_state is 0, the update failed, but we need to move on.
	 */
	if ( (eme_state==0 || eme_state==1) && i==2 )
	    {
	    /* Put dsp into receive */
	    eme_state=2;		/* Mark the activity */
	    dsp_cmd('T', 0X0001, 2, fgainl[xmit_power], fgainh[xmit_power]);

	    if (transmit) t_r_toggle();	/* Into receive */
	    /* Transmit pulse is finished - at 2.56 sec, trigger receive */
	    /* The delay is shortened by T/R sequencing delays */
	    delay(dly_emet2r);		/* Sit here during eme_state=2 (400) */

	    fft_fcn |= 0X0400;		/* Trigger FFT for data collection */
	    dsp_cmd('F', fft_fcn, fft_fcn>>8, 3, 4);
	    fft_fcn &= 0XFBFF;		/* So next guy doesn't trigger */

	    beep(5);			/* Little click */
	    integrate_count=0;		/* Keep display timing in sync */
	    eme_state=3;		/* Receive/process data & next xmit */
	    diagnostics(1);		/* Update screen */
	    }
	}

/* lhl7_timing()  can be called anytime, during receive.
 * Checks to see if sec has just passed a multiple of 2.
 * If so, another receive fft sequence is started. Note that DSP only needs
 * a start trigger and then it finishes up and waits for trigger by itself.
 *
 * lhl7_state keeps track of where we are.  This requires a call to this
 * routine at least once per sec. Also used in exactly the same ay for PUA43.
 *   lhl7_state=0   Waiting for start of trigger sequence
 *   lhl7_state=1   Triggered, but during first second
 */
void lhl7_timing(void)
	{
	int i, ri;
	static long int  t_last=-1L;
	long int       t_new;
	struct tm      *tp;

	if (transmit) return;

	t_new=update_clock(1);
	if (t_new == t_last)   return;

	t_last=t_new;			/* For next time */
        t_new += dsp_clock_base;	/* Reuse t_new */
	tp=gmtime(&t_new);		/* Get struct that includes .tm_sec */
	ri = tp->tm_sec % 2;		/* Remainder after div by 2 */

	if (lhl7_state==0 && ri==0)	/* Every 2 seconds */
	    {
	    fft_fcn |= 0X0400;		/* Software trigger DSP */
	    /* 4*tp->tm_sec is DSP time tag */
	    dsp_cmd('F', fft_fcn, fft_fcn>>8, 4*tp->tm_sec, 4);
	    fft_fcn &= 0XFBFF;

	    setviewport(90, 470, 127, 479, DO_CLIP);
	    clearviewport();
 if(!eme_doppler)				/*<<<<<< */
	    {
	    settextstyle(0, 0, 1);
	    settextjustify(RIGHT_TEXT, TOP_TEXT);
	    moveto(36, 0);
	    sprintf(cc, "R:%d", tp->tm_sec);
	    outtext(cc);
	    }

	    if (tp->tm_sec==0)		/* Start of a new minute */
		{
		code_min=tp->tm_min;	/* Used to randomize pua43 code */
		code_min += (60*tp->tm_hour);
		}
            if (lpt_port==0)
                beep(tick_time);
            else
               {outp(lpt_port, 1);
                delayu(ntau);
                outp(lpt_port, 0);
               }

	    lhl7_state=1;		/* This prevents a retrigger */
	    }
	else if (ri==1)
	    {
	    lhl7_state=0;
	    }
	}
	
void print_stars(void)
	{
	 unsigned int cnt;
	 for(cnt = 0; cnt < 79; cnt ++)
		  printf("*");
	 }

/* t_r_toggle()   - Responds to hitting t/r toggling key.  Adds sequencing
 * for external antenna relays and power amps. Bits are controlled by
 * uint shift_reg with  0=CW on, 1=trans, 2=recv, 3=PA enable,
 * 4=Ant relay transmit.
 */
void t_r_toggle(void)
	{
	static unsigned int save_rfgain;
	char  ct[20];
	
	cw_on = 0;			/* Start with cw key up */
	transmit ^= 1U;			/* Toggle transmit/receive */

	if (transmit)
		{
		save_rfgain=rfg[rfgain].gainbyte;
	 	rfg[rfgain].gainbyte=0;	/* Disconnect receiver from antenna */
		setcolor(c_text_note);
		sprintf(ct, "%s", "TRANSMIT");
/* Bob-There are 3 new_freq()'s - Is one at beginning adequate?? */
		new_freq();		/* T & R freq may be different */
		if (hardware)
		    {  
		    dsp_cmd(10, 0, 0, 0, 4);	/* Set audio gain to 0 */
		    shift_reg = 0X0010;    /*(shift_reg & 0X00E0) | 0X0010;*/
		    send_shift_data(R10MHZ); /* Ant ry to transmit, recv off */
		    beep(10); delay(dly_ant2amp);
		    shift_reg=0X0018;    /* |= 0X0008;*/
		    send_shift_data(R10MHZ);	/* Enable power amp */
		    delay(dly_amp2xmit);
		    shift_reg=0X001A;    /*  |= 0X0002; */
		    send_shift_data(R10MHZ);	/* Hardware into transmit */
		    delay(10);
		    }
		dsp_cmd('T', 0X0003, 2,         /* Xmit, t status, key up */ 
		    fgainl[xmit_power], fgainh[xmit_power] );
		}
	else
		{
		/* Get back out of transmit to allow receiving command */
		dsp_cmd('T', 0X0002, 2,        /*Rcv, t status, key up  */ 
		    fgainl[xmit_power], fgainh[xmit_power] );
		setcolor(c_text_bold);
		if(input_if_af)		/* If audio mode */
		    sprintf(ct, "%s", "AUDIO");
		else
		    sprintf(ct, "%s", "RECEIVE");
		new_freq();		/* T & R freq may be different */
		if (hardware)
		    {
		    shift_reg = 0X0018;
		    send_shift_data(R10MHZ);	/* Hardware xmit off */

		    delay(dly_xmit2amp);
		    shift_reg = 0X0010;
		    send_shift_data(R10MHZ);	/* Amp off */

		    delay(dly_amp2ant);
		    shift_reg = 0X0000;
		    send_shift_data(R10MHZ);	/* Ant relay to receive */

		    delay(2);
	 	    rfg[rfgain].gainbyte=save_rfgain;  /* Restore rcvr gain */
		    shift_reg = 0X0004;
 /* Bob-are these needed or can new_freq() do the last  of these?? */
		    send_shift_data(R10MHZ);	/* Hardware into receive */
		    delay(20);
		    set_afgain(0);		/* Pop over ? */
		    }
		error_on_tr=2;		/* Prevent error after T/R change */
		}
	setviewport(0,60,101,70,DO_CLIP);
	clearviewport();
	settextstyle(0, 0, 1);
	settextjustify(1,2);
	moveto(50,1);	outtext(ct);
	setcolor(c_text_bold);

	new_freq();			/* T & R freq may be different */
	}

/* rit_toggle()  -  Turns RIT on and off.  When turned back on the old
 * offset is restored.  Not restored after shutdown.
 */
void rit_toggle(void)
	{
	rit_on ^= 1;			/* Toggle rit on/off */
	if (rit_on)
		{
		rit = rit_save;
		}
	else
		{
		rit_save = rit;
		rit = 0;
		}
	new_freq();
	}

/* new_band(int nband)  -  Changes parameters to nband.
 */
void new_band(int nband)
	{
	if (input_if_af)		/* Non-operative in audio mode */
	    {
	    do_t_error(13, "W13 Illeg Audio", 0);
	    return;
	    }
	if (tvtr[nband].lo_type==999)
	    {
	    do_t_error(14, "W14 No Tvtr", 0);
	    return;			/* No transverter for this band */
	    }
	setviewport(0, 3, 102, 16, DO_CLIP);
	clearviewport();
	settextjustify(1,2);
	setcolor(c_text_note);
	moveto(51, 3);
	outtext(tvtr[nband].name);

	tvtr[band].last_tfreq=tfreqL;	/* Save last freq and mode */
	tvtr[band].last_mode=mode;
	tfreqL=(long)tvtr[nband].last_tfreq;  /* Load new freq and mode */
	mode=tvtr[nband].last_mode;
	band=nband;
	new_freq();
	set_mode(0);
	if (diag_mode==3) set_cur_loc(-1);	/* Update doppler EME data */
	}

/* The first conversion synthesizer freq is set by the long int in the
 * call to send_shift_data. The dsp lo is set by global uint's dsp_deltap_lo
 * and dsp_deltap_hi.  These are sent to the dsp by the '7' command.
 * For transmit the above frequencies are calculated here, based on
 *      global transmit		   1=transmit, 0=receive
 *      global tfreqL		   Transmit carrier freq in Hz at 2 meters
 *					(except FM, where it is rcvr freq
 *					 with .control for transmit offset)
 *	global tvtr[band].lo_freq  Offset of displayed freq due to transverter
 *	global tvtr[band].sb	   -1 if i-f=rf-lo, else +1
 *	global tvtr[band].control  FM offset (FM Only)
 *	global mode		
 * For receive, all the above are used, plus
 *	global rit_on		1 if RIT is on, else 0
 *	global rit		Offset in Hz, +RIT raises the receiver lo
 *	global cwoffset		Receive is usb and the displayed frequency
 *				  is higher by cwoffset than the zero-beat.
 * The displayed receive frequency includes RIT which is also displayed
 * separately. When the receive frequency is not obvious, such as with Doppler
 * correction or random stir, it is displayed below the transmit freq.
 *
 * DSP lo freq ranges from 12.501 to 17.500 kHz as main synthesizer
 * goes in 5 KHz steps.
 */
void new_freq(void)
	{
	int i, x;
	unsigned int cmd2;
	unsigned int dsp_dph;
	double synth_freq, dt_freq;
	double et_freq;			/* Receive frequency for display */
	long int    disp_MHz;		/* MHz part of display */
	long int    disp_Hz;
	long int    disp_f;
	long int    aL, bL, xfreqL;

	double      freq_eme2_base;

	static long int   int_freqL=0L;
	static long int   synth_freqL=0L;
	static long int   n_synthL=0L;

	static int  last_time_transmit=FALSE;
	int         lock_synth;		/* TRUE=Don't change synth */
	static long int  last_n_synthL = 0;

	/* USB/LSB direction, mode by mode
	 *                          CW  USB   LSB  FM  LHL7 PUA43 EME2 LTI  */
/*later	static double mode_dir[8]={1.0, 1.0, -1.0, 0.0, 1.0, 1.0, 1.0, 1.0}; */

	/* These are variables for time keeping for EME doppler */
	time_t t0;
	double eme_dop;
	/* And time keeping for LTI randomization  */
	long int       t_new;
	struct tm      *tp;

	/* If in transmit, don't change the synthesizer.
	 * But, retuning in these modes (while
	 * transmitting) can move one out of passband--so limit excursion */
	if (!transmit)			/* In receive */
	    {
	    lock_synth=FALSE;
	    last_time_transmit=FALSE;
	    }
	else 				/* In transmit */
	    {
	    lock_synth=last_time_transmit;	/* Lock if return to here */
	    last_time_transmit=TRUE;
	    }
	
	if (rit_on)
		setcolor(c_text_bold);
	else
		setcolor(c_text);
	setviewport(0,303,126,312,DO_CLIP);
	clearviewport();
	settextstyle(0, 0, 1);
	settextjustify(2,2);
	moveto(126,0);
	sprintf(cc, "%d", rit);
	outtext(cc);
	settextjustify(0,2);
	moveto(2,0);
	outtext("RIT");
	setcolor(c_text_bold);

	if (!waterfall)
	    {
	    setviewport(X2R-47, Y2T+20, X2R-1, Y2T+40, DO_CLIP);
	    clearviewport();
	    settextjustify(CENTER_TEXT, TOP_TEXT);
	    moveto(23, 1);
	    outtext("RIT");
	    moveto(23, 11);
	    if (rit_on)
		outtext(cc);
	    else
		outtext("Off");
	    }

	/* Calculate 'xfreqL', the carrier frequency needed at 2 meters
	 * Note that the transmit frequency, kept in global tfreqL is the base
	 * from which all frequencies are derived. It should NEVER be
	 * modified here. Thus the need for xfreqL.
	 */
	xfreqL = tfreqL;			/* Base for all freqs */
	if (!transmit && rit_on)
	    xfreqL += (long int)rit;	/* For receive only */
	if (transmit && mode==CW)
	    xfreqL += (long int)cwoffset;
	if (transmit && (mode==LHL7 || mode==PUA43))
	    xfreqL += (long int)fsk;
	if (transmit && mode==FM)	/* FM Offset */
	    xfreqL += 1000L*(long int)tvtr[band].control;

	/* In EME2 the randomizing must occur once per 5 sec period. It must
	 * be the same value for T & R. random_freqL is global and survives
	 * this function call.
	 */
	if (transmit && mode==EME2 && eme_state==0)
	    {
	    if (eme2_format & 2)
		random_freqL=(long int)
		    (eme_freq_range*0.000030518*(double)(rand()-16384));
	    else
		random_freqL=0L;
	    }

	/* Base freq, 323, 600 or 900 Hz */
	freq_eme2_base=df[fft_bw]*(double)lt_i3;

	if (mode==EME2)
	    {
	    fsk=(int)random_freqL+(int)freq_eme2_base;
	    set_fsk(2);
	    if(transmit)
		xfreqL += random_freqL+(long int)freq_eme2_base;
	    else
		xfreqL += random_freqL;
	    }

	/* Several conditions require EME doppler. This, in turn, needs
	 * UTC. For all users this is calculated here.
	 */
	if(!transmit && ((mode==EME2 && (eme2_format&1)) || eme_doppler))
	    {
	    if (ut_available)
		{
		if (ms_available)
		    t0=update_clock(1)+dsp_clock_base;
		else
		    (void)time(&t0);
		/* Convert to TDT for Moon calculations */
		cnvut2tdt(&t0, &tdt);	/* t0 is local, tdt is global struct */
		eme_dop=update_moon(&tdt, 1);	/* 1 for display update  */
		}
	    else
		{
		do_t_error(15, "E15 No UTime", 200);
		eme_dop=0.0;
		}
	    }

/*>>>>>TEMP */
	if(transmit && ((mode==EME2 && (eme2_format&1)) || eme_doppler))
	    {
	    if (ut_available)
		{
		if (ms_available)
		    t0=update_clock(1)+dsp_clock_base;  /* Get clock time */
		else
		    (void)time(&t0);
		/* Convert to TDT for Moon calculations */
		cnvut2tdt(&t0, &tdt);	/* t0 is local, tdt is global struct */
		update_moon(&tdt, 1);	/* 1 for display update  */
		}
	    else
		do_t_error(15, "E15 No UTime", 200);
	    }

	/* To use the EME doppler, (a)must be receive, (b)must be activated
	 * if EME2 mode, or eme_doppler must be true if other modes.
	 */
	if (!transmit && diag_mode==3 && 
	    ((mode==EME2 && (eme2_format & 1)) || eme_doppler))
	        xfreqL += (long int)eme_dop;

	/* LTI mode gets randomized every minute, both T & R. This is
	 * updated here using data from lti_freq().
	 */
	if (mode==LTI)
	    {
	    t_new=update_clock(1) + dsp_clock_base;
	    tp=gmtime(&t_new);		/* Need sec of min */
	    code_min=tp->tm_min;	/* Randomize (like pua43) */
	    code_min += (60*tp->tm_hour); /* lti_freq() needs this */
	    fsk=(int)(lti_freq()+0.5);  /* Set code freq */
	    if (transmit)
		{
		xfreqL += (long int)fsk;  /* Includes lt_i3 cntr and random */
		set_fsk(2);		/* Display fsk */
		}
	    else
		{
		xfreqL += (long int)lti_frandom; /* random only */
		set_fsk(0);		/* Clear, no display */
		}
	    }

	/* Receive frequency for case of offsets such as Doppler or random */
	setviewport(9, 162, 119, 171, DO_CLIP);
	clearviewport();
	settextjustify(CENTER_TEXT,TOP_TEXT);
	settextstyle(0, 0, 1);
	setcolor(c_text_bold);
	moveto(55,1);

	/* et_freq is receive freq for display only. Shown during receive,
	 * below xmit freq, if there is a modification from transmit freq */
	if (!transmit &&
		(mode==EME2 || (rit_on && rit!=0) ||
		eme_doppler || (mode==LTI && lti_freq_range>0.4)))
	    {
	    /* et_freq is for display purposes only */
	    et_freq=tvtr[band].lo_freq-
		((double)tvtr[band].sb)*(double)xfreqL;
	    sprintf(cc, "%.6lf", 0.000001*et_freq);
	    outtext(cc);
	    }

	/* Display an R if rcvr freq is randomized */
	show_er();

	/* The insides of the frequency display box for waterfall:    */
	setviewport(2,124,125,156,DO_CLIP);
	clearviewport();
	settextjustify(1,2);
	settextstyle(0, 0, 1);
	setcolor(c_text_bold);
	moveto(62,2);

	if (input_if_af==0)		/* For rf processing */
	    {
	    /* Correct for reference frequency error, first:  */
	    xfreqL=(long int)(0.5+ref_offset*(double)xfreqL);
	    if (!lock_synth)	/* Don't change synth if still in transmit */
		{
		/* int_freqL is desired carrier freq at 2-meters, quantized to
		 * the next lower 5 kHz.  The differences for USB and LSB
		 * are due to birdie avoidance at 9600 and 19200 Hz. 
		 */
		if (mode==FM)
		    int_freqL=5000L*((xfreqL+2500L)/5000L);
		else if (mode==LSB)
		    int_freqL=5000L*((xfreqL+250L)/5000L);
		else
		    int_freqL=5000L*((xfreqL+4000L)/5000L);

		synth_freqL=int_freqL-19665000L;
		n_synthL=synth_freqL/5000L;
		}
	    dsp_dph=(unsigned int)
		(1.365333333*(double)(int_freqL-xfreqL+15000UL));
	    dsp_deltap_lo=0XFF & dsp_dph;
	    dsp_deltap_hi=0XFF & (dsp_dph >> 8);
	    /* In 7 cmd, byte 3 is strobe on shift regs. Set here to 0 */
	    ASYNCOUT(ESC_CHAR);      ASYNCOUT(7);   /* This sets dsp lo */
	    ASYNCOUT(0);             ASYNCOUT(1);
	    ASYNCOUT(dsp_deltap_lo);  ASYNCOUT(dsp_deltap_hi);

	    if(dsp_dph<13653  || dsp_dph>27307)
		do_t_error(16, "E16 3rd LO", 20);

	    if(current_info_box == &lb_freq)
		freq_data(2, synth_freqL, n_synthL, xfreqL,
		    xfreqL-synth_freqL, int_freqL-xfreqL+15000UL, dsp_dph);
	    setviewport(2,124,125,156,NO_CLIP);  /* The small frequency box */
	    clearviewport();
	    settextjustify(CENTER_TEXT, TOP_TEXT);
	    settextstyle(0, 0, 1);
	    setcolor(c_text_bold);
	    moveto(62,2);

	    if (n_synthL != last_n_synthL)	/* If need new synth freq */
		{
		bL=( (unsigned long)( (double)n_synthL / 64.0 ) ) & 0X7FFUL;
		aL=(n_synthL - 64UL * bL) & 0X7FUL;
		send_shift_data( (aL<<1) | (bL<<8) );  /* To dsp to prog PLL */
		last_n_synthL=n_synthL;
		}

	    /* Now find the freq to actually display.  Note that in cw the
	     * displayed freq is that which receives a signal producing
	     * a pitch of cwoffset.
	     * In FM the display is the receive freq in receive, transmit if
	     * in transmit 
	     */
	    dt_freq=tvtr[band].lo_freq-((double)tvtr[band].sb)*(double)tfreqL;
/*>>>>	    if (!transmit) dt_freq += (double)rit;  */
	    if (mode==CW)
	        dt_freq += (double)cwoffset;
	    if (mode==FM && transmit)
		dt_freq += -((double)tvtr[band].sb)*
			1000.0*(double)tvtr[band].control;
	    disp_MHz=(long int)(0.000001*(dt_freq+0.0000005));
	    disp_Hz=(long int)(dt_freq-1.0e6*(double)disp_MHz);

	    if (waterfall)
		{
		sprintf(cc, "%ld", disp_MHz);
		outtext(cc);
		settextstyle(0, 0, 2);
		if (mode==FM)
		    {
		    sprintf(cc, ".%3.3ld", disp_Hz/1000L);
		    moveto(62,15);  outtext(cc);
		    settextstyle(0, 0, 1);
		    }
		else
		    {
		    sprintf(cc, ".%3.3ld", disp_Hz/1000L);
		    moveto(30,15);  outtext(cc);
		    sprintf(cc, "%3.3ld", disp_Hz-1000L*(disp_Hz/1000L));
		    moveto(95,15);  outtext(cc);
		    settextstyle(0, 0, 1);
		    }
		}
	    else			/* Large frequency dial */
		{
		setviewport(0,122,127,158,DO_CLIP);	/* Small Freq Box */
		clearviewport();	/* Remove it */

		/* 128, 639   d=511  */
		setviewport(X2L+53, Y2T+8, X2L+460, Y2T+39, DO_CLIP);
		clearviewport();
		for(i=0; i<5; i++)
		    {
		    x = 43+i*80;
		    line(x, 35, x, 24);
		    }
		for(i=0; i<25; i++)
		    line(11+i*16, 35, 11+i*16, 29);
		/* The frequency cursor: */
		x=3+(int)((xfreqL-143493750)/12500);
		setcolor(c_plot_marker);
		line(x, 34, x, 29);
		setlinestyle(SOLID_LINE, 0, THICK_WIDTH);
		line(x, 28, x, 23);
		/* Frequency Display: */
		disp_f=1000000L*disp_MHz+disp_Hz;

		setlinestyle(SOLID_LINE, 0, NORM_WIDTH);
		setcolor(c_text_bright);
		settextjustify(CENTER_TEXT, TOP_TEXT);
		settextstyle(DEFAULT_FONT, HORIZ_DIR, 3);
		if (mode==FM)
		    {
		    if (disp_MHz < 1000)
			moveto(170, 0);
		    else if (disp_MHz>=1000 && disp_MHz<10000)
			moveto(158, 0);
		    else if (disp_MHz>=10000 && disp_MHz<100000)
			moveto(146, 0);
		    else 			/* disp_MHz>=100000 */
			moveto(134, 0);
		    sprintf(cc, "%ld.", disp_MHz);
		    outtext(cc);
		    moveto(252, 0);
		    disp_f=disp_Hz/1000;
		    sprintf(cc, "%3.3ld", disp_f);
		    outtext(cc);

		    /* Now print FM Offset */
		    setviewport(X2L+1, Y2T+30, X2L+49, Y2T+40, DO_CLIP);
		    clearviewport();
		    settextjustify(CENTER_TEXT, TOP_TEXT);
		    settextstyle(0, 0, 1);
		    if (tvtr[band].control)  setcolor(c_text_bold);
		    else  setcolor(c_text);
		    moveto(24, 1);
		    sprintf(cc, "%d", tvtr[band].control);
		    outtext(cc);
		    setcolor(c_text_bold);
		    }			/* End mode==FM */
		else
		    {
		    if (disp_MHz < 1000)
			moveto(143, 0);
		    else if (disp_MHz>=1000 && disp_MHz<10000)
			moveto(131, 0);
		    else if (disp_MHz>=10000 && disp_MHz<100000)
			moveto(119, 0);
		    else 			/* disp_MHz>=100000 */
			moveto(107, 0);
		    sprintf(cc, "%ld.", disp_MHz);
		    outtext(cc);
		    moveto(223, 0);
		    disp_f=disp_Hz/1000;
		    sprintf(cc, "%3.3ld", disp_f);
		    outtext(cc);
		    moveto(305, 0);
		    disp_f=disp_Hz-1000*disp_f;
		    sprintf(cc, "%3.3ld", disp_f);
		    outtext(cc);

		    /* Now clear any FM Offset */
		    setviewport(X2L+1, Y2T+30, X2L+49, Y2T+40, DO_CLIP);
		    clearviewport();
		    }
		setcolor(c_text_bold);
		}
	    }
	else
	    {				/* For audio processing */
	    dsp_dph = (unsigned int)( 1.365333333 * (double)tfreqL );
	    if (transmit)
		dsp_dph=0;
	    if (transmit && (mode==CW))
		dsp_dph= (unsigned int)( 1.365333333 * (double)cwoffset );
	    dsp_deltap_lo=0XFF & dsp_dph;
	    dsp_deltap_hi=0XFF & (dsp_dph >> 8);
	    /* In 7 cmd, byte 3 is strobe on shift regs. Set here to 0 */
	    ASYNCOUT(ESC_CHAR);      ASYNCOUT(7);
	    ASYNCOUT(0);             ASYNCOUT(1);
	    ASYNCOUT(dsp_deltap_lo);  ASYNCOUT(dsp_deltap_hi);

	    disp_Hz=tfreqL;
	    outtext("Offset Hz");
	    settextstyle(0,0,2);
	    sprintf(cc, "%ld", disp_Hz);
	    moveto(62,15);  outtext(cc);
	    settextstyle(0, 0, 1);
	    }				/* End audio input */
	}				/* End new_freq() */

/* The af gain is set in the dsp by a multiplicative constant and a shift
 * value that ends up in the se register.  Up to 50, se=0.  Each increase
 * of 1 in se is 6 dB of added gain.
 * There are two separate software gain controls for FM and another for
 * other modes.
 */
void set_afgain(int dbstep)
	{
	int gain, afg, ifg;
	int se;				/* Shift value for dsp se register */

	if (mode==FM)
	    {
	    fm_afgain += dbstep;
	    if (fm_afgain>100)  fm_afgain=100;
	    if (fm_afgain<0)  fm_afgain=0;
	    gain=fm_afgain;		/* Local copy */
	    }
	else
	    {
	    afgain += dbstep;
	    if (afgain>100)  afgain=100;
	    if (afgain<0)  afgain=0;
	    gain=afgain;
	    }
	setviewport(63,393,126,402,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	sprintf(cc, "%d", gain);
	outtext(cc);
	if(gain==0)
	    {
	    se=0;
	    afg=0;
	    }
	else if(gain>50)
	    {
	    se=(gain-45)/6;
	    afg=95+(gain-45)%6;
	    }
	else
	    {
	    se=0;
	    afg=50+gain;
	    }
	dsp_cmd(10, fgainl[afg], fgainh[afg], se, 4);
	}

void set_mike_gain(void)
	{
	setviewport(103,258,127,267,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(RIGHT_TEXT, TOP_TEXT);
	moveto(24, 1);
	sprintf(cc, "%d", mike_gain);
	outtext(cc);
	set_mode(1);			/* Send new mike_gain to dsp */
	}

void set_cwoffset(void)
	{
	int sidetone_ph;

	setviewport(34, 288, 100, 297, DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	setcolor(c_text);
	settextjustify(CENTER_TEXT, TOP_TEXT);
	moveto(33, 1);
	sprintf(cc, "%d %d", sidetone_lev, cwoffset);
	outtext(cc);
	setcolor(c_text_bold);

	sidetone_ph=(unsigned int)(1.365333333*(double)cwoffset);
	dsp_cmd('H', sidetone_ph, sidetone_ph>>8,
	   fgainl[sidetone_lev], fgainh[sidetone_lev]);
	new_freq();
	}

void set_eme_doppler(void)
	{
	setviewport(1, 162, 127, 171, DO_CLIP);
	clearviewport();
	if (mode==FM) return;
	show_er();
	new_freq();
	}

void show_er(void)
	{
	/* Display an E on the right, if eme doppler being corrected */
	setviewport(119, 162, 127, 171, DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(CENTER_TEXT, TOP_TEXT);
	setcolor(c_text_bold);
	if (eme_doppler || mode==EME2)
	    {
	    settextstyle(0,0,1);
	    if (diag_mode!=3)
		setcolor(c_text);
	    moveto(4, 1);
	    outtext("E");
	    setcolor(c_text_bold);	/* Always leave this way */
	    }

	/* Display an R on left, if randomization of freq being used */
	setviewport(1, 162, 9, 171, DO_CLIP);
	clearviewport();
	if ((mode==LTI && lti_freq_range>0.4) ||
		 (mode==EME2 && eme_freq_range>0.4))
	    {
	    moveto(4, 1);
	    outtext("R");
	    }
	}

/* set_fsk(do_what)
 * If do_what==0  Clears the display space above the small freq box.
 * If do_what==1  Displays the current fsk, sets the DSP and does a new_freq().
 * If do_what==2  Displays the current fsk & sets the DSP only.
 */
void set_fsk(int do_what)
	{
	int sidetone_ph;

	setviewport(1, 110, 127, 120, DO_CLIP);
	clearviewport();
	if (do_what==0)  return;	/* Just clear the box */
	settextstyle(0,0,1);
	setcolor(c_text);
	settextjustify(CENTER_TEXT, TOP_TEXT);
	moveto(63, 1);
	sprintf(cc, "FSK=%d", fsk);
	outtext(cc);
	setcolor(c_text_bold);
	sidetone_ph=(unsigned int)(1.365333333*(double)fsk);
	dsp_cmd('H', sidetone_ph, sidetone_ph>>8, 
	    fgainl[sidetone_lev], fgainh[sidetone_lev]);
	if (do_what==2)   return;	/* To be called from new_freq() */
	new_freq();
	}

void set_fm_squelch(void)
	{
	int sq;

	if (mode != FM) return;
	setviewport(34, 288, 84, 297, DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(1,2);
	setcolor(c_text);
	moveto(25, 1);
	sprintf(cc, "%d", fm_squelch);
	outtext(cc);
	setcolor(c_text_bold);
	sq=2*(100-fm_squelch)*(100-fm_squelch);
	dsp_cmd('S', sq, sq>>8, 3, 4); 
	}

void set_xmit_power(void)
	{
	setviewport(63,278,126,287,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	sprintf(cc, "%d", xmit_power);
	outtext(cc);
	if (transmit)			/* Update dsp */
	    {ASYNCOUT(ESC_CHAR);  ASYNCOUT('T');
	     ASYNCOUT(0X0003 | (cw_on<<4));  ASYNCOUT(2); 
	     ASYNCOUT( fgainl[xmit_power] ); ASYNCOUT( fgainh[xmit_power] );
	    }
	}

void set_rfgain(void)
	{
	setviewport(63,403,126,412,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	sprintf(cc, "%d", rfg[rfgain].dbdisplay);
	outtext(cc);
	send_shift_data(R10MHZ);
	}

void set_cw_speed(void)
	{
	setviewport(66,268,127,277,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(60, 1);
	outtext(t[cw_speed].description);
	}

void set_contrast(void)
	{
	setviewport(63,233,126,242,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	sprintf(cc, "%d", contrast);
	outtext(cc);
	dis_gain=0.0025+0.0035*(double)(contrast);
	contrast_line_count=20;
	}

void set_brightness(void)
	{
	setviewport(63,243,126,252,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	sprintf(cc, "%d", bright);
	outtext(cc);
	dis_offset=200.0-3.0*(double)bright;
	contrast_line_count=20;
	}

/* Set window has evolved into a general setup of the screen in the spectral
 * display area! It also sets the constants for varying conditions and keeps
 * the variables consistant.
 */
void set_window(void)
	{
	int i;
	if (do_window==0)
	    {
	    sprintf(cc, "%s", "Non");
	    win_adj_loss=0.0;		/* Coherent loss of window fcn, dB */
	    win_offset=0;		/* 3 dB steps in dsp sptrm output */
	    win_adj_residual=0.0;	/* After 3 dB steps */
	    }
	else if (do_window==1)
	    {
	    sprintf(cc, "%s", "Tuk");
	    win_adj_loss=2.3;		/* Book says 1.11 dB ? */
	    win_offset=1;
	    win_adj_residual=-0.5;
	    }
	else if (do_window==2)
	    {
	    sprintf(cc, "%s", "Ham");
	    win_adj_loss=5.35;
	    win_offset=2;
	    win_adj_residual=-0.7;
	    }
	else  /* do_window==3 */
	    {
	    sprintf(cc, "%s", "B-h");
	    win_adj_loss=8.87;
	    win_offset=3;
	    win_adj_residual=-0.2;
	    }
	setviewport(63,378,90,387,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	setcolor(c_text);
	settextjustify(CENTER_TEXT, TOP_TEXT);
	moveto(15, 1);
	outtext(cc);
	setcolor(c_text_bold);
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('W');
	ASYNCOUT(do_window); ASYNCOUT(2);
	ASYNCOUT(3); ASYNCOUT(4);

	/* In EME2 or LTI, a change of fft_bw or do_window needs fresh data */
	if (mode==EME2 || mode==LTI)
	    {
/* USE LTERM_INT_UPDATE() */
	    for(i=0; i<256; i++)  lt_dnorm[i]=0.0;
	    for(i=0; i<21;  i++)  eme_pwr_sum[i]=0.0;
	    eme_pulse_count=0;
	    set_bins();			/* Set lt_i1,... */
	    }
	fft_fcn &= 0XFFF0;		/* Clear last 4 bits */
	if (fft_bw==0)			/* 1200 Hz BW (2400 samp/sec)*/
	    {
	    sprintf(cc, "%s", "1200");
	    fft_fcn |= 0X0001;
	    bin_low=(int)(0.5+freq_low/2.34375);
	    }
	else if (fft_bw==1)		/* 2400 Hz BW (4800 samp/sec)*/
	    {
	    sprintf(cc, "%s", "2400");
	    fft_fcn |= 0X0002;
	    bin_low=(int)(0.5+freq_low/4.6875);
	    }
	else if (fft_bw==2)		/* 4800 Hz BW (9600 samp/sec)*/
	    {
	    sprintf(cc, "%s", "4800");
	    fft_fcn |= 0X0004;
	    bin_low=(int)(0.5+freq_low/9.375);
	    }
	setviewport(91,378,126,387,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	setcolor(c_text_bold);
	settextjustify(RIGHT_TEXT, TOP_TEXT);
	moveto(35, 1);
	outtext(cc);

	set_display_width();

	ASYNCOUT(ESC_CHAR);  ASYNCOUT('F');
	ASYNCOUT(fft_fcn&0XFF);  ASYNCOUT((fft_fcn>>8)&0XFF);
	ASYNCOUT(3);  ASYNCOUT(4);

	if (fft_fcn & 0X0080)  sprintf(cc, "%s", "W");
	else                   sprintf(cc, "%s", "N");
	setviewport(55,313,64,322,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	setcolor(c_text);
	settextjustify(CENTER_TEXT, TOP_TEXT);
	moveto(4, 0);
	outtext(cc);
	setcolor(c_text_bold);

	/* For information, put the noise bandwidth in the upper right */
	if (waterfall)			/* Don't use for simple display */
	    {
	    sprintf(cc, "%.1f Hz", bin_width[fft_bw]*noise_bw[do_window]);
	    if (text_area)
		setviewport(X1R-62, Y1T+53, X1R-3, Y1T+62, DO_CLIP);
	    else
		setviewport(X1R-62, Y1T+20, X1R-3, Y1T+29, DO_CLIP);
	    clearviewport();
	    settextjustify(RIGHT_TEXT, TOP_TEXT);
	    setcolor(c_text);
	    moveto(58, 1);	outtext(cc);
	    setcolor(c_text_bold);
	    }

	set_noncohave(0);		/* Altered by fft_bw ? */
	}

/* set_mode(do_what)  -  Arrive with global int mode=0, 1, 2,...
 * do_what is 1 if save_fine_freqL is valid.
 *            2 increments the mode until an active one is found
 *              relative to mode_mask.  Note that CW (mode=0) is always
 *              available.
 *            0 if came from a new VFO, or cold start.
 */
void set_mode(int do_what)
	{
	int cmd1, cmd2;
	int mode_test;
	int last_mode;
	char cm[20];		/* cc doesn't work here as too many users */

	/* save_fine_freqL is used when going into and out of the fm mode
	   to allow fm to go in 5 kHz steps, but return to the 1 Hz freq */
	static long	save_fine_freqL=0;

	/* We may be about to change the mode. Sometimes we need to know
	 * which mode we came from:         */
	last_mode=mode;

	if (!(mode==EME2 || mode==LTI))  save_8_16=bits_8_16;

	dsp_cmd(10, 0, 0, 0, 4);	/* Set audio gain to 0 */

	set_fsk(0);			/* Clears FSK annotation, if any */

	/* Clear the time in the lower left that is used for code modes */
	setviewport(90, 470, 127, 479, DO_CLIP);
	clearviewport();

	if (do_what==0)
	    save_fine_freqL=tfreqL;	/* Initialize */

	if (do_what==2)			/* Mode change, rotates through them */
	    {
	    while (1)
		{
	        if (++mode>7)  mode=0;	/* Maximum installed mode */
		mode_test=near_int(pow(2.0, (double)mode));
		if (mode_test & mode_mask)  break;	/* An active mode */
		}
	    /* In audio, USB only to avoid inversion of tuning: */
/* Not AF fcn	    if (input_if_af && mode==LSB) mode=USB;   */

	    /* Data modes only when ms clock available:   */
	    if (!ms_available && mode>=4)
		{
		mode=0;
		do_t_error(25, "W25 ms Unavail", 0);
		}
	    if (!waterfall && mode>=4) mode=0;
	    do_what=1;			/* From here on, it is a 1 */
	    }				/* End do_what==2 */

	/* Fall through here for any do_what */
	/* text_area means top 32 pixels of spectrum plot is for text data */
	if ((waterfall && mode==CW) || mode==LHL7 || mode==PUA43)
	    text_area=TRUE;
	else
	    text_area=FALSE;

	/* First clear old data from Mode line  */
	setviewport(34, 288, 126, 297, DO_CLIP);
	clearviewport();

	/* If there are Alt A info boxes running, they get closed when we
	 * change modes:     */
	if (mode != EME2  &&  current_info_box==&lb_eme2) /* If displayed */
	    eme2_update(0);		/* Close */
	if (mode != LHL7  &&  current_info_box==&lb_lhl7)
	    lhl7_update(0, 0);
	if (mode != PUA43  &&  current_info_box==&lb_pua43)
	    pua43_update(0, 0);
	if (mode != LTI  &&  current_info_box==&lb_lterm_int)
	    lterm_int_update(0);

	/* fft_bw must be 2 for LHL7 & FM. Save old BW if new fft_bw = 2 */
	if ((mode==FM || mode==LHL7) && fft_bw != 2)
	    {
	    fft_bw_save=fft_bw;
	    fft_bw=2;
	    }
	if (mode!=LHL7 && mode!=FM && (last_mode==FM || last_mode==LHL7))
	    fft_bw=fft_bw_save;

	/* do_window must be Hamming in FM. Save old  */
	if (mode==FM)
	    {
	    do_window_save=do_window;
	    do_window=2;		/* Hamming window function */
	    }
	if (mode!=FM && last_mode==FM)
	    do_window=do_window_save;

	/* Bit 7 of fft_fcn must be set (I-F bandwidth is wide) in FM.
	 * Save old value  */
	if (mode==FM)
	    {
	    fir_save=fft_fcn & 0X0080;	/* Save the two bits */
	    fft_fcn |= 0X0080;		/* Select wide filter (not interp) */
	    }
	if (mode!=FM && last_mode==FM)
	    fft_fcn |= fir_save;	/* OR in the old value */

	fft_fcn &= 0XE080;		/* Retain I-Q & Interp filts */
	if (mode==LHL7)
	  {
	  integrate_count=0;
	  fft_fcn |= 0X0244;  /* Software trig, 9.6kHz, 1024pt, 3 data/2 sec */
	  }

	/* EME2 and PUA43 allow variable fft bandwidth */
	if (mode==EME2 || mode==PUA43)
	  {
	  integrate_count=0;
	  if(fft_bw==0)       fft_fcn |= 0X0241;  /* Trig, 1024 pt, 1200 Hz */
	  else if(fft_bw==1)  fft_fcn |= 0X0242;  /* Trig, 1024 pt, 2400 Hz */
	  else                fft_fcn |= 0X0244;  /* Trig, 1024 pt, 4800 Hz */
	  }
	else
	  {
	  if(fft_bw==0)       fft_fcn |= 0X0041;  /* 1024 pt, 1200 Hz */
	  else if(fft_bw==1)  fft_fcn |= 0X0042;  /* 1024 pt, 2400 Hz */
	  else                fft_fcn |= 0X0044;  /* 1024 pt, 4800 Hz FFT BW */
	  }

	/* Long term integration forces 16 bit data. Others use what is set */
	if (mode==EME2 || mode==LTI)
	    {
	    bits_8_16 = 1;
	    set_8_16(9);
	    set_bins();			/* Also set up i0,... values */
	    }
	else				/* Go back to last value */
	    {
	    bits_8_16 = save_8_16;
	    set_8_16(9);
	    }

	/* Now program DSP for fft_fcn */
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('F');
	ASYNCOUT(fft_fcn&0XFF);  ASYNCOUT((fft_fcn>>8)&0XFF);
	ASYNCOUT(0);  ASYNCOUT(4);

	set_window();			/* Update display */

	if (mode==FM )  /* Use 5 kHz resolution */
	    {
	    save_fine_freqL=tfreqL;
	    tfreqL = 5000*((tfreqL+2500)/5000);
	    }
	if (last_mode==FM && do_what>0 && save_fine_freqL>0L)
	    {
	    tfreqL=save_fine_freqL;	/* Get old freq to 1 Hz */
	    }
	if (mode==FM)
	    set_agc(-1);		/* Turn off AGC for FM */
	else
	    set_agc(agc_decay);

	if (narrow)
	    {
	    display_width=FULL_FREQ;	/* All except LHL7, EME2, LTI */
	    switch (mode)
		{
		case CW: {
			sprintf(cm, "%s", "   CW");
			cmd1=1;		/* 1=CW */
			cmd2=0;		/* 0=USB (LSB in dsp) */
			/* set_cwoffset(); Done below */
			print_medium(0, -1);	/* Initialize */
			break;
			}
		case USB: {
			sprintf(cm, "%s", "  USB");
			cmd1=0;		/* 0=SSB */
			cmd2=0;		/* 0=USB (LSB in dsp) */
			break;
			}
		case LSB: {
			sprintf(cm, "%s", "  LSB");
			cmd1=0;		/* 0=SSB */
			cmd2=1;		/* 1=LSB (USB in dsp) */
/* cmd2=2; 2=Binaural */
			break;
			}
		case FM: {
			sprintf(cm, "%s", "   FM");
			cmd1=2;		/* 2=FM */
			cmd2=0;		/* 0=no ssb */
			set_fm_squelch();
			break;
			}
		case LHL7: {
			sprintf(cm, "%s", "LHL-7");
			cmd1=1;		/* 1=CW */
			cmd2=0;		/* 0=USB (LSB in dsp) */
			d_nx2=0;
			set_code_nx2();		/* Update screen settings */
			lhl7_update(0, -1);
			fsk=600;  set_fsk(1);
			morse(INIT_CW, 0);
			print_medium_1(0, -1);	/* Initialize */
			print_medium_2(0, -1);	/* Initialize */
			print_big(0, -1, 0);	/* Initialize */
			sync_flag=TRUE; 	/* Time align displays */
			/* HALF_FREQ is not generally available */
			display_width=HALF_FREQ;
			break;
			}
		case PUA43: {
			sprintf(cm, "%s", "PUA43");
			cmd1=0x01;		/* 80=Triggered 2 sec, 1=CW */
			cmd2=0;			/* 0=USB */
			print_big(0, -1, 0);	/* Initialize */
			print_super_row(0, -1, 0);	/* Initialize */
			pua43_update(0, -1);	/* Also initialize */
			fsk=0; set_fsk(1);
			sync_flag=TRUE;
			eme_pulse_count=0;	/* New start for Noise Blank */
			break;
			}
		case EME2: {
			sprintf(cm, "%s", "EME2");
			cmd1=0x01;		/* For now: 1=CW */
			cmd2=0;			/* 0=USB */
			eme2_update(-1);	/* Also initialize */
			set_bins();
			fsk=(int)(0.5+df[fft_bw]*(double)lt_i3);
			set_fsk(0);		/* No display, use Alt-A Box*/
			display_width=HALF_FREQ;
			break;
			}
		case LTI: {
			sprintf(cm, "%s", "  LTI");
			cmd1=1;		/* 1=CW */
			cmd2=0;		/* 0=USB (LSB in dsp) */
			print_medium(0, -1);	/* Initialize */
			fsk=(int)(0.5+lti_freq());
			set_fsk(0);
			display_width=HALF_FREQ;
			lterm_int_update(-1);	/* Also initialize */
			break;
			}

		/* AM=7, ... */
		default: break;
		}
	    }
	else				/* else if wideband modes */
	    {
#if 0
	    /* This needs to be integrated into things.
	     * The most useful are binaural and USB left, LSB right.
	     */
	    if (binaural && iq)
		{
		sprintf(cm, "%s",  "WB IQ");
		cmd1=0;			/* Not used in wideband */
		cmd2=0X96;		/* Wide, I-Q, Binaural */
		}
	    else if (!binaural && iq)
		{
		sprintf(cm, "%s",  "WM  I");
		cmd1=0;			/* Not used in wideband */
		cmd2=0X94;		/* Wide, I-Q, Monaural */
		}
	    else if (binaural && !iq)
		{
		sprintf(cm, "%s",  "B SSB");
		cmd1=0;			/* Not used in wideband */
		cmd2=0X9A;		/* Wide, SSB, Binaural */
		}
	    else if (!binaural && !iq)
		{
		sprintf(cm, "%s",  "M USB");
		cmd1=0;			/* Not used in wideband */
		cmd2=0X98;		/* Wide, SSB, Monaural */
		}
#endif
 	    sprintf(cm, "%s", " ");   /* Temp before reinstalling wb stuff */
	    }				/* End 'if wideband modes' */
	setviewport(84,288,126,297,DO_CLIP); /* Area for mode on left */
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(RIGHT_TEXT, TOP_TEXT);
	moveto(42, 1);
	outtext(cm);			/* Print the current mode */

	new_freq();			/* Get cw offset, etc */

	/* In CW and LTI the sidetone is at the same pitch and level
	 * even though the FSK tone changes in LTI with random freq.
	 */
	if (mode==CW || mode==LTI) set_cwoffset();

	/* The CW data is a little wide! Print now to have room */
	if (!waterfall)
	    {
	    setviewport(X2L+1, Y2T+20, X2L+49, Y2T+30, DO_CLIP);
	    clearviewport();
	    settextjustify(CENTER_TEXT, TOP_TEXT);
	    moveto(24, 1);
	    outtext(cm+2);
	    }
	/* Note: when these commands get to dsp they produce the reverse
	   sideband.  This is because of upper side conversion at 19.68 MHz.
	   Notation, USB, LSB here refers to signal at 2 meters */
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('M');
	ASYNCOUT(cmd1); ASYNCOUT(cmd2);
	ASYNCOUT( fgainl[mike_gain] ); ASYNCOUT( fgainh[mike_gain] );
	delay(50);			/* Stop pop after mode change */
	set_afgain(0);			/* Set but don't change */
	set_noncohave(0);		/* Mode affects these, sometimes */
	set_display_width();
	set_eme_doppler();
	}

/* set_code_nx2()  -  sets/displays the number of 2 second periods being
 * used in codes LHL7. The variable d_nx2 determines the action:
 *   d_nx2==-1  Reduce code_nx2 to the next lower value, if there is one.
 *   d_nx2==0   No change to code_nx2. Redisplay and alter spec_ave if needed.
 *   d_nx2==1   Increase code_nx2 to the next higher value, if there is one.
 * Allowable values of code_nx2 are 1, 2, 4, 10 and 30 with the times being
 * twice these values.
 */
void set_code_nx2(void)
	{
	setviewport(34, 288, 84, 297, DO_CLIP);
	clearviewport();

	if (mode != LHL7) return;

	if (d_nx2 == 0) ;		/* Display, but don't change */
	else if (d_nx2 == -1)		/* Reduce code_nx2 to next lower */
	    {
	    if (code_nx2==1) ;		/* Just redisplay */
	    else if (code_nx2==2) code_nx2=1;
	    else if (code_nx2==5) code_nx2=2;
	    else if (code_nx2==10) code_nx2=5;
	    else code_nx2=10;
	    }
	else if (d_nx2 == 1)
	    {
	    if (code_nx2==30) ;		/* Just redisplay */
	    else if (code_nx2==1) code_nx2=2;
	    else if (code_nx2==2) code_nx2=5;
	    else if (code_nx2==5) code_nx2=10;
	    else code_nx2=30;
	    }
	settextstyle(0,0,1);
	setcolor(c_text);
	settextjustify(CENTER_TEXT, TOP_TEXT);
	moveto(25, 1);
	sprintf(cc, "%d s", 2*code_nx2);
	outtext(cc);
	setcolor(c_text_bold);

	set_noncohave(2);		/* Keep this tracking */
	}

void set_data_save(void)
	{
	do_data_save ^= 1;	
	setviewport(10,80,101,90,DO_CLIP);
	clearviewport();
	setcolor(c_text);
	settextstyle(0, 0, 1);
	settextjustify(1,2);
	moveto(50,1);
	if (do_data_save)
	    outtext("RECORDING");
	setcolor(c_text_bold);
	}

void normalize_plot(int k)
	{
	int i;

	for (i=0; i<512; i++)
		dnorm[i]=0.0;
	setviewport(0,90,101,100,DO_CLIP);
	clearviewport();
	settextstyle(0, 0, 1);
	settextjustify(1,2);
	moveto(50,1);
	if (k==0)			/* Clear normalization */
	    {
	    normalize_on=0;
	    }
	else if (k==1)			/* Add fast normalization */
	    {
	    normalize_on=25;		/* Count down to 1 */
	    normalize_num=25.0;
	    outtext("Wait...");
	    }
	else
	    {
	    normalize_on=1000;		/* Big countdown for precision */
	    normalize_num=1000.0;
	    outtext("Wait...");
	    }
	}

/* do_auto_disp(k)  -  corrects for analog gain variations so that the
 * colors in the spectral display don't drift.  The down side to this
 * is that strong signals tend to push the noise down.  So it works very
 * well for weak signals in noise.
 */
void do_auto_disp(int k)
	{
	int i;

	setviewport(63,423,126,432,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	if (k==0)			/* Stop color agc */
	    {
	    auto_disp_on=FALSE;
	    setcolor(c_text);
	    outtext("OFF");
	    setcolor(c_text_bold);
	    }
	else				/* Turn on auto_disp */
	    {
	    auto_disp_on=-1;		/* A flag to initialize base value */
	    setcolor(c_text_bold);
	    outtext("ON");
	    }
	}

/* set_8_16(k)  -  Programs the DSP for 8 or 16 bit data (either 512 bytes
 * or 256 words of logged spectral data).  Changes the resolution of the
 * data coming to the PC from the DSP.  If k=0, resets bits_8_16=0, the
 * value for 8 bit data.  If k=1 it toggles 8/16.  If k=9 it just programs
 * the DSP to the current value of bits_8_16.
 * See also set_disp_offset() for more info.
 */
void set_8_16(int k)
	{
	if (k==0)  bits_8_16=0;		/* Reset to 8 bit on 0 */
	else if (k==1)  bits_8_16 ^= 1;		/*   or toggle on 1 */
	/* Fall through on 9 */

	ASYNCOUT(ESC_CHAR);  ASYNCOUT(9);
	ASYNCOUT(bits_8_16);
	ASYNCOUT(power_exp);
	ASYNCOUT(LOG_OFFSET&0XFF);  ASYNCOUT((LOG_OFFSET>>8)&0XFF);

	setviewport(70,368,90,377,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(CENTER_TEXT, TOP_TEXT);
	moveto(10, 1);
	if (bits_8_16&2)
	   setcolor(c_text_bold);	/* Indicates FFT data not coming */
	else
	   setcolor(c_text);		/* Normal FFT data */
	if (bits_8_16&1)
	    outtext("16");
	else
	    outtext("8");
	setcolor(c_text_bold);

	set_disp_offset(1);  /* << CHECK - This probably does same dsp cmds*/

	if (mode==EME2)
	    {
	    eme2_update(-1);
	    eme_pulse_count=0;
	    }
	if (mode==LTI)
	    {
	    lterm_int_update(-1);
	    eme_pulse_count=0;
	    }
	}

/* set_input()  -  Arrive with global input_if_af=0 for i-f or 1 for af.
 */
void set_input(void)
	{
	static long	save_af_freqL=0;
	static long	save_tfreqL=146000000;
	static int 	save_mode_if=USB;
	
	if (transmit) t_r_toggle();	/* Either if or af, it's receive */

	setviewport(0,60,101,70,DO_CLIP);
	clearviewport();
	settextstyle(0, 0, 1);
	settextjustify(1,2);
	setcolor(c_text_bold);
	moveto(50,1);
	if (input_if_af==0)
		{
		outtext("RECEIVE");
		save_af_freqL=tfreqL;	/* Going from af to if input */
		tfreqL=save_tfreqL;
		mode=save_mode_if;
		set_mode(0);
		}
	else
		{
		outtext("AUDIO");
		save_tfreqL=tfreqL;	/* Going from if to af input */
		tfreqL=save_af_freqL;
/* Patch now to only allow USB at audio (LSB on screen) and no transmit */
/*		if (mode==USB)
		    {
		    mode=LSB;
		    set_mode(0);
		    }
 */
		save_mode_if=mode;		/* For return to rf */
		mode=LSB; set_mode(0);		/* Always LSB in audio */
		ASYNCOUT(ESC_CHAR);  ASYNCOUT('M');
		ASYNCOUT(0); ASYNCOUT(1);	/* This sets LSB (dsp usb)*/
		ASYNCOUT( fgainl[mike_gain] ); ASYNCOUT( fgainh[mike_gain] );
		}
	new_freq();

	setviewport(0, 3, 102, 16, DO_CLIP); /* Now transverter designation */
	clearviewport();
	settextjustify(1,2);
	setcolor(c_text_note);
	moveto(51, 3);
	if (input_if_af==0)  outtext(tvtr[band].name);

	setcolor(c_text_bold);
	setviewport(2,100,104,110,DO_CLIP);
	clearviewport();
	if (input_if_af && eme_doppler)		/* Only for audio processor */
	    {
	    sprintf(cc, "Fd=%dMHz", near_int(0.000001*f_eme) );
	    settextstyle(0,0,1);
	    settextjustify(CENTER_TEXT, TOP_TEXT);
	    moveto(51, 0);
	    outtext(cc);
	    }

	ASYNCOUT(ESC_CHAR);  ASYNCOUT('J');
	ASYNCOUT(input_if_af);
	ASYNCOUT(0); ASYNCOUT(3); ASYNCOUT(4);   /* Audio to L & R */
	}

/* set_affilt_sw()  -  Arrive with do_af_filt=0 for no af filter.
 */
void set_affilt_sw(void)
	{
	annotate_af_filt();
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('A');
	ASYNCOUT(do_af_filt);
	ASYNCOUT(do_LMS+1);
	ASYNCOUT(3); ASYNCOUT(4);
	}

/* set_affilt_select()  -  Arrive with af_coef (0, 7).
 * Sets the filter number (byte 1) and the gain inherent in the FIR coeffs
 * Variable kfilt is from the configuration file and adjusts whether
 * the noise level is constant as the bandwidth changes or the center
 * freq sine-wave gain is constant.
 *
 * filter_mask is a bit-by-bit mask for the 8 DSP audio filters.
 */
void set_affilt_select(void)
	{
	unsigned int byte2;
	int filter_test;

	/* First insure that a valid mask exists and then rotate through the
	 * filters until an active one is found */
	if (filter_mask==0 || (filter_mask&0XFF)==0)	/* Invalid situation */
	    af_coef=0;			/* So, use the first filter anyway */
	else
	    {
	    while (1)
		{
		filter_test=near_int(pow(2.0, (double)af_coef));
		if (filter_test & filter_mask)  break;	/* An active mode */
		if (++af_coef > 7) af_coef=0;
		}
	    }
	/* Following the DSP FIR there is a gain reduction in steps of 2.
	 * The number of steps is byte2. */
	byte2=0X000F & (unsigned)(kfilt*(double)filt[af_coef].fil_gain);
	annotate_af_filt();
	dsp_cmd('B', af_coef, byte2, 3, 4);
	}

/* set_delay_right(do_what)   Sets amount of delay put into right channel
 * in dsp.  do_what=0 means toggle delay on and off.
 */
void set_delay_right(int do_what)
	{
	int del;

	del=(int)(0.001+pow(2.0, (double)(delay_right+1)));
	if (do_what==0)
	    {
	    delay_state ^= 1;
	    }
	setviewport(63,333,126,342,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(64, 1);
	if (delay_state)
	    {
	    dsp_cmd('E', del, del>>8, 3, 4);
	    sprintf(cc, "%d", del);
	    outtext(cc);
	    }
	else
	    {
	    dsp_cmd('E', 1, 0, 3, 4);	/* Delay off is a delay of 1 */
	    outtext("Off");
	    }
	}

void select_denoise(void)
	{
	do_LMS += 1;
	if (do_LMS>1)  do_LMS = -1;
	set_denoise();			/* This updates the screen */
	dsp_cmd('A', do_af_filt, do_LMS+1, 3, 4);
	}

void set_denoise(void)
	{
	unsigned int autonotch_beta;

	unsigned int byte2;
	unsigned int dcynum;

	/* The beta number used in the DSP is the 0.5*(denoise_level)^2.
	   As denoise_level ranges (0, 100) beta ranges (1, 5202)
	   which as 1.15 is (0.0, 0.159). The square root compression
	   puts most of the range towards 0.0, e.g. 50 -> 0.041.
	 */

	/* The decay in the DSP is the number sent in byte 2, doubled
	   and subtracted from 32735. As denoise_decay ranges (0, 100)
	   the number sent, dcynum, ranges (254, 0) in steps of 2. In the
	   DSP the range is (32227, 32735) corresponding to 1.15 values
	   of (0.983, 0.999). The DSP decay is per W9GR and decays only
	   one of the coefficients per processing cycle, so the effective
	   decay rate is modified by 31 or 61, depending on the length.
	   The effective range is (0.99973 to 0.99998) for 61 length.
	   W9GR recommended a value around 1-0.00004=0.99996, or a
	   denoise_decay value of 75.
	 */
	dcynum=2*(int)(0.5+1.27*(double)(100-denoise_decay));
	byte2=dcynum | denoise_del_len;	/* Byte 2 */

	/* As autonotch_level ranges (0, 100) autonotch_beta
	   ranges (15, 1500). */
	autonotch_beta=(unsigned)
	    (1500.0/pow(1.0472,(double)(100-autonotch_level)));

	/* This is made wide on left to insure clearing the old decay data */
 	setviewport(95,323,127,332,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(RIGHT_TEXT, TOP_TEXT);
	moveto(32, 0);
	if (do_LMS==1)	/* De-Noise */
	    {
	    sprintf(cc, "%d", denoise_level);
	    outtext(cc);
	    setviewport(2,323,96,332,DO_CLIP);
	    clearviewport();
	    settextjustify(LEFT_TEXT, TOP_TEXT);
	    moveto(0, 0);
	    outtext("LMS Denoise");

	    setviewport(87,323,103,332,DO_CLIP);
	    clearviewport();
	    settextjustify(LEFT_TEXT, TOP_TEXT);
	    setcolor(c_text);
	    moveto(0, 0);
	    sprintf(cc, "%d", denoise_decay);
	    outtext(cc);
	    setcolor(c_text_bold);
	    }
	else if (do_LMS== -1)	/* Auto-Notch */
	    {
	    sprintf(cc, "%d", autonotch_level);
	    outtext(cc);
	    setviewport(2,323,96,332,DO_CLIP);
	    clearviewport();
	    settextjustify(LEFT_TEXT, TOP_TEXT);
	    moveto(0, 0);
	    outtext("LMS Notch");
	    }
	else
	    {
	    outtext("Off");
	    setviewport(2,323,96,332,DO_CLIP);
	    clearviewport();
	    settextjustify(LEFT_TEXT, TOP_TEXT);
	    moveto(0, 0);
	    outtext("LMS");
	    }
	/* Add 2 to beta to prevent beta in DSP going negative */
	dsp_cmd('C',2+denoise_level,byte2,autonotch_beta, autonotch_beta>>8);
	}


/* set_agc(int decay)  -  Sets the AGC decay rate in the DSP, or turns
 * AGC off by setting decay to 0.
 * Special case: No AGC in FM.
 */
void set_agc(int decay)
	{
	static int agcd[]={0, 32, 24, 16, 12, 8, 6, 4, 3, 2, 1};

	setviewport(63,413,126,422,DO_CLIP);
	clearviewport();
	setcolor(c_text_bold);
	settextstyle(0,0,1);
	settextjustify(RIGHT_TEXT, TOP_TEXT);
	moveto(64, 1);

	if (decay == -1)		/* Special case--AGC off for FM */
	    {
	    outtext("Off");
	    ASYNCOUT(ESC_CHAR);  ASYNCOUT('V');
	    ASYNCOUT(2); ASYNCOUT(3); ASYNCOUT(0); ASYNCOUT(0);
	    return;			/* Do not alter last_agc */
	    }
	    
	if (decay>0)
	    {
	    sprintf(cc, "%d", decay);
	    outtext(cc);
	    /* AGC on raises gain by 48 dB in DSP--compensate here.
	     * This does not affect fm_afgain, as there is no AGC in FM.
	     */
	    if (last_agc==0)  set_afgain(-48);
	    }
	else
	    {
	    outtext("Off");
	    if (last_agc > 0)  set_afgain(48);
	    }
	last_agc=decay;
	/* AGC Settings: */
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('V');
	ASYNCOUT(2); ASYNCOUT(3);
	ASYNCOUT(agcd[decay]&0XFF); ASYNCOUT((agcd[decay]>>8)&0XFF);
	}

void annotate_af_filt(void)
	{
	if (do_af_filt==0)
	   sprintf(cc, "%s", "None");
	else
	   sprintf(cc, "%s", filt[af_coef].name);
	setcolor(c_text_bold);
	setviewport(69,313,126,322,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(RIGHT_TEXT, TOP_TEXT);
	moveto(57, 0);
	outtext(cc);

	sprintf(cc, "%d", 1+af_coef);
	setcolor(c_text);
	settextjustify(LEFT_TEXT, TOP_TEXT);
	moveto(0, 0);
	outtext(cc);
	setcolor(c_text_bold);
	}

void set_dot(void)
	{dot_type += 1;
	 if  (dot_type>2) dot_type=0;
	}

/* disp_toggle()  As a resuly of the Alt Y command sets the display format
 * to either Large freq display (!waterfall), or to waterfall.
 */
void disp_toggle(void)
	{
	setviewport(0, 232, 70, 252, DO_CLIP);	/* For Contrast & Bright */
	clearviewport();
	settextjustify(0,2);
	settextstyle(0,0,1);
	if (waterfall)
	    {
	    setcolor(c_text_bold);
	    moveto(2,1);	outtext("Contrast");
	    moveto(2,11);	outtext("Bright");
	    setviewport(0, 0, 639, 479, DO_CLIP);
	    line(X2L, Y2T, X2R, Y2T);
	    setviewport(X2L+1, Y2T+1, X2R-1, Y2B-1, DO_CLIP);
	    clearviewport();
	    scroll_line=0;
	    setcolor(c_outline_bright);		/* Freq display border */
	    setviewport(0, 122, 127, 158, DO_CLIP);
	    setcolor(c_outline_bright); /* Freq display border */
	    rectangle(1,0,126,36);	/* Leave off corner pixels */
	    rectangle(0,1,127,35);
	    setcolor(c_text_bold);
	    new_freq();
	    if(mode==CW)
		{
		text_area=TRUE;
		print_medium(0, -1);	/* Initialize */
		}
	    if(mode==LHL7 || mode==PUA43)
		{
		text_area=TRUE;
		print_medium_1(0, -1);	/* Initialize */
		print_big(0, -1, 0);
		print_medium_2(0, -1);
		}
	    }
	else				/* Slide rule dial display */
	    {
	    setcolor(c_text);		/* Keep first uses viewport above */
	    moveto(2,1);	outtext("Contrast");
	    moveto(2,11);	outtext("Bright");
	    if (mode==CW)
		{
		morse(ADD_CH, 32);		/* Force a redraw of cw text */
		morse(ADD_CH, 8);		/* Back it out */
		}
	    if (mode==LHL7 || mode==PUA43 || mode==EME2 || mode==LTI)
		{
		mode=USB;
		set_mode(1);
		}
	    setcolor(c_text_bold);
	    draw_panel();
	    new_freq();
	    diag_mode=0;
	    switch (mode)
		{
		case CW: {
			sprintf(cc, "%s", "   CW");
			break;
			}
		case USB: {
			sprintf(cc, "%s", "  USB");
			break;
			}
		case LSB: {
			sprintf(cc, "%s", "  LSB");
			break;
			}
		case FM: {
			sprintf(cc, "%s", "   FM");
			break;
			}
		default: break;
		}
	    setviewport(X2L+1, Y2T+15, X2L+49, Y2T+45, DO_CLIP);
	    clearviewport();
	    settextstyle(0, 0, 1);
	    settextjustify(CENTER_TEXT, TOP_TEXT);
	    moveto(24, 10);
	    outtext(cc+2);
	    text_area=FALSE;
	    }			/* End !waterfall */
	}
	    
/* set_display_width()  -  Arrive with global display_width 0 or 1,
 * i.e., FULL_FREQ or HALF_FREQ.
 */
void set_display_width(void)
	{
	if (display_width)
		{if      (fft_bw==0) mark_frequency(500);
		 else if (fft_bw==1) mark_frequency(1000);
		 else                mark_frequency(2000);
		}
	else
		{if      (fft_bw==0) mark_frequency(1000);
		 else if (fft_bw==1) mark_frequency(2000);
		 else                mark_frequency(4000);
		}
	}

/* set_db_div(changed) is used to change the setting (changed==1) and
 * to redraw the display (changed==0).
 */
void set_db_div(int changed)
	{
	int i, yy, kdb;

	if (db_div_set==0)
		{
		kdb_div=3.2;
		kdb=10;
		}
	else if (db_div_set==1)
		{
		kdb_div=6.4;
		kdb=5;
		}
	else if (db_div_set==2)
		{
		kdb_div=16.0;
		kdb=2;
		}
	else                            /* if(db_div_set==3) */
		{
		kdb_div=32.0;
		kdb=1;
		}
	if (text_area)
	    setviewport(X1L+1, Y1T+35, X1L+68, Y1T+44, DO_CLIP);
	else
	    setviewport(X1L+1, Y1T+2, X1L+68, Y1T+11, DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(0,2);
	setcolor(c_text);
	moveto(1, 1);
	sprintf(cc, "%d dB/", kdb);
	outtext(cc);
	setcolor(c_text_bold);

	if (changed)
	    {
	    setviewport(102, Y1T, X1L-1, Y1B-8, DO_CLIP);
	    clearviewport();
	    settextjustify(1,1);
	    for (i=0; i<=7; i++)
		{
		yy =  1 + i*16;
		moveto(14, yy);
		sprintf(cc, "%d", kdb*(8 -i));
		outtext(cc);		/* dB level markings */
		}
	    setcolor(c_text_bold);
	    set_disp_offset(1);
	    }
	}

void set_disp_offset(int update_dsp)
	{
	if ( near_int(koffset) != 0)
		{
		if (text_area)
		    setviewport(X1R-60, Y1T+35, X1R-4, Y1T+44, DO_CLIP);
		else
		    setviewport(X1R-60, Y1T+2, X1R-4, Y1T+11, DO_CLIP);
		clearviewport();
		settextstyle(0,0,1);
		settextjustify(RIGHT_TEXT, TOP_TEXT);
		setcolor(c_text);
		moveto(55, 1);
		sprintf(cc, "%d dB", near_int(koffset) );
		outtext(cc);
		setcolor(c_text_bold);
		}
	/* power_exp is sent to the dsp to scale the power data before it is
	 * log converted to 8 bits.  The total dynamic range of the 8 bit data
	 * is 80 dB (lsb=80/256) and this factor keeps it centered in the
	 * 80 dB display area.
	 * In the following we add 99.0 to the offset to prevent (int)
	 * confusion at 0.  It is removed by an extra 33 in front.
	 * Note: not used except with 10 dB/div.  Therefore you can still
	 * overload if you offset more than -50 dB in 5 dB/div.
	 * Limit to this amount back in command.
	 * Variable win_offset is the approximate coherent loss of the window
	 * function, in 3 dB steps. This gives the power of 2 for scaling
	 * the dsp power spectrum data to almost correct for windowing.
	 * The remainder of the correction is win_adj_residual.
	 */
	if (db_div_set==0)		/* 0 --> 10 dB/div  */
	    {
	    power_exp = 43 + win_offset - ((int)(99.0+koffset))/3;
	    win_adj=win_adj_residual;	/* Part not included above */
	    }
	else				/* 5, 2, or 1 dB/div */
	    {
	    power_exp=10;
	    win_adj=win_adj_loss;	/* Coherent loss factor for window */
	    }

	if (update_dsp)
	    dsp_cmd(9, bits_8_16, power_exp, LOG_OFFSET, (LOG_OFFSET>>8));
	}

/* SET_NOCOHAVE(WHY) - Sets the amount of non-coherent averaging of 
 * spectral data. Why is the reason for being here, so that scrolling
 * through the modes does not leave a wierd value from EME2, or the like.
 * 
 * why==0  ==> here from a mode change.
 * why==1  ==> here from changing amount of noncohave.
 * why==2  ==> special case for LHL7, change in code_nx2 may need new noncohave
 */
void set_noncohave(int why)
	{
	double x;
	static int ncsave=-1;		/* -1 = never been set */
	static nclhlsave=-1;

	/* LHL7 mode allows a number of different tone lengths, all multiples
	 * of 2 sec. The noncoherent averaging time needs to be a multiple
	 * of 18 to be consistent with the basic 2 second multiple, and also
	 * needs to be less than the tone length. The following trees implement
	 * these constraints. Note LHL7 uses only fft_bw=2 (4800 Hz max)   */
	if (mode==LHL7)
	    {
	    if (why==0)			/* Mode change */
		{
		if (nclhlsave>0)
		    noncohave=nclhlsave;
		else
		    noncohave=3;
		}
	    else if (why==1)
		{
		if (noncohave==3 && d_noncoh==-1) return;  /* Never less */
	        switch(code_nx2)
		   {
		   case 30:
		    if (noncohave==3 && d_noncoh==1) {noncohave=6; break;}
		    if (noncohave==6 && d_noncoh==-1){ noncohave=3; break;}
		    if (noncohave==6 && d_noncoh==1) {noncohave=9; break;}
		    if (noncohave==9 && d_noncoh==-1) {noncohave=6; break;}
		    if (noncohave==9 && d_noncoh==1) {noncohave=15; break;}
		    if (noncohave==15 && d_noncoh==-1) {noncohave=9; break;}
		    if (noncohave==15 && d_noncoh==1) {noncohave=30; break;}
		    if (noncohave==30 && d_noncoh==-1) {noncohave=15; break;}
		    if (noncohave==30 && d_noncoh==1) {noncohave=45; break;}
		    if (noncohave==45 && d_noncoh==-1) {noncohave=30; break;}
		    if (noncohave==45 && d_noncoh==1) {noncohave=90; break;}
		    if (noncohave==90 && d_noncoh==-1) {noncohave=45; break;}
		    if (noncohave==90 && d_noncoh==1) break;
		   case 10:
		    if (noncohave==3 && d_noncoh==1) {noncohave=6; break;}
		    if (noncohave==6 && d_noncoh==-1) {noncohave=3; break;}
		    if (noncohave==6 && d_noncoh==1) {noncohave=15; break;}
		    if (noncohave==9) {noncohave=6; break;} /*Chng code_nx2*/
		    if (noncohave==15 && d_noncoh==-1) {noncohave=6; break;}
		    if (noncohave==15 && d_noncoh==1) {noncohave=30; break;}
		    if (noncohave==30 && d_noncoh==-1) {noncohave=15; break;}
		    if (noncohave==30 && d_noncoh==1) break;
		    if (noncohave>30) {noncohave=30; break;}
		   case 5:
		    if (noncohave==3 && d_noncoh==1) {noncohave=15; break;}
		    if (noncohave==6 || noncohave==9) {noncohave=3; break;}
		    if (noncohave==15 && d_noncoh==-1) {noncohave=3; break;}
		    if (noncohave==15 && d_noncoh==1) break;
		    if (noncohave>15) {noncohave=15; break;}
		   case 2:
		    if (noncohave==3 && d_noncoh==1) {noncohave=6; break;}
		    if (noncohave==6 && d_noncoh==-1) {noncohave=3; break;}
		    if (noncohave==6 && d_noncoh==1) break;
		    if (noncohave>6) {noncohave=6; break;}
		   case 1:
		    noncohave=3;
		    break;
		   }			/* End, switch(code_nx2) */
		nclhlsave=noncohave;
		}			/* End, why==1 */
	    else if (why==2)		/* Adjust noncohave to legal value */
		{
	        switch(code_nx2)
		   {
		   case 30:
		    break;		/* They are all valid */
		   case 10:
		    if (noncohave<=6) break;
		    if (noncohave==9) {noncohave=6; break;}
		    if (noncohave<=30) break;
		    if (noncohave>30) {noncohave=30; break;}
		   case 5:
		    if (noncohave==3 || noncohave==15) break;
		    if (noncohave==6 || noncohave==9) {noncohave=3; break;}
		    noncohave=15;
		    break;
		   case 2:
		    if (noncohave==3 || noncohave==6) break;
		    noncohave=6;
		    break;
		   case 1:
		    noncohave=3;
		    break;
		   }			/* End, switch(code_nx2) */
		nclhlsave=noncohave;
		}			/* End, why==2 */
	    }				/* End LHL7 */
	else if (mode==EME2)
	    {
	    /*  if (why==0) */      noncohave=3;
	    }
	else if (mode==PUA43)
	    {
	    noncohave=3;
	    }
	else				/* All other modes,  CW, USB, etc */
	    {
	    if (why==0 && ncsave>0) noncohave=ncsave;
	    if (d_noncoh==-1)
		{
		noncohave--;
		if (noncohave<1) noncohave=1;
		}
	     else if (d_noncoh==1)
		{
		noncohave++;
		if (noncohave>1667) noncohave=1667;
		}
		ncsave=noncohave;	/* Needed if mode is changed */
	    }				/* End, other modes */

	d_noncoh=0;			/* In case reenter for mode change */

	/* Print as integer if not xx.5 */
	x=fft_per[fft_bw]*(double)noncohave;
	if ((int)(x+.01) == (int)(x+0.51))
	    sprintf(cc, "%d", (int)(x+0.01) );
	else
	    sprintf(cc, "%.1f", x);
	setviewport(92,368,126,377,DO_CLIP);
	clearviewport();
	settextstyle(0,0,1);
	settextjustify(RIGHT_TEXT, TOP_TEXT);
	moveto(33, 1);	outtext(cc);
	}

/* read_config_file()  -  Restores transceivers to previous settings
 * using UHFA.CFG.
 * Returns:	 0  Non-fatal error, or no file found.
 *		 1  Successful file read.
 */
int read_config_file(void)
	{
	int i;
	static char entrysep[]=" \t,=\n\r";	/* Allow sp, tab, comma,
						   equal, lf, cr  */
	char *ci, *cd;
	char cline[260], cline2[260];

        if (NULL == ( config_file=fopen("UHFA.CFG", "rt") ))
	    {
	    printf("No UHFA.CFG file.  Using default configuration\n");
	    return 0;
	    }

	if (fgets(cline, 256, config_file) == NULL)
	    {
	    printf("Empty UHFA.CFG file.  Using default configuration\n");
	    fclose(config_file);
	    return 0;
	    }
 	ci = strtok(cline, entrysep);	/* Should be file identifier line */
	if (strcmp(ci, "UHFACFG2"))
	    {
	    printf("Illegal config file, not UHFACFG2. Use default config\n");
	    fclose(config_file);
	    return 0;
	    }
	while (fgets(cline, 256, config_file))
	    {
	    strcpy(cline2, cline);	/* Extra for when string token first */
	    /* Get identifier and skip return:   */
	    if ((ci = strtok(cline, entrysep)) == NULL)  break;
	    /* Get a token to cd */
	    if ((cd = strtok(NULL, entrysep)) == NULL)  strcpy(cd, "BLANK");

	    if      (!stricmp(ci, "contrast")) contrast=atoi(cd);
	    else if (!stricmp(ci, "bright"))  bright=atoi(cd);
	    else if (!stricmp(ci, "afgain"))  afgain=atoi(cd);
	    else if (!stricmp(ci, "fm_afgain"))  fm_afgain=atoi(cd);
	    else if (!stricmp(ci, "rfgain"))  rfgain=atoi(cd);
	    else if (!stricmp(ci, "rfgdb"))
		{
		rfg[1].dbgain=atof(cd);
		for (i=2; i<=6; i++)
		    {
		    if ((cd=strtok(NULL, entrysep))==0)
			{
			printf("Missing rfgdb entry in UHFA.CFG (need 6)\n");
			return 0;
			}
		    else
			{
			rfg[i].dbgain=atof(cd);
			}
		    }
		}
	    else if (!stricmp(ci, "gain_adj"))  gain_adj=atof(cd);
	    else if (!stricmp(ci, "cw_speed"))  cw_speed=atoi(cd);
	    else if (!stricmp(ci, "agc_decay"))
		{agc_decay=atoi(cd);
		 last_agc=agc_decay;	/* Get started right */
		}
	    else if (!stricmp(ci, "mike_gain"))  mike_gain=atoi(cd);
	    else if (!stricmp(ci, "xmit_power"))  xmit_power=atoi(cd);
	    else if (!stricmp(ci, "fm_squelch"))  fm_squelch=atoi(cd);
	    else if (!stricmp(ci, "cwoffset"))  cwoffset=atoi(cd);
	    else if (!stricmp(ci, "eme_doppler")) eme_doppler=atoi(cd);
	    else if (!stricmp(ci, "sidetone_lev"))  sidetone_lev=atoi(cd);
	    else if (!stricmp(ci, "noncohave"))
		{noncohave=atoi(cd);
		 if (noncohave<1) noncohave=1;
		}
	    else if (!stricmp(ci, "do_window"))  do_window=atoi(cd);
	    else if (!stricmp(ci, "fft_bw"))  { fft_bw=atoi(cd);
						fft_bw_save=fft_bw;}
	    else if (!stricmp(ci, "knoise"))  knoise=atof(cd);
	    else if (!stricmp(ci, "kfilt"))  kfilt=atof(cd);
	    else if (!stricmp(ci, "fft_fcn"))  fft_fcn=atoi(cd);
	    else if (!stricmp(ci, "do_af_filt"))  do_af_filt=atoi(cd);
	    else if (!stricmp(ci, "af_coef"))  af_coef=atoi(cd);
	    else if (!stricmp(ci, "do_LMS"))  do_LMS=atoi(cd);
	    else if (!stricmp(ci, "denoise_level"))  denoise_level=atoi(cd);
	    else if (!stricmp(ci, "denoise_decay"))  denoise_decay=atoi(cd);
	    else if (!stricmp(ci, "denoise_del_len")) denoise_del_len=atoi(cd);
	    else if (!stricmp(ci, "autonotch_level")) autonotch_level=atoi(cd);
	    else if (!stricmp(ci, "dot_type"))  dot_type=atoi(cd);
	    else if (!stricmp(ci, "display_width"))  display_width=atoi(cd);
	    else if (!stricmp(ci, "freq_low"))  freq_low=atof(cd);
	    else if (!stricmp(ci, "db_div_set"))  db_div_set=atoi(cd);
	    else if (!stricmp(ci, "koffset"))  koffset=atof(cd);
	    else if (!stricmp(ci, "mode"))  mode=atoi(cd);
	    else if (!stricmp(ci, "mode_mask"))  mode_mask=(1 | atoi(cd));
	    else if (!stricmp(ci, "bits_8_16")) { bits_8_16=atoi(cd);
						  if (mode==EME2 || mode==LTI)
							 save_8_16=0;
						  else save_8_16=bits_8_16; }
	    else if (!stricmp(ci, "input_if_af"))  input_if_af=atoi(cd);
	    else if (!stricmp(ci, "hardware"))  hardware=atoi(cd);
	    else if (!stricmp(ci, "hardware_ptt"))  hardware_ptt=atoi(cd);
	    else if (!stricmp(ci, "rit"))  rit=atoi(cd);
	    else if (!stricmp(ci, "rit_on"))  rit_on=atoi(cd);
	    else if (!stricmp(ci, "band"))  band=atoi(cd);
	    else if (!stricmp(ci, "tfreqL"))  tfreqL=atol(cd);
	    else if (!stricmp(ci, "ref_offset"))  ref_offset=atof(cd);
	    else if (!stricmp(ci, "diag_mode"))  diag_mode=atoi(cd);
	    else if (!stricmp(ci, "startup_detail")) startup_detail=atoi(cd);
	    else if (!stricmp(ci, "waterfall"))  waterfall=atoi(cd);
	    else if (!stricmp(ci, "shift_reg"))  shift_reg=atoi(cd);
	    else if (!stricmp(ci, "eme2_te"))  eme2_te=atof(cd);
	    else if (!stricmp(ci, "eme_freq_range"))  eme_freq_range=atof(cd);
	    else if (!stricmp(ci, "code_nb_ratio"))  code_nb_ratio=atof(cd);
	    else if (!stricmp(ci, "lterm_int_flow"))  lterm_int_flow=atof(cd);
	    else if (!stricmp(ci, "lterm_int_fhigh")) lterm_int_fhigh=atof(cd);
	    else if (!stricmp(ci, "lterm_int_center"))
		 lterm_int_center=atof(cd);
	    else if (!stricmp(ci, "lti_freq_range")) lti_freq_range=atof(cd);
	    else if (!stricmp(ci, "lhl_bins"))  lhl_bins=atoi(cd);
	    else if (!stricmp(ci, "lhl_peak_ave")) lhl_peak_ave=atoi(cd);
	    else if (!stricmp(ci, "lhl7_2line")) lhl7_2line=atoi(cd);
	    else if (!stricmp(ci, "pua43_noisedecay"))
			 pua43_noisedecay=atof(cd);
	    else if (!stricmp(ci, "pua43_dorandom")) pua43_dorandom=atoi(cd);
	    else if (!stricmp(ci, "pua43_xstring"))
		{
		cd=strtok(cline2, "|");		/* Go to string */
		cd=strtok(NULL, "|");		/* Get string */
		cd[28]=0;			/* Eventual end */
		if (strlen(cd) < 28)
		    for (i=strlen(cd); i<28; i++)
			cd[i]=32;
		strcpy(pua43_xstring, cd);
		}
	    else if (!stricmp(ci, "pua14_28")) pua14_28=atoi(cd);
	    else if (!stricmp(ci, "dly_ant2amp"))  dly_ant2amp=atoi(cd);
	    else if (!stricmp(ci, "dly_amp2xmit"))  dly_amp2xmit=atoi(cd);
	    else if (!stricmp(ci, "dly_xmit2amp"))  dly_xmit2amp=atoi(cd);
	    else if (!stricmp(ci, "dly_amp2ant"))  dly_amp2ant=atoi(cd);
	    else if (!stricmp(ci, "dly_emet2r"))  dly_emet2r=atoi(cd);
	    else if (!stricmp(ci, "dut1"))  dut1=atof(cd);
	    else if (!stricmp(ci, "time_type"))  time_type=atoi(cd);
	    else if (!stricmp(ci, "show_secs"))  show_secs=atoi(cd);
	    else if (!stricmp(ci, "time_str"))
		{
		cd=strtok(cline2, "|");		/* Go to time_str */
		cd=strtok(NULL, "|");		/* Get string */
		strcpy(time_str, cd);
		}
	    else if (!stricmp(ci, "tick_time")) tick_time=atoi(cd);
	    else if (!stricmp(ci, "clock_speed"))  clock_cor_factor=atof(cd);
	    else if (!stricmp(ci, "code_nx2")) code_nx2=atoi(cd);
	    else if (!stricmp(ci, "code_id"))
		{
		cd=strtok(cline2, "|");		/* Go to code_id */
		cd=strtok(NULL, "|");		/* Get string */
		cd[9]=0;			/* For excessive length */
		strcpy(code_id, cd);
		}
	    else if (!stricmp(ci, "call_str"))
		{
		cd=strtok(cline2, "|");		/* Go to call_str string */
		cd=strtok(NULL, "|");		/* Get string */
		strcpy(call_str, cd);
		}
	    else if (!stricmp(ci, "quit_save_state")) quit_save_state=atoi(cd);
	    else if (!stricmp(ci, "beep_on_error"))  beep_on_error=atoi(cd);
	    else if (!stricmp(ci, "palette"))
		{
		upalette.colors[0]=atoi(cd);
		for (i=1; i<16; i++)
		    {
		    if ((cd=strtok(NULL, entrysep))==0)
			{
			printf("Missing color entry in UHFA.CFG\n");
			return 0;
			}
		    else
			{
			upalette.colors[i]=atoi(cd);
			}
		    }
		}
	    else if (!stricmp(ci, "c_background"))  c_background=atoi(cd);
	    else if (!stricmp(ci, "c_outline")) c_outline=atoi(cd);
	    else if (!stricmp(ci, "c_outline_bold"))  c_outline_bold=atoi(cd);
	    else if (!stricmp(ci, "c_outline_bright"))
		c_outline_bright=atoi(cd);
	    else if (!stricmp(ci, "c_text"))  c_text=atoi(cd);
	    else if (!stricmp(ci, "c_text_bold"))  c_text_bold=atoi(cd);
	    else if (!stricmp(ci, "c_text_bright"))  c_text_bright=atoi(cd);
	    else if (!stricmp(ci, "c_text_note"))  c_text_note=atoi(cd);
	    else if (!stricmp(ci, "c_plot_bold"))  c_plot_bold=atoi(cd);
	    else if (!stricmp(ci, "c_plot_bright"))  c_plot_bright=atoi(cd);
	    else if (!stricmp(ci, "c_plot_marker"))  c_plot_marker=atoi(cd);
	    else if (!stricmp(ci, "c_bar_dark"))  c_bar_dark=atoi(cd);
	    else if (!stricmp(ci, "c_bar_low"))  c_bar_low=atoi(cd);
	    else if (!stricmp(ci, "c_bar_medium"))  c_bar_medium=atoi(cd);
	    else if (!stricmp(ci, "c_bar_high"))  c_bar_high=atoi(cd);
	    else if (!stricmp(ci, "kbd_alt1"))  kbd_alt1=atoi(cd);
	    else if (!stricmp(ci, "kbd_alt2"))  kbd_alt2=atoi(cd);
	    else if (!stricmp(ci, "kbd_alt3"))  kbd_alt3=atoi(cd);
/*	    else if (!stricmp(ci, "scs_key"))  scs_key=atoi(cd); */
	    else if ( !stricmp(ci,"tvtr") && (i=atoi(cd))>=0 && i<=20 )
		{
		cd=strtok(NULL, "|");		/* Move to string */
		cd=strtok(NULL, "|");		/* Get name string */
		strcpy(tvtr[i].name, cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].lo_type=atoi(cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].sb=atoi(cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].lo_freq=atof(cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].chain_mult=atoi(cd);
		strtok(NULL, "|");
		cd=strtok(NULL, "|");
		strcpy(tvtr[i].init_str, cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].gain_db=atof(cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].control=atoi(cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].bus_adr=atoi(cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].last_tfreq=atof(cd);
		cd=strtok(NULL, entrysep);
		tvtr[i].last_mode=atoi(cd);
		}
	    else if ( !stricmp(ci,"loc") && (i=atoi(cd))>=0 && i<=9 )
		{
		loc[i].time_zone=0;		/* Mark as existant */
		cd=strtok(NULL, "|");		/* Move to string */
		cd=strtok(NULL, "|");		/* Get short name string */
		if (strlen(cd)>6)
		    {
		    printf("\nLoc[%d].nm exceeds 6 ch; Loc not used", i);
		    loc[i].time_zone=999;
		    }
		else
		    strcpy(loc[i].nm, cd);
		strtok(NULL, "|");
		cd=strtok(NULL, "|");		/* Get description */
		if (strlen(cd)>31)
		    {
		    printf("\nLoc[%d].name exceeds 31 ch; Loc not used", i);
		    loc[i].time_zone=999;
		    }
		else
		    strcpy(loc[i].name, cd);
		cd=strtok(NULL, entrysep);
		loc[i].lat=d2r(atof(cd));
		cd=strtok(NULL, entrysep);
		loc[i].lon=d2r(atof(cd));
		}
	    else if (!stricmp(ci, "f_eme"))  f_eme=atof(cd);
	    else if (!stricmp(ci, "cur_loc"))  cur_loc=atoi(cd);
	    else if (!stricmp(ci, "delay_state"))  delay_state=atoi(cd);
	    else if (!stricmp(ci, "delay_right"))  delay_right=atoi(cd);
	    else if (!stricmp(ci, "lpt_port"))  lpt_port=atoi(cd);
	    else if (!stricmp(ci, "filter_mask"))  filter_mask=atoi(cd);
	    /* Filter 8 is not available from .CFG as it is "custom" */
	    else if ( !stricmp(ci,"fildat") && (i=atoi(cd))>=1 && i<=7 )
		{
		cd=strtok(NULL, "|");		/* Move to string */
		cd=strtok(NULL, "|");		/* Get name string */
		if (strlen(cd)>6)
		    {
		    printf("\nfildat[%d].name exceeds 6 ch; Data not used", i);
		    sprintf(filt[i-1].name, "FILT%d", i);
		    }
		else
		    strcpy(filt[i-1].name, cd);

		strtok(NULL, "|");
		cd=strtok(NULL, "|");
		if (strlen(cd)>30)
		    {
		    printf("\nfildat[%d].file_name > 30 ch; Data not used", i);
		    strcpy(filt[i-1].file_name, " ");
		    }
		else
		    strcpy(filt[i-1].file_name, cd);
		cd=strtok(NULL, entrysep);
		filt[i-1].fil_type=atoi(cd);
		cd=strtok(NULL, entrysep);
		filt[i-1].param1=atof(cd);
		cd=strtok(NULL, entrysep);
		filt[i-1].param2=atof(cd);
		cd=strtok(NULL, entrysep);
		filt[i-1].param3=atoi(cd);
		cd=strtok(NULL, entrysep);
		filt[i-1].fclow=atoi(cd);
		cd=strtok(NULL, entrysep);
		filt[i-1].fchigh=atoi(cd);
		cd=strtok(NULL, entrysep);
		filt[i-1].fil_gain=atoi(cd);
		}
	    else if (!stricmp(ci, "flow"))  flow=atof(cd);
	    else if (!stricmp(ci, "fhigh"))  fhigh=atof(cd);
	    else if (!stricmp(ci, "kaiser_db"))  kaiser_db=atof(cd);
	    else if (!stricmp(ci, "fil8_use"))  fil8_use=atoi(cd);
	    else if (!stricmp(ci, "fil8_name"))
		{
		cd=strtok(cline2, "|");		/* Move to string */
		cd=strtok(NULL, "|");		/* Get short name string */
		cd[6]=0;			/* Limit length */
		strcpy(fil8_name, cd);
		}
	    else if (!stricmp(ci, "aeqrdb"))
		{
		aeqrdb[0]=atof(cd);
		for (i=1; i<NFRB-1; i++)
		    {
		    if ((cd=strtok(NULL, entrysep))==0)
			{
			printf("Missing aeqrdb equalizer entry in UHFA.CFG\n");
			return 0;
			}
		    else
			{
			aeqrdb[i]=atof(cd);
			}
		    }
		}
	    /* Note: comments will fall through here */
	    }
	fclose(config_file);
	return 1;
	}

/* save_state()  -  Used at shut down to bring the state of the transceiver
 * back to last state.  This will replace uhfa.cfg file.
 * The first line must be "UHFACFG2".  The fprintf below summarizes the
 * the flexible format.  The special rules for strings come from strtok()
 * that gags on empty strings.  The use of the '|' allows spaces within the
 * string for titles like "432 MHz Tvtr".
 */
void save_state(void)
	{
	int i;

	if (NULL==( config_file=fopen("UHFA.CFG", "wt") ))
	    do_t_error(42, "E42 CFG File", 200);
	else
	    {
	    rewind(config_file);
	    fprintf(config_file, "UHFACFG2\n");
	    fprintf(config_file,
		"// Note: Use double space around .name and .init_str\n");
	    fprintf(config_file,
		"// and demark each string with vertical bars '|'.\n");
	    fprintf(config_file,
		"// Empty strings should have at least a space in them.\n");
	    fprintf(config_file,
		"// Comments start with // and go to end of line\n");
	    fprintf(config_file,
		"// Fields can be delimited by space, tab, ',' ';' or '='\n");
	    fprintf(config_file,
		"//  \n");
	    fprintf(config_file,
		"// Up  to 10 locations are given by:\n");
	    fprintf(config_file,
		"//     'loc i  |Name|  |Description|  lat lon'\n");
	    fprintf(config_file,
		"// where i is (0,9) and lat and lon are decimal degrees,\n");
	    fprintf(config_file,
		"// with West being negative longitude. i=0 is home loc.\n");
	    for (i=0; i<=9; i++)
	      {
	      if (loc[i].time_zone != 999)
		    {
		    if (strlen(loc[i].nm)==0)
			strcpy(loc[i].nm, " ");
		    if (strlen(loc[i].name)==0)
			strcpy(loc[i].name, " ");
		    fprintf(config_file,
			"loc %d  |%s|  |%s|  %.5lf %.5lf\n",
			i, loc[i].nm, loc[i].name,
			r2d(loc[i].lat), r2d(loc[i].lon));
		    }
	      }
	    fprintf(config_file, "cur_loc %d\n", cur_loc);
	    fprintf(config_file,
		"// f_eme in Hz, used for audio processing only\n");
	    fprintf(config_file, "f_eme %.1lf\n", f_eme);
	    for (i=0; i<21; i++)
	      {
	      if (tvtr[i].lo_type < 999)
		{
		if (strlen(tvtr[i].name)==0)
			strcpy(tvtr[i].name, " ");
		if (strlen(tvtr[i].init_str)==0)
			strcpy(tvtr[i].init_str, " ");
	        fprintf(config_file,
		    "tvtr %d  |%s|  %d %d %.1f %d  |%s|  %.3f %d %d %.1f %d\n",
		    i, tvtr[i].name, tvtr[i].lo_type, tvtr[i].sb,
		    tvtr[i].lo_freq, tvtr[i].chain_mult, tvtr[i].init_str,
		    tvtr[i].gain_db, tvtr[i].control, tvtr[i].bus_adr,
		    tvtr[i].last_tfreq, tvtr[i].last_mode);
		}
	      }
	    fprintf(config_file, "contrast %d\n", contrast);
	    fprintf(config_file, "bright %d\n",  bright);
	    fprintf(config_file, "afgain %d\n",  afgain);
	    fprintf(config_file, "fm_afgain %d\n",  fm_afgain);
	    fprintf(config_file, "rfgain %d\n",  rfgain);
	    fprintf(config_file,
		"// The following six entries are the actual attenuation\n");
	    fprintf(config_file,
		"// for rf gain settings 94, 88,...,64. Nominal values\n");
	    fprintf(config_file,
		"// of -6.0, -12.0,...,-36.0.\n");
	    fprintf(config_file,
		"rfgdb %.2lf %.2lf %.2lf %.2lf %.2lf %.2lf\n", 
		rfg[1].dbgain, rfg[2].dbgain, rfg[3].dbgain, 
		rfg[4].dbgain, rfg[5].dbgain, rfg[6].dbgain);
	    fprintf(config_file,
		"// gain_adj is small dB value to adjust power displays.\n");
	    fprintf(config_file, "gain_adj %.2lf\n", gain_adj);
	    fprintf(config_file, "cw_speed %d\n",  cw_speed);
	    fprintf(config_file, "agc_decay %d\n",  agc_decay);
	    fprintf(config_file, "mike_gain %d\n",  mike_gain);
	    fprintf(config_file, "xmit_power %d\n",  xmit_power);
	    fprintf(config_file, "fm_squelch %d\n",  fm_squelch);
	    fprintf(config_file, "cwoffset %d\n",  cwoffset);
	    fprintf(config_file, "eme_doppler %d\n",  eme_doppler);
	    fprintf(config_file, "sidetone_lev %d\n",  sidetone_lev);
	    fprintf(config_file, "noncohave %d\n",  noncohave);
	    fprintf(config_file,
		"// do_window=0,1,2,3 for None, Tukey25, Hamming or Bh-92.\n");
	    fprintf(config_file, "do_window %d\n",  do_window);
	    fprintf(config_file,
		"// fft_bw=0,1,2 for 1200, 2400, 4800 Hz.\n");
	    fprintf(config_file, "fft_bw %d\n",  fft_bw);
	    fprintf(config_file,
	     "// knoise:1.0 constant sig; 0.0 constant noise pwr, display\n");
	    fprintf(config_file, "knoise %lf\n",  knoise);
	    fprintf(config_file,
	     "// kfilt:1.0 constant sig; 0.0 constant noise pwr, FIR filt\n");
	    fprintf(config_file, "kfilt %lf\n",  kfilt);
	    fprintf(config_file, "fft_fcn %d\n",  fft_fcn);
	    fprintf(config_file, "do_af_filt %d\n",  do_af_filt);
	    fprintf(config_file, "af_coef %d\n",  af_coef);
	    fprintf(config_file, "do_LMS %d\n",  do_LMS);
	    fprintf(config_file, "denoise_level %d\n",  denoise_level);
	    fprintf(config_file, "denoise_decay %d\n",  denoise_decay);
	    fprintf(config_file,
		 "// denoise_del_len=0 for 31; =1 for 61 delay line length\n");
	    fprintf(config_file, "denoise_del_len %d\n",  denoise_del_len);
	    fprintf(config_file, "autonotch_level %d\n",  autonotch_level);
	    fprintf(config_file, "delay_state %d\n", delay_state);
	    fprintf(config_file, "delay_right %d\n", delay_right);
	    fprintf(config_file, "dot_type %d\n",  dot_type);
	    fprintf(config_file, "display_width %d\n",  display_width);
	    fprintf(config_file, "// freq_low is lowest freq marker\n");
	    fprintf(config_file, "freq_low %lf\n",  freq_low);
	    fprintf(config_file, "db_div_set %d\n",  db_div_set);
	    fprintf(config_file, "koffset %lf\n",  koffset);
	    fprintf(config_file, "mode %d\n",  mode);
	    fprintf(config_file,
		"// mode_mask is sum of values for desired modes:\n");
	    fprintf(config_file,
	 "// 1=CW, 2=USB, 4=LSB, 8=FM, 16=LHL7, 32=PUA43, 64=EME2, 128=LTI\n");
	    fprintf(config_file, "mode_mask %d\n",  mode_mask);
	    fprintf(config_file, "bits_8_16 %d\n",  bits_8_16);
	    fprintf(config_file, "input_if_af %d\n",  input_if_af);
	    fprintf(config_file, "hardware %d\n",  hardware);
	    fprintf(config_file, "hardware_ptt %d\n",  hardware_ptt);
	    fprintf(config_file, "rit %d\n",  rit);
	    fprintf(config_file, "rit_on %d\n",  rit_on);
	    fprintf(config_file, "band %d\n",  band);
	    if (input_if_af==0)
		fprintf(config_file, "tfreqL %ld\n",  tfreqL);
	    fprintf(config_file, "ref_offset %15.13lf\n",  ref_offset);
	    fprintf(config_file, "diag_mode %d\n",  diag_mode);
	    fprintf(config_file,
		"// startup_detail=1 for diagnostic text screens:\n");
	    fprintf(config_file, "startup_detail %d\n",  startup_detail);
	    fprintf(config_file,
		"// waterfall=1 for weak-signal waterfall,\n");
	    fprintf(config_file,
		"//          =0 for big dial & help summary screen\n");
	    fprintf(config_file, "waterfall %d\n",  waterfall);
	    fprintf(config_file, "shift_reg %d\n",  shift_reg);
	    fprintf(config_file, "eme2_te %lf\n", eme2_te);
	    fprintf(config_file, "eme_freq_range %lf\n", eme_freq_range);
	    fprintf(config_file, "code_nb_ratio %lf\n", code_nb_ratio);
	    fprintf(config_file, "lterm_int_flow %.1lf\n", lterm_int_flow);
	    fprintf(config_file, "lterm_int_fhigh %.1lf\n", lterm_int_fhigh);
	    fprintf(config_file, "lterm_int_center %.1lf\n", lterm_int_center);
	    fprintf(config_file, "lti_freq_range %.1lf\n", lti_freq_range);
	    fprintf(config_file, "lhl_bins %d\n",  lhl_bins);
	    fprintf(config_file, "lhl_peak_ave %d\n",  lhl_peak_ave);
	    fprintf(config_file, "lhl7_2line %d\n",  lhl7_2line);
	    fprintf(config_file, "pua43_noisedecay %.6lf\n", pua43_noisedecay);
	    fprintf(config_file, "pua43_dorandom %d\n", pua43_dorandom);
	    fprintf(config_file, "pua43_xstring  |%s|\n", pua43_xstring);
	    fprintf(config_file, "pua14_28 %d\n",  pua14_28);
	    fprintf(config_file, "dly_ant2amp %d\n",  dly_ant2amp);
	    fprintf(config_file, "dly_amp2xmit %d\n",  dly_amp2xmit);
	    fprintf(config_file, "dly_xmit2amp %d\n",  dly_xmit2amp);
	    fprintf(config_file, "dly_amp2ant %d\n",  dly_amp2ant);
	    fprintf(config_file, "dly_emet2r %d\n",  dly_emet2r);
	    fprintf(config_file, "dut1 %.3lf\n",  dut1);
	    fprintf(config_file, "quit_save_state %d\n",  quit_save_state);
	    fprintf(config_file, "// Following defines palette and screen:\n");
	    fprintf(config_file,
		"palette %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n", 
		 upalette.colors[0],  upalette.colors[1],  upalette.colors[2],
		 upalette.colors[3],  upalette.colors[4],  upalette.colors[5],
		 upalette.colors[6],  upalette.colors[7],  upalette.colors[8],
		 upalette.colors[9],  upalette.colors[10], upalette.colors[11],
		 upalette.colors[12], upalette.colors[13], upalette.colors[14],
		 upalette.colors[15]);
	    fprintf(config_file, "c_background %d\n", c_background);
	    fprintf(config_file, "c_outline %d\n", c_outline);
	    fprintf(config_file, "c_outline_bold %d\n", c_outline_bold);
	    fprintf(config_file, "c_outline_bright %d\n",c_outline_bright);
	    fprintf(config_file, "c_text %d\n", c_text);
	    fprintf(config_file, "c_text_bold %d\n", c_text_bold);
	    fprintf(config_file, "c_text_bright %d\n", c_text_bright);
	    fprintf(config_file, "c_text_note %d\n", c_text_note);
	    fprintf(config_file, "c_plot_bold %d\n", c_plot_bold);
	    fprintf(config_file, "c_plot_bright %d\n", c_plot_bright);
	    fprintf(config_file, "c_plot_marker %d\n", c_plot_marker);
	    fprintf(config_file, "c_bar_dark %d\n", c_bar_dark);
	    fprintf(config_file, "c_bar_low %d\n", c_bar_low);
	    fprintf(config_file, "c_bar_medium %d\n", c_bar_medium);
	    fprintf(config_file, "c_bar_high %d\n", c_bar_high);
	    fprintf(config_file, "kbd_alt1 %d\n", kbd_alt1);
	    fprintf(config_file, "kbd_alt2 %d\n", kbd_alt2);
	    fprintf(config_file, "kbd_alt3 %d\n", kbd_alt3);
/*	    fprintf(config_file, "scs_key %d\n", scs_key); */
	    fprintf(config_file, "beep_on_error %d\n", beep_on_error);
	    fprintf(config_file, "code_nx2 %d\n", code_nx2);
	    fprintf(config_file,
		"// code_id for CW ID during FSK code, e.g. |DE W7XYZ|\n");
	    if (strlen(code_id)==0)
		strcpy(code_id, "DE ?");
	    fprintf(config_file, "code_id  |%s|\n", code_id);
	    fprintf(config_file,
		"// time_type=0 is GMT display, 1=computer time\n");
	    fprintf(config_file, "show_secs %d\n", show_secs);
	    fprintf(config_file, "time_type %d\n", time_type);
	    if (strlen(time_str)==0)
		strcpy(time_str, ".");
	    fprintf(config_file, "time_str  |%s|\n", time_str);
	    fprintf(config_file, "tick_time %d\n", tick_time);
	    fprintf(config_file, "clock_speed %.7lf\n", clock_cor_factor);
	    fprintf(config_file,
		"// call_str is a string printed at top of upper-left\n");
	    fprintf(config_file,
		"// identification box, e.g. |W7XYZ|\n");
	    if (strlen(call_str)==0)
		strcpy(call_str, ".");
	    fprintf(config_file, "call_str  |%s|\n", call_str);
	    fprintf(config_file,
		"// lpt_port=0 for none, 888 for LPT1 (0x378),\n");
	    fprintf(config_file,
		"// or 956 for LPT2 (0x3BC) (confirm your computer)\n");
	    fprintf(config_file, "lpt_port %d\n", lpt_port);
	    fprintf(config_file,
	      "//The fildat data is informational input to the PC program.\n");
	    fprintf(config_file,
	      "//This data does NOT design filters.\n");
	    for (i=0; i<7; i++)
		{
		if (strlen(filt[i].name)==0)
			strcpy(filt[i].name, " ");
		if (strlen(filt[i].file_name)==0)
			strcpy(filt[i].file_name, " ");
	        fprintf(config_file,
		    "fildat %d  |%s|  |%s|  %d %.3lf %.6f %d %d %d %d\n",
		    i+1, filt[i].name, filt[i].file_name,
		    filt[i].fil_type, filt[i].param1,
		    /* filt[i].param2 */   0.0,
		    /* filt[i].param3 */   0,
		    filt[i].fclow, filt[i].fchigh,
		    filt[i].fil_gain);
		}
	    fprintf(config_file,
		"// filter mask=sum[1=Filter1, 2=Filter2,...,128=Filter8]\n");
	    fprintf(config_file, "filter_mask %d\n", filter_mask);
	    fprintf(config_file, "flow %.1lf\n", flow);
	    fprintf(config_file, "fhigh %.1lf\n", fhigh);
	    fprintf(config_file, "kaiser_db %.1lf\n", kaiser_db);
	    fprintf(config_file, "fil8_use %d\n", fil8_use);
	    fprintf(config_file, "fil8_name  |%s|\n", fil8_name);
	    fprintf(config_file, "// Rcv Equalizer dB:\n");
	    fprintf(config_file, "aeqrdb ");
	    for (i=0; i<NFRB-1; i++)
		fprintf(config_file, "%6.1lf ", aeqrdb[i]);
	    fprintf(config_file, "\n");
	    fprintf(config_file, "//     0      200    400    600    800   1000  ");
	    fprintf(config_file, "1300   1600   2000   2400   2800   3200");
	    fprintf(config_file, "   3600   4000   4400\n");

	    if (fclose(config_file) != NULL)
		do_t_error(42, "E42 CFG File", 200);
	    }
	}

/* Save long term power for extracting norm[] off-line.
 */
void save_lt_pwr(void)
	{
	char ct[15];
	int i;
	double lt_pave_center;
	double rel_pwr_db;
	time_t t0;
	struct tm *tgp;

	cc[0]='P';

	if (ms_available)
	    t0=update_clock(1)+dsp_clock_base;
	else
	    (void)time(&t0);
	if (time_type)
	    tgp = localtime(&t0);	/* time_type==1 for local */
	else
	    tgp = gmtime(&t0);		/*          ==0 for gmt */
	itoa( (1+tgp->tm_mon), ct, 13);   /* Jan=1,..., Dec=C */
	cc[1]=ct[0];
	itoa(tgp->tm_mday, ct, 10);
	if (tgp->tm_mday < 10)
	    {
	    cc[2]='0';
	    cc[3]=ct[0];
	    }
	else
	    {
	    cc[2]=ct[0];
	    cc[3]=ct[1];
	    }
	itoa( (100*tgp->tm_hour + tgp->tm_min), ct, 10);
	if (tgp->tm_hour < 10)
	    {
	    cc[4]='0'; cc[5]=ct[0]; cc[6]=ct[1]; cc[7]=ct[2];
	    }
	else
	    {
	    cc[4]=ct[0]; cc[5]=ct[1]; cc[6]=ct[2]; cc[7]=ct[3];
	    }
	cc[8]='.'; cc[9]='L'; cc[10]='T'; cc[11]='D'; cc[12]=0;

	if (NULL==( lt_pave_file=fopen(cc, "wt") ))
	    {
	    do_t_error(43, "E43 LTD File", 200);
	    return;
	    }
	else
	    {
	    lt_pave_center=0.5*(lt_pwrdat[94] + lt_pwrdat[98]);
	    for (i=49; i<144; i++)		/* [96] +47 on either side */
		{
		rel_pwr_db=p2db(lt_pwrdat[i]/lt_pave_center);
		fprintf(lt_pave_file, "%d\t%10.5lf\n", i, rel_pwr_db);
		}
	    if (fclose(lt_pave_file) != NULL)
		do_t_error(43, "E43 LTD File", 200);
	    }
	}

/*
 * draw_grid()  - Assumes that BGI graphics are initialized
 * and draws grid for plotting of SA levels and threshold crossings.
 * Adds annotation for readability.
 * Set up for VGA 640 x 480 pixels ONLY
 */
void draw_grid(void)
	{
	unsigned int x, i;
	int xx, yy;

	new_freq();

	setviewport(0,0,639,479,DO_CLIP);
	setcolor(c_outline_bright);
	line(3,19,101,19); line(3,53,101,53);
	line(3,19,3,53); line(101,19,101,53);
	setcolor(c_text_bright);
	settextstyle(0,0,1);
	settextjustify(1,2);
	moveto(51,23); outtext(call_str);
	moveto(51,32); outtext("DSP TRANSCVR");
	moveto(51,43); outtext("W7PUA DSP10");

	setcolor(c_outline_bright);		/* Freq display border */
	rectangle(1,122,126,158);		/* Leave off corner pixels */
	rectangle(0,123,127,157);

	setcolor(c_text_bold);
	moveto(51,61);	outtext("RECEIVE");
	if (!hardware)
	    {
	    moveto(51, 71);
	    outtext("No RF Hdw");
	    }

	if (input_if_af && eme_doppler)		/* Only for audio processor */
	    {
	    moveto(51, 101);
	    sprintf(cc, "Fd=%dMHz", near_int(0.000001*f_eme) );
	    outtext(cc);
	    }
	for(i=1; i<=9; i++)
		{
		x = X2L + (int)(53.229 * (double)i);
		line(x, Y2T-2, x, Y2T);		/* Vertical ticks */
		}

	line(X2L, Y2T, X2R, Y2T);		/* Lower spectral box */
	line(X2L, Y2B, X2R, Y2B);
	line(X2L, Y2T, X2L, Y2B);
	line(X2R, Y2T, X2R, Y2B);

	/* Now status info along left side */
	settextjustify(1,2);
	settextstyle(0, 0, 1);
	setcolor(c_text);
	moveto(64,S_METER_POS);	outtext("Signal Level");

	settextjustify(0,2);
	settextstyle(0,0,1);
	setcolor(c_text);
	line(1,229,127,229);
	setcolor(c_text_bold);
	moveto(2,233);	outtext("Contrast");
	moveto(2,243);	outtext("Bright");
	setcolor(c_text);
	line(1,255,127,255);
	setcolor(c_text_bold);
	moveto(2,258);  outtext("MikeGain");
	moveto(2,268);  outtext("CW Speed");
	moveto(2,278);  outtext("Xmit Pwr");
	moveto(2,288);	outtext("Mode");
	setcolor(c_text);
	line(1,299,127,299);
	moveto(2,303);	outtext("RIT");
	setcolor(c_text_bold);
	moveto(2,313);	outtext("Filter");
	moveto(2,323);  outtext("DeNoise");
	moveto(2,333);  outtext("Binaural");

/*	setcolor(c_text);
	moveto(2,333);  outtext("ANL"); */
/*	moveto(2,243);  outtext(""); */		/* Empty line */

	setcolor(c_text);
	line(1,354,127,354);
/*	moveto(2,358);  outtext("BinAve"); */   /* Not used now */
	setcolor(c_text_bold);
	moveto(2,368);	outtext("SpecAve");
	moveto(2,378);	outtext("SpecAnl");

	setcolor(c_text);
	line(1,389,127,389);
	setcolor(c_text_bold);
	moveto(2,393);	outtext("AF Gain");
	moveto(2,403);	outtext("RF Gain");
	moveto(2,413);	outtext("AGC");
	moveto(2,423);  outtext("AutDisp");

	setcolor(c_text);
	line(1,434,127,434);

	setcolor(c_text_bold);
	setviewport(1,1,639,479,DO_CLIP);
	settextstyle(0,0,1);
	settextjustify(2,2);
	moveto(126, 233);
	sprintf(cc, "%d", contrast);
	outtext(cc);
	moveto(126, 243);
	sprintf(cc, "%d", bright);
	outtext(cc);
	moveto(126, 258);
	sprintf(cc, "%d", mike_gain);
	outtext(cc);
	moveto(126, 268);
	outtext(t[cw_speed].description);
	moveto(126, 278);
	sprintf(cc, "%d", xmit_power);
	outtext(cc);
/*	setcolor(c_text);
	moveto(126, 358);
	sprintf(cc, "%d", 1); */		/* BinAve - not used now */
	outtext(cc);
	setcolor(c_text_bold);
/*	moveto(126, 368);
	sprintf(cc, "%d", noncohave);
	outtext(cc);
 */
	set_8_16(9);

	moveto(126, 413);
	sprintf(cc, "%d", rfg[rfgain].dbdisplay);
	outtext(cc);
/*	setcolor(c_text);
	moveto(126, 423);
	sprintf(cc, "%d", 100);
	outtext(cc);
	setcolor(c_text_bold);
*/
	setviewport(0, S_METER_POS-11, 127, S_METER_POS-5, DO_CLIP);
	/* bar_graph()  bars are every 6 pixels, starting at 20 */
	micro_char(1, 21, 6); micro_char(5, 44, 6); micro_char(9, 68, 6);
	micro_char(2, 79, 6); micro_char(0, 83, 6);
	micro_char(4, 91, 6); micro_char(0, 95, 6);
	micro_char(6, 103, 6); micro_char(0, 107, 6);

	if(!waterfall)   draw_panel();
	sa_graph();			/* Put spectrum analyzer graph up */
	set_db_div(1);			/* Add numbers on left */

	normalize_plot(0);		/* Clear normalization */
	}

/* Draw the upper right portion of the screen for the spectrum analyzer
 * display.  Separated so that this same area can be used for a modal
 * dialog box.  Box outline  and markings only.  Grid is drawn during
 * the plotting of the data.
 */
void sa_graph(void)
	{
	int i, yy, kdb;

	setviewport(X1L-12, Y1T, X1R, Y1B, DO_CLIP);	
	setcolor(c_outline);
	line(12,         0,       12+X1R-X1L, 0);
	line(12,         Y1B-Y1T, 12+X1R-X1L, Y1B-Y1T);
	line(12,         0,       12,         Y1B-Y1T);
	line(12+X1R-X1L, 0,       12+X1R-X1L, Y1B-Y1T);

	settextstyle(0, 0, 1);
	settextjustify(1,1);
	/* db_div_set:  0=10db/div, 1=5db, 2=2db, 3=1db */
	if (db_div_set==0) kdb=10;
	else if (db_div_set==1) kdb=5;
	else if (db_div_set==2) kdb=2;
	else kdb=1;
	for (i=0; i<=8; i++)
		{
		yy = i*16;
		moveto(0, yy);
		sprintf(cc, "%d", kdb*(8 -i));
		outtext(cc);		/* dB level markings */
		}
	}

void mark_frequency(int fmax)
	{
	int i, k;

	setcolor(c_text);
	settextstyle(0, 0, 1);
	settextjustify(1, 2);
	setviewport(178, 134, 638, 144, DO_CLIP);
	clearviewport();
	moveto(430, 0);
	if (fmax<10000)
		{
		k=fmax/4;
		outtext("Hz");
		}
	else
		{
		k=fmax/4000;
		outtext("kHz");
		}
	for(i=0; i<4; i++)
		{
		moveto(57+i*107, 0);
		sprintf(cc, "%d", k*(i+1));
		outtext(cc);			/* Freq markings */
		}
	setcolor(c_text_bold);
	}

/* draw_panel() is used for the conventional panel display.  The lower
 * waterfall window is divided into two parts, the lower 2/3 for help
 * info and this frequency display in the top 1/3.  This display is never
 * used with the waterfall display.
 */
void draw_panel(void)
	{
	int i, x;

	/* 128, 639   d=511  */
	setviewport(X2L+1, Y2T, X2R-1, Y2T+95, DO_CLIP);
	clearviewport();
	setcolor(c_outline);
	setlinestyle(SOLID_LINE, 0, THICK_WIDTH);
	rectangle(50, 2, 461, 55);	/* Big Frequency Box */
	setcolor(c_outline_bold);
	line(55, 41, 455, 41);
	setlinestyle(SOLID_LINE, 0, NORM_WIDTH);
	line(0, 0, 46,0);  line(0, 3, 46, 3);	/* Trim lines at top */
	line(465, 0, 510, 0);  line(465, 3, 510, 3);

	for(i=0; i<5; i++)
	    {
	    x = 95+i*80;
	    line(x, 43, x, 32);
	    tiny_char(1,   x-9, 51);   /* Print with bottom left at coords */
	    tiny_char(4,   x-3, 51);
	    tiny_char(i+4, x+3, 51);
	    }
	for(i=0; i<25; i++)
	    line(63+i*16, 43, 63+i*16, 37);

	/* Now put the general help screen at the bottom: */
	setcolor(c_text_bold);
	settextstyle(DEFAULT_FONT, HORIZ_DIR, 1);

	settextjustify(0,2);		/* Left and top */
	/* Works out to be 224 pixels high: */
	setviewport(X2L+1, Y2T+96, X2R-1, Y2B-12, DO_CLIP);
	getimage(0,  0, 510, 160, big_area_ptr1);
	getimage(0, 161, 510, 224, big_area_ptr2);
	clearviewport();
	line(1, 5, 509, 5);		/* Half into "first" line (blank) */
	line(1, 223, 509, 223);
 /* 63 characters:
          123456789012345678901234567890|23456789012345678901234567890123 */
 moveto(3, 3);			/* Line 1-1 */
 /* Blank */
 moveto(3, 13);			/* Line 1-2 */
 outtext("               DSP-10 KEYBOARD COMMANDS - Summary");
 moveto(3, 23);			/* Line 1-3 */
 outtext("       FREQUENCY              |                 GAIN");
 moveto(3, 33);			/* Line 1-4 */
 outtext("1 Hz   Ctrl-Alt-Sh F9 & F10   |   AF Gain               F7 & F8");

 moveto(3, 43);			/* Line 1-5 */
 if(kbd_alt1 && kbd_alt3)
 outtext("10 Hz Down            Sh F9   |   RF Gain          Scrl F7 & F8");
 else if (!kbd_alt1 && kbd_alt3)
 outtext("10 Hz Down               F9   |   RF Gain          Scrl F7 & F8");
 else if (kbd_alt1 && !kbd_alt3)
 outtext("10 Hz Down            Sh F9   |   RF Gain            Sh F7 & F8");
 else
 outtext("10 Hz Down               F9   |   RF Gain            Sh F7 & F8");

 moveto(3, 53);			/* Line 1-6 */
 if (kbd_alt1)
 {
 outtext("10 Hz Up              Sh F10  |   Mike Gain      Alt-Sh F7 & F8");
 moveto(3, 63);			/* Line 1-7 */
 outtext("100 Hz              F9 & F10  |");
 }
 else
 {
 outtext("10 Hz Up                 F10  |   Mike Gain      Alt-Sh F7 & F8");
 moveto(3, 63);			/* Line 1-7 */
 outtext("100 Hz           Sh F9 & F10  |");
 }

 moveto(3, 73);			/* Line 1-8 */
 outtext("1 kHz           Alt F9 & F10  |   CW Send Speed     Alt F7 & F8");
 moveto(3, 83);			/* Line 1-9 */
 outtext("10 kHz         Ctrl F9 & F10  |   Require Scroll Lock On:");
 moveto(3, 93);			/* Line 1-10 */
 outtext("100 kHz    Ctrl-Alt F9 & F10  |   AGC Release        Scrl G & H");
 moveto(3, 103);		/* Line 1-11 */
 outtext("1 MHz       Ctrl-Sh F9 & F10  |   Transmit Power     Scrl O & P");
 moveto(3, 113);		/* Line 1-12 */
 outtext("                              |   CW Sidetone Lev    Scrl S & D");
 moveto(3,123);			/* Line 1-13 */
 outtext("FM Squelch        Alt S & D   |   ");
 moveto(3,133);			/* Line 1-14 */
 outtext("                              |");
 moveto(3,143);			/* Line 1-15 */
 outtext("                              |       TRANSMIT/RECEIVE");
 moveto(3, 153);		/* Line 1-16 */
 outtext("RIT On/Off            Alt O   |   T/R Toggle               Home");
 moveto(3, 163);		/* Line 1-17 */
 outtext("RIT Down/Up 10Hz  Alt R & T   |");
 moveto(3, 173);		/* Line 1-18 */
 outtext("RIT Zero              Alt Z   |          AF FILTERS");
 moveto(3, 183);		/* Line 1-19 */
 outtext("Mode change           Alt M   |   Filters On/Off             F3");
 moveto(3, 193);		/* Line 1-20 */
 outtext("Help Screens             F1   |   Filter Select              F4");

 moveto(3, 203);		/* Line 1-21 */
 if (kbd_alt3)
 outtext("Display Format       Scrl Y   |   DeNoise & Notch On/Off  Sh F3");
 else
 outtext("Display Format        Alt Y   |   DeNoise & Notch On/Off  Sh F3");

 moveto(3, 213);		/* Line 1-22 */
 outtext("Exit to DOS     Scrl-Alt F4   |   De-Noise Gain    Ctrl F3 & F4");
	}

/* set_bins()  -  Sets up the global bin numbers  lt_i1, lt_i2,... used
 * in both EME2 and LTI long term integration.
 */
void set_bins(void)
	{
	int mid;

	/* For EME2 (center freq 323, 600 and 900 Hz):
	 *   FFT  low disp  low detail  center  high detail  high disp
	 *  1200     90        128       138        148         186
	 *  2400     80        118       128        138         176
	 *  4800     48         86        96        106         144
	 */
	if (mode==EME2)
	    {
	    if (fft_bw==0)
		{
		lt_i1=90;
		lt_i2=128;
		lt_i3=138;
		lt_i4=148;
		lt_i5=186;
		}
	    else if (fft_bw==1)
		{
		lt_i1=80;
		lt_i2=118;
		lt_i3=128;
		lt_i4=138;
		lt_i5=176;}
	    else if (fft_bw==2)
		{
		lt_i1=48;
		lt_i2=86;
		lt_i3=96;
		lt_i4=106;
		lt_i5=144;
		}
	    }
	else if (mode==LTI)
	    {
	    lt_i1=(int)(0.5+(lterm_int_flow/df[fft_bw]));
	    if (lt_i1<1 || lt_i1>234) lt_i1=1;

	    lt_i5=(int)(0.5+(lterm_int_fhigh/df[fft_bw]));
	    if (lt_i5>254 || lt_i5<21) lt_i5=254;

	    /* Be sure there is enough room for 21 detailed points */
	    if ((lt_i5-lt_i1) <= 21)
		{
		lt_i1=1;
		lt_i5=254;
		}

	    /* lt_i3 is the desired center. Be  sure it is inside the range */
	    lt_i3=(int)(0.5+(lterm_int_center/df[fft_bw]));
	    if (lt_i3<=(lt_i1+10) || lt_i3>=(lt_i5-10))
		 lt_i3=(lt_i1+lt_i5)/2;

	    lt_i2=lt_i3-10;		/* Set limits for 21 detailed pts */
	    lt_i4=lt_i3+10;
	    }
	}

/* fsk2sec() sends the selected fsk tone for 1.8 seconds. This gives
   enough time for changing tone and getting back before next 2 sec start.
 */
void fsk2sec(void)
	{
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('T');
	ASYNCOUT(0X0011); ASYNCOUT(2);     /* Xmit, no stat, key dwn */ 
	ASYNCOUT(fgainl[xmit_power]); ASYNCOUT(fgainh[xmit_power]);
	delay(450);
	update_clock(0);
	delay(450);
	update_clock(0);
	delay(450);
	update_clock(0);
	delay(450);
	update_clock(0);
	ASYNCOUT(ESC_CHAR);  ASYNCOUT('T');
	ASYNCOUT(0X0001); ASYNCOUT(2);	/* Npw key up */
	ASYNCOUT(fgainl[xmit_power]); ASYNCOUT(fgainh[xmit_power]);
	}

/* update_clock(int fcn)  -  can be called often to
 * add time to the 'millisec' clock. 
 * If (fcn==0) the return value is 0 unless an
 * integral second was passed by the update.
 * If (fcn==2) there is no 2.6 sec EME offset. This allows clock set during
 * LHL7 and PUA43.
 * The time returned is in seconds, suitable for localtime() etc, after
 * adding dsp_clock_base.
 */
long int  update_clock(int fcn)
	{
	long int          cl_diff, milli_new, m;
	static long int   cl_old=0;
	static long int   m_last=0;

	cl_ms=clock();			/* Returns milliseconds */
	cl_diff=cl_ms-cl_old;		/* Integer millisecs */
        cl_old=cl_ms;			/* To keep accumulating millisec */
	millisec += (clock_cor_factor*(double)cl_diff); /* Update dsp clock */
	/* millisec_offset, a double, moves the tick times around in a
	   second and also offsets the time by a few seconds to bring
	   computer set time to correct time */
	milli_new=(long int)(millisec+millisec_offset);

	/* Offset by a eme delay, if appropriate */
	if ((!transmit) && (fcn!=2) && eme_doppler
		 && (mode==LHL7 || mode==PUA43))
	    milli_new -= 2600;

	m=milli_new/1000L;		/* Whole seconds in milli_new */
        if ( m > m_last || fcn==1 )
	   {m_last=m;
	    return m;
	   }
	else
	   {
	    return 0L;
	   }
	}

/* set_clock(cmd)  -  Uses line at screen bottom to set/change the internal
   dsp-10 clock.
   cmd=0  ==> initialization of clock.
   cmd=1  ==> changing of clock.
   This function is modal and cannot be left running or the other functions
   will stop.
   Note: clock() starts fresh with startup, but will rollover after 594 hours
   with errors at that point. Work on that later! [Maybe update_clock()<0 ? ]
 */
void set_clock(int cmd)
	{
	char  c3[50];
	double         dsp_clock;
	unsigned int   ichrr;
	long int       t_new, ttt;
	struct tm      *tp;

	if (ms_available==0) return;

	/* Set dsp10 clock to computer clock.
           Then for dsp10 use millisec + a base offset to convert
           to a long int that is display time.
	 */
	if (cmd==0)		/* Initialization stuff, done at startup:   */
	   {dsp_clock_base=time(NULL);
	    millisec=0.0;
	    return;
	   }
	else if (cmd==1)	/* Setclock */
	   {
	    clock_been_set=TRUE;
	    setviewport(129, 390, 638, 469, DO_CLIP); /* Clear text space */
	    clearviewport();
	    settextjustify(LEFT_TEXT,TOP_TEXT);
	    moveto(20, 1);
	    outtext("------------------- DSP-10 CLOCK SET ------------------");
	    moveto(20, 11);
	    outtext(" Back  1 sec     (       |      Forward  1 sec      )");
	    moveto(20, 21);
	    outtext(" Back 10 mSec    {       |      Forward 10 mSec     }");
	    moveto(20, 31);
	    outtext(" Back  1 mSec    [       |      Forward  1 mSec     ]");
	    moveto(20, 41);
	    outtext(" Rate slower     <       |      Rate faster         >");
	    moveto(20, 61);
	    outtext(" Exit by *    Reenter by ALT K     PC clock unchanged.");
	    setviewport(129, 470, 638, 479, DO_CLIP); /* Clear main line */
	    clearviewport();
	    settextjustify(0,2);
	    while (1)
		{if ( (t_new=update_clock(2)) > 0L)
        	     {if (lpt_port==0)
                          beep(tick_time);	/* Do this once per second */
                      else
                          {outp(lpt_port, 1);
                           delayu(ntau);
                           outp(lpt_port, 0);
                          }
                      ttt=dsp_clock_base+t_new;
		      if (time_type)
			tp=localtime(&ttt);	/* time_type==1 for local */
		      else
			tp=gmtime(&ttt);	/*          ==0 for gmt */
                      strcpy(c3, asctime(tp));	/* Rearrange the characters */
                      c3[19]=0;
                      sprintf(cc,
			  "DSP-10 time=%s   Clock rate=%.7lf   Offset=%.3lf",
			  &c3[11], clock_cor_factor, 0.001*millisec_offset);
		      clearviewport();
                      moveto(5, 0); outtext(cc);
                     }
                 if (kbhit() != 0)
                     {ichrr = 0X7F & _bios_keybrd(_KEYBRD_READ);
                      if      (ichrr == '{') millisec_offset-=10.0;
                      else if (ichrr == '}') millisec_offset+=10.0;
                      else if (ichrr == '[') millisec_offset-=1.0;
                      else if (ichrr == ']') millisec_offset+=1.0;
                      else if (ichrr == '(') millisec_offset-=1000.0;
                      else if (ichrr == ')') millisec_offset+=1000.0;
                      else if (ichrr == '<') clock_cor_factor -= 0.0000001;
                      else if (ichrr == '>') clock_cor_factor += 0.0000001;
                      else if (ichrr == '*')
			  {setviewport(129, 390, 638, 479, DO_CLIP);
			   clearviewport();
			   return;
			  }
                      /* Ignore everything else */
                     }
              }				/* End while(1) */
	   }				/* End cmd==1 */
	}				/* End fcn */

void point1(int x, int y)	/* Cross */
	{
	line(x, y-1, x, y+1);
	line(x-1, y, x+1, y);
	}

void point5(int x, int y)	/* Block rectangle */
	{
	line(x-1, y-1, x+2, y-1);
	line(x-1, y-1, x-1, y+1);
	line(x+2, y-1, x+2, y+1);
	line(x-1, y+1, x+2, y+1);
	}

/* tiny_time( ) puts the time along the left side of the spectral map.
 * The area 9 high and 32 wide should be cleared.
 * If show_secs==TRUE there is a colon between the first two and the
 * last two digits.
 */
void tiny_time(int hhmm, int x0, int y0)
	{
	int h1, h2, m1, m2;
	h1=hhmm/1000;
	hhmm-=h1*1000;
	h2=hhmm/100;
	hhmm-=h2*100;
	m1=hhmm/10;
	m2=hhmm-m1*10;
	tiny_char(h1, x0, y0);
	tiny_char(h2, x0+6, y0);
	if (show_secs)
	    {
	    putpixel(x0+13, y0-1, c_text_bold);
	    putpixel(x0+13, y0-4, c_text_bold);
	    tiny_char(m1, x0+15, y0);
	    tiny_char(m2, x0+21, y0);
	    line(x0+27, y0, x0+33, y0);	/* Tick for actual time */
	    }
	else
	    {
	    tiny_char(m1, x0+12, y0);
	    tiny_char(m2, x0+18, y0);
	    line(x0+24, y0, x0+30, y0);	/* Tick for actual time */
	    }
	}

/* Print n with bottom left at (x0, y0) */
void tiny_char(int n, int x0, int y0)
	{
	if (n<0 || n>9) return;
	switch (n)
	  {
	  case 0:{
		line(x0, y0, x0+4, y0);
		line(x0, y0-6, x0+4, y0-6);
		line(x0, y0, x0, y0-6);
		line(x0+4, y0, x0+4, y0-6);
		break;
		}
	  case 1:{
		line(x0+2, y0, x0+2, y0-6);
		break;
		}
	  case 2:{
		line(x0, y0, x0+4, y0);
		line(x0, y0-3, x0+4, y0-3);
		line(x0, y0-6, x0+4, y0-6);
		line(x0, y0, x0, y0-3);
		line(x0+4, y0-3, x0+4, y0-6);
		break;
		}
	  case 3:{
		line(x0, y0, x0+4, y0);
		line(x0, y0-3, x0+4, y0-3);
		line(x0, y0-6, x0+4, y0-6);
		line(x0+4, y0, x0+4, y0-6);
		break;
		}
	  case 4:{
		line(x0, y0-3, x0+4, y0-3);
		line(x0, y0-3, x0, y0-6);
		line(x0+4, y0, x0+4, y0-6);
		break;
		}
	  case 5:{
		line(x0, y0, x0+4, y0);
		line(x0, y0-3, x0+4, y0-3);
		line(x0, y0-6, x0+4, y0-6);
		line(x0+4, y0, x0+4, y0-3);
		line(x0, y0-3, x0, y0-6);
		break;
		}
	  case 6:{
		line(x0, y0, x0+4, y0);
		line(x0, y0-3, x0+4, y0-3);
		line(x0, y0, x0, y0-6);
		line(x0+4, y0, x0+4, y0-3);
		break;
		}
	  case 7:{
		line(x0, y0-6, x0+4, y0-6);
		line(x0+4, y0, x0+4, y0-6);
		break;
		}
	  case 8:{
		line(x0, y0, x0+4, y0);
		line(x0, y0-3, x0+4, y0-3);
		line(x0, y0-6, x0+4, y0-6);
		line(x0, y0, x0, y0-6);
		line(x0+4, y0, x0+4, y0-6);
		break;
		}
	  case 9:{
		line(x0, y0-3, x0, y0-6);
		line(x0, y0-3, x0+4, y0-3);
		line(x0, y0-6, x0+4, y0-6);
		line(x0+4, y0, x0+4, y0-6);
		break;
		}
	  }
	}

/* Print smaller n with bottom left at (x0, y0) */
void micro_char(int n, int x0, int y0)
	{
	if (n<0 || n>9) return;
	switch (n)
	  {
	  case 0:{
		line(x0, y0, x0+2, y0);
		line(x0, y0-4, x0+2, y0-4);
		line(x0, y0, x0, y0-4);
		line(x0+2, y0, x0+2, y0-4);
		break;
		}
	  case 1:{
		line(x0+1, y0, x0+1, y0-4);
		break;
		}
	  case 2:{
		line(x0, y0, x0+2, y0);
		line(x0, y0-1, x0+2, y0-3);
		line(x0, y0-4, x0+2, y0-4);
		break;
		}
	  case 3:{
		line(x0, y0, x0+2, y0);
		line(x0, y0-2, x0+2, y0-2);
		line(x0, y0-4, x0+2, y0-4);
		line(x0+2, y0, x0+2, y0-4);
		break;
		}
	  case 4:{
		line(x0, y0-2, x0+2, y0-2);
		line(x0, y0-2, x0, y0-4);
		line(x0+2, y0, x0+2, y0-4);
		break;
		}
	  case 5:{
		line(x0, y0, x0+2, y0);
		line(x0, y0-2, x0+2, y0-2);
		line(x0, y0-4, x0+2, y0-4);
		line(x0+2, y0-1, x0+2, y0-2);
		line(x0, y0-2, x0, y0-4);
		break;
		}
	  case 6:{
		line(x0, y0, x0+2, y0);
		line(x0, y0-2, x0+2, y0-2);
		line(x0, y0, x0, y0-2);
		line(x0, y0-2, x0+1, y0-4);
		line(x0+2, y0, x0+2, y0-2);
		break;
		}
	  case 7:{
		line(x0, y0-4, x0+2, y0-4);
		line(x0+1, y0, x0+2, y0-4);
		break;
		}
	  case 8:{
		line(x0, y0, x0+2, y0);
		line(x0, y0-2, x0+2, y0-2);
		line(x0, y0-4, x0+2, y0-4);
		line(x0, y0, x0, y0-4);
		line(x0+2, y0, x0+2, y0-4);
		break;
		}
	  case 9:{
		line(x0, y0-2, x0, y0-4);
		line(x0, y0-2, x0+2, y0-2);
		line(x0, y0-4, x0+2, y0-4);
		line(x0+2, y0, x0+2, y0-4);
		break;
		}
	  }
	}

/* Draw bar graph, starting from right bar. If value smaller than test
 * value, just draw an empty bar. When smaller, continue to fill in all bars.
 */
void bar_graph(int x0, int y0, unsigned int value)
	{
	int i, color_on;
	unsigned int test_value;
	setfillstyle(SOLID_FILL, c_bar_dark);   /* Unlit bar color */
	test_value = 0X4000;
	for(i=14; i>=0; i--)			/* Start from right */
		{
		color_on = ( value >= test_value );
		test_value /= 2;
		if (color_on)
			setfillstyle(SOLID_FILL, bar_color[i]);
		bar(x0+i*6, y0, x0+i*6+4, y0+8);
		}
	}

void beep(int tau)
	{sound(600);
	delay(tau);
	nosound();
	}

/* Diagnostics usually uses the bottom line on the display to display info
 * about the inner workings of the transceiver.
 * Apr99 Added Moon/Sun track as 'diagnostic' option.
 * The data depends on the value of diag_mode:
 *	0  No diagnostic info is shown.
 *	1  The 20 bytes of status_buf data are printed in hex.
 *	2  Two bar graphs of A/D lvel and D/A level are shown.
 *	3  Moon data
 *	4  Sun data
 */
void diagnostics(int do_what)
	{
	int i;

	if (do_what==0)			/* do_what==0 for toggle diag_mode */
		{
		diag_mode++;

		if ((diag_mode) > 4)  diag_mode=0;

		setviewport(129, 470, 639, 479, DO_CLIP); /* Clear main line */
		clearviewport();
		settextjustify(0,2);
		moveto(10, 0);
		sprintf(cc, "%.1d", diag_mode);
		outtext(cc);

		setviewport(0, 470, 40, 479, DO_CLIP); /* Clear left line */
		clearviewport();

		if (diag_mode==3 || diag_mode==4)
		    set_cur_loc(-1);		/* Update screen at bottom */
		set_eme_doppler();		/* Update eme mark */
		return;
		}
	/* do_what==1 for display according to diag_mode */
	if (diag_mode>0)
	  {
	  setviewport(0, 470, 40, 479, DO_CLIP);
	  clearviewport();
	  settextstyle(0, 0, 1);
	  settextjustify(LEFT_TEXT, TOP_TEXT);
	  moveto(0, 0);
	  sprintf(cc, "%d", ASYNCINSTAT() );  /* Always put loading info out */
	  outtext(cc);
#if 0
	  settextjustify(CENTER_TEXT, TOP_TEXT);
	  moveto(64, 0);
	  sprintf(cc, "%d", dsp_status[15]);  /* and the fft scaling factor */
	  outtext(cc);
	  settextjustify(LEFT_TEXT, TOP_TEXT);
#endif
	  }
	if (diag_mode==0)
	  {
	  setviewport(0, 470, 40, 479, DO_CLIP); /* Clear left corner */
	  clearviewport();
	  setviewport(129, 470, 639, 479, DO_CLIP); /* Clear main line */
	  clearviewport();

	  return;
	  }
	else if (diag_mode==1)
	    {setviewport(129, 470, 579, 479, DO_CLIP); /* Clear line */
	     clearviewport();
	     settextjustify(0,2);
	     for(i=0; i<18; i++)
		{
		moveto(10+i*20, 0);
		sprintf(cc, "%2.2X", dsp_status[i]);
		outtext(cc);
		}
	    moveto(373, 0);
	    if (transmit)
		outtext("T");
	    else
		outtext("R");
	    moveto(385, 0);
	    outtext("Ver2.01");

	    /* This is disabled because there is no data here to show:
	    if(transmit)
		{
		setviewport(150, 30, 450, 50, DO_CLIP);
		clearviewport();
		settextstyle(0, 0, 1);
		settextjustify(0, 2);
		for(i=0; i<10; i++)
		    {
		    moveto(10+i*20, 0);
		    sprintf(cc, "%2.2X", dsp_xmit_status[i]);
		    outtext(cc);
		    if (4*(i/4)==i)  outtext(" ");
		    }
		}
	    */

	    if ((!transmit) && auto_disp_on)
		{moveto(400, 0);
		 if (ave_spectral_pwr<1e-12)
			ave_spectral_pwr=1e-12;
		 sprintf(cc, "%7.2lf", p2db(ave_spectral_pwr));
		 outtext(cc);
		}
	    }
	else if (diag_mode==2)
	    {
	    setviewport(129, 470, 579, 479, DO_CLIP); /* Clear line */
	    clearviewport();

	    moveto(10, 0);
	    outtext("A/D Level:");
	    bar_graph( 100, 0, 256*dsp_status[10]+dsp_status[9] );

	    moveto(220, 0);
	    outtext("D/A Level:");
	    bar_graph( 300, 0, 256*dsp_status[12]+dsp_status[11] );

	    if ((!transmit) && auto_disp_on)
		{moveto(400,0);
		 if (ave_spectral_pwr<1e-12)
			ave_spectral_pwr=1e-12;
		 sprintf(cc, "%7.2lf", p2db(ave_spectral_pwr));
		 outtext(cc);
		}
	    }
	}

/* Preliminaries and tables - this is preliminary, but it must follow
 * get_mem() so that arrays to be initialized are in existance.
 */
void initialize(void)
	{
	int i, ifg;
	double fg;
	long  a, b, fr;
	unsigned int xu;
	long int li1, li2;

	ASYNCINIT(com_port);
	ASYNCSET(BAUD_RATE, COM_SETTINGS);
	ASYNCCLEAR();

	srand(1);			/* Seed random num gen */
	set_clock(0);			/* Synchronize to PC clock */

	/* Following vector converts (0,255), representing 0 to
	    80 dB into a power ratio (1.0, 10^8).  Use vector for speed */
	for (i=0; i<256; i++)		/* Convert log to power */
		power10[i]=pow(10.0, 0.03125*(double)i );

	/* Gain function in 1 dB steps for use in dsp */
	for(i=1; i<=100; i++)		/* Setup gain fcn as 2 bytes */
		{fg = pow(10.0, (double)(i-100) / 20. );
		 fgain[i] = fg;
		 ifg = near_int(32767.0*fg);
		 fgainl[i] = ifg & 0XFF;
		 fgainh[i] = (ifg >> 8) & 0XFF;
		}

	for(i=0; i<515; i++)  raw_data[i]=0;
	for(i=0; i<256; i++)  pd16[i]=0;
	for(i=0; i<515; i++)  dnorm[i]=0.0;
	for(i=0; i<256; i++)  lt_dnorm[i]=0.0;
	for(i=0; i<21;  i++)  eme_pwr_sum[i]=0.0;

/*	do_warm_boot(); */    	/* Fully set DSP & screen to current state */

	/* Setup reference divider in synthesizer */
	send_shift_data(R10MHZ);	/* R1MHZ, R5MHZ, R6_4MHZ or R10MHZ */
	delay(250);

	if (!waterfall)
	    {
	    draw_panel();
	    setviewport(0,124,127,161,DO_CLIP);	/* Small Freq Box */
	    clearviewport();
	    }

	if(input_if_af==0) new_band(band);

	setviewport(0,60,101,70,DO_CLIP);
	clearviewport();
	settextstyle(0, 0, 1);
	settextjustify(1,2);
	setcolor(c_text_bold);
	moveto(50,1);
	if (input_if_af==0)
		outtext("RECEIVE");
	else
		outtext("AUDIO");
	print_big(0, -1, 0);		/* Initialize scroll printing */
	print_medium(0, -1);
	print_medium_1(0, -1);
	print_medium_2(0, -1);
	lhl7_update(0, -1);		/* Init statics */
	pua43_update(0, -1);
	morse(INIT_CW, 0);
	set_contrast();
	set_brightness();
	contrast_line_count=0;		/* Don't show "Contrast Line" */
	do_auto_disp(0);		/* Start with auto display off */
	set_disp_offset(1);
	set_window();
	set_mode(0);

	set_afgain(0);
	set_rfgain();
	set_agc(agc_decay);
	set_8_16(9);			/* 9==> set to current value */
	set_affilt_sw();
	set_affilt_select();
	set_denoise();
	set_delay_right(delay_right);
	set_display_width();
	set_eme_doppler();

	new_freq();			/* Initialize freq */

	/* Following is a kludge to make  sure that hardware is properly
	 * properly initialized.  Someday find out why needed!!
	 */
	tfreqL -= 1000000;
	new_freq();
	delay(10);
	tfreqL += 1000000;
	new_freq();

	/* Calibrate ntau delay to 20 useconds */
	li1=biostime(0,0);
	for (li2=0L; li2<1000000; li2++)  delayu(20);
	li1=biostime(0,0) - li1;
	ntau=(unsigned int)(20*(1+(int)( 503.0 / (double)li1 )));
	}

/* Print frequency data in a text box. 20 characters wide, 14 lines
 * high.
 * 
 */
void freq_data(int what, long int synth_fL, long int n_synL,
       long int fL,  long int if1L,  long int if2L, unsigned int dsp_dp)
    {
    static char cb[350];

	if (what==0)		/* Close FREQ, but no reset */
	   {
	   lb_close(&lb_freq, 0);
	   current_info_box=0;	/* Mark as none */
	   return;
	   }
	else if (what==1)		/* Open FREQ */
	   {
	   if (current_info_box != &lb_freq) /* If not open */
		{
		/* Only one at a time:     */
		if(current_info_box)  lb_close(current_info_box, 0);
		lb_create(&lb_freq, "INTERNAL FREQS",
		     474, 148, 638, 300, 0, NON_MODAL);
	    	current_info_box=&lb_freq;

		lb_open(&lb_freq, 0);

		/* Fixed text in FREQ box:     */
		lb_line(&lb_freq, 13, TXT, "  Close Box=Alt A", 0);
		}
	    lb_update(&lb_freq);	/* Draw the empty box */
	   }
	else if (what==2)
	   {
	    lb_open(&lb_freq, 0);
	    sprintf(cb+50,  " Carrier= %9ld", fL);
	    lb_line(&lb_freq, 2, BOLD, cb+50, 0);
	    sprintf(cb+75,  " LO#1   = %9ld", synth_fL);
	    lb_line(&lb_freq, 3, BOLD, cb+75, 0);
	    sprintf(cb+100, " LO#2   = %9ld", 19680000L);
	    lb_line(&lb_freq, 4, BOLD, cb+100, 0);
	    sprintf(cb+125, " IF#1   = %9ld", if1L);
	    lb_line(&lb_freq, 5, BOLD, cb+125, 0);
	    sprintf(cb+150, " IF#2 (DSP)  %6ld", if2L);
	    lb_line(&lb_freq, 6, BOLD, cb+150, 0);

	    sprintf(cb+175, " N-Synth     %6ld", n_synL);
	    lb_line(&lb_freq, 7, BOLD, cb+175, 0);

	    if (mode==CW)
		sprintf(cb+200, " CW Offset    %5d", cwoffset);
	    else if (mode==LTI || mode==LHL7 || mode==PUA43)
		sprintf(cb+200, " FSK          %5d", fsk);
	    else
		sprintf(cb+200, " No CW Offset or FSK");
	    lb_line(&lb_freq, 8, BOLD, cb+200, 0);

	    if (rit_on)
		sprintf(cb+225, " RIT          %5d", rit);
	    else
		sprintf(cb+225, " No RIT");
	    lb_line(&lb_freq, 9, BOLD, cb+225, 0);

	    sprintf(cb+250, " DSP, HL: 0X%.2X 0X%.2X",
		 (0XFF & (dsp_dp >> 8)), (0XFF & dsp_dp) );
	    lb_line(&lb_freq, 10, BOLD, cb+250, 0);

	    lb_update(&lb_freq);
	   }
	else return;			/* Safety on what */
    }

/* The following preset values are entered for now, but may be overwritten
 * by the .CFG file if the associated entries are present.
 * The tvtr[] structure provides one preset for each of the 16 VHF/UHF/uWave
 * bands plus 5 more 'Memory' channels that can be used on any band by the
 * UHFA.CFG file.  These are invoked by Ctrl-Alt x where x is either a band
 * from A to P or a memory channel 1 to 5.
 */
void presets(void)
	{
	int i;

	strcpy(tvtr[0].name, "50 MHz");
	tvtr[0].lo_type=0;
	tvtr[0].sb=-1;
	tvtr[0].lo_freq=-94000000.0;		/* - sign for upconversion */
	tvtr[0].chain_mult=1;
	strcpy(tvtr[0].init_str, "  ");
	tvtr[0].gain_db=20.0;
	tvtr[0].control=0;
	tvtr[0].bus_adr=0X00;
	tvtr[0].last_tfreq=144125000.0;
	tvtr[0].last_mode=USB;

	/* 2 Meters is a dummy transverter with LO frequency of 0.0 Hz */
	strcpy(tvtr[1].name, "144 MHz");
	tvtr[1].lo_type=0;
	tvtr[1].sb=-1;
	tvtr[1].lo_freq=0.0;
	tvtr[1].chain_mult=1;
	strcpy(tvtr[1].init_str, "  ");
	tvtr[1].gain_db=0.0;
	tvtr[1].control=0;
	tvtr[1].bus_adr=0X10;
	tvtr[1].last_tfreq=144200000.0;
	tvtr[1].last_mode=USB;

	for (i=2; i<9; i++)
	    tvtr[i].lo_type=999;		/* Transverter not installed */

	strcpy(tvtr[9].name, "10 GHz");
	tvtr[9].lo_type=0;
	tvtr[9].sb=-1;
	tvtr[9].lo_freq=10224000000.0;
	tvtr[9].chain_mult=1;
	strcpy(tvtr[9].init_str, "  ");
	tvtr[9].gain_db=28.0;
	tvtr[9].control=0;
	tvtr[9].bus_adr=0X20;
	tvtr[9].last_tfreq=144100000.0;
	tvtr[9].last_mode=USB;

	for (i=10; i<15; i++)
	    tvtr[i].lo_type=999;		/* Transverter not installed */

	strcpy(tvtr[15].name, "15 kHz Input");
	tvtr[15].lo_type=0;
	tvtr[15].sb=-1;
	tvtr[15].lo_freq=-143085000.0;
	tvtr[15].chain_mult=1;
	strcpy(tvtr[15].init_str, "  ");
	tvtr[15].gain_db=17.0;
	tvtr[15].control=0;
	tvtr[15].bus_adr=0X10;
	tvtr[15].last_tfreq=144000000.0;
	tvtr[15].last_mode=USB;

	strcpy(tvtr[16].name, "144MHz VFO 1");
	tvtr[16].lo_type=0;
	tvtr[16].sb=-1;
	tvtr[16].lo_freq=0.0;
	tvtr[16].chain_mult=1;
	strcpy(tvtr[16].init_str, "  ");
	tvtr[16].gain_db=0.0;
	tvtr[16].control=0;
	tvtr[16].bus_adr=0X10;
	tvtr[16].last_tfreq=144100000.0;
	tvtr[16].last_mode=CW;

	strcpy(tvtr[17].name, "144MHz VFO 2");
	tvtr[17].lo_type=0;
	tvtr[17].sb=-1;
	tvtr[17].lo_freq=0.0;
	tvtr[17].chain_mult=1;
	strcpy(tvtr[17].init_str, "  ");
	tvtr[17].gain_db=0.0;
	tvtr[17].control=0;
	tvtr[17].bus_adr=0X10;
	tvtr[17].last_tfreq=144200000.0;
	tvtr[17].last_mode=USB;

	strcpy(tvtr[18].name, "144MHz VFO 3");
	tvtr[18].lo_type=0;
	tvtr[18].sb=-1;
	tvtr[18].lo_freq=0.0;
	tvtr[18].chain_mult=1;
	strcpy(tvtr[18].init_str, "  ");
	tvtr[18].gain_db=0.0;
	tvtr[18].control=0;
	tvtr[18].bus_adr=0X10;
	tvtr[18].last_tfreq=144250000.0;
	tvtr[18].last_mode=USB;

	strcpy(tvtr[19].name, "144MHz VFO 4");
	tvtr[19].lo_type=0;
	tvtr[19].sb=-1;
	tvtr[19].lo_freq=0.0;
	tvtr[19].chain_mult=1;
	strcpy(tvtr[19].init_str, "  ");
	tvtr[19].gain_db=0.0;
	tvtr[19].control=0;
	tvtr[19].bus_adr=0X10;
	tvtr[19].last_tfreq=146580000.0;
	tvtr[19].last_mode=FM;

	strcpy(tvtr[20].name, "144MHz VFO 5");
	tvtr[20].lo_type=0;
	tvtr[20].sb=-1;
	tvtr[20].lo_freq=0.0;
	tvtr[20].chain_mult=1;
	strcpy(tvtr[20].init_str, "  ");
	tvtr[20].gain_db=0.0;
	tvtr[20].control=-600;
	tvtr[20].bus_adr=0X10;
	tvtr[20].last_tfreq=146940000.0;
	tvtr[20].last_mode=FM;

	/* Initialize the locations to non-existant except 0 which is at
	 * lat=0 and lon=0.  Non-existant is time_zone=999.
	 */
	for (i=0; i<=9; i++)
	    {
	    loc[i].lat=0.0;
	    loc[i].lon=0.0;
	    loc[i].ht=0.0;
	    loc[i].time_zone=999;
	    loc[i].dst=0;
	    strcpy(loc[i].nm, "");
	    strcpy(loc[i].name, "");
	    }
	loc[0].time_zone=0;
	strcpy(loc[0].nm, "(0,0)");
	strcpy(loc[0].name, "Lat=0 Lon=0");

	/* Initialize filters */
	strcpy(filt[0].name, "MTCH12");
	strcpy(filt[0].file_name, " ");
	filt[0].fil_type=3;
	filt[0].param1=0.0;
	filt[0].fclow=550;
	filt[0].fchigh=650;
	filt[0].fil_gain=6;

	strcpy(filt[1].name, "6CW200");
	strcpy(filt[1].file_name, " ");
	filt[1].fil_type=1;
	filt[1].param1=40.0;
	filt[1].fclow=500;
	filt[1].fchigh=700;
	filt[1].fil_gain=3;

	strcpy(filt[2].name, "6CW300");
	strcpy(filt[2].file_name, " ");
	filt[2].fil_type=1;
	filt[2].param1=70.0;
	filt[2].fclow=450;
	filt[2].fchigh=750;
	filt[2].fil_gain=3;

	strcpy(filt[3].name, "6CW450");
	strcpy(filt[3].file_name, " ");
	filt[3].fil_type=1;
	filt[3].param1=70.0;
	filt[3].fclow=375;
	filt[3].fchigh=825;
	filt[3].fil_gain=3;

	strcpy(filt[4].name, "7CW600");
	strcpy(filt[4].file_name, " ");
	filt[4].fil_type=1;
	filt[4].param1=70.0;
	filt[4].fclow=400;
	filt[4].fchigh=1000;
	filt[4].fil_gain=2;

	strcpy(filt[5].name, "SSB-N");
	strcpy(filt[5].file_name, " ");
	filt[5].fil_type=1;
	filt[5].param1=70.0;
	filt[5].fclow=250;
	filt[5].fchigh=2500;
	filt[5].fil_gain=0;

	strcpy(filt[6].name, "SSB-M");
	strcpy(filt[6].file_name, " ");
	filt[6].fil_type=1;
	filt[6].param1=70.0;
	filt[6].fclow=200;
	filt[6].fchigh=2900;
	filt[6].fil_gain=0;

	strcpy(filt[7].name, "8CW300");
	strcpy(filt[7].file_name, " ");
	filt[7].fil_type=1;
	filt[7].param1=60.0;
	filt[7].fclow=650;
	filt[7].fchigh=950;
	filt[7].fil_gain=3;
	}

void print_time(void)
	{int i;
	 time_t t0;
	 struct tm *tp;

	 if (ms_available)		/* use update_clock() */
	    t0=update_clock(1) + dsp_clock_base;
	 else				/* if not avail, use regular PC time */
	    time(&t0);

	 if (time_type)			/* time_type==1 --> use local time */
	    tp=localtime(&t0);
	 else
	    tp=gmtime(&t0);		/* time_type==0 --> use GMT */

	 /* Un-indent--or else qualify the passage of a second */
	     strcpy(cc, asctime(tp));	/* Rearrange the characters */
	     cc[0]=cc[8]; cc[1]=cc[9]; cc[2]=32;	/* Get Date */
	     cc[3]=cc[4]; cc[4]=cc[5]; cc[5]=cc[6]; cc[6]=32;
	     cc[7]=cc[22]; cc[8]=cc[23]; cc[9]=0;
	     setviewport(0,437,127,457,DO_CLIP);
	     clearviewport();
	     setcolor(c_text_bold);
	     settextstyle(0, 0, 1);
	     settextjustify(1, 2);
	     moveto(63,1);
	     outtext(cc);
	     /* Now create string with 'hhmm:ss XXX' where XXX is GMT, etc. */
	     cc[0]=cc[11];  cc[1]=cc[12];	/* Get hh */
	     for(i=14; i<=18; i++)		/* Get mm:ss */
	         cc[i-12]=cc[i];
	     cc[7]=32;			/* Space */
	     for (i=0; i<=3; i++)
	        cc[i+8]=time_str[i];	/* XXX */
	     cc[12]=0;			/* End of string */
	     moveto(63,11);
	     outtext(cc);

	     /* Now update Moon &/| Sun data, Process if GMT is available */
	     if(ut_available)		/* We can update Moon/Sun data */
	       {
	       cnvut2tdt(&t0, &tdt);
	       if (diag_mode==3)
		  (void)update_moon(&tdt, 1);  /* Update */
	       else if (diag_mode==4)
		  update_sun(&tdt, 0);	/* Update if new minute */
	       }
	}

/* get_dsp_ptrs() - Used at initialization time to fill in a global table of
 * addresses, dsp_ptr[i]. These come from the DSP. At DSP reset the
 * pointer to the addresses is in DSP status_buf bytes 7 (low) and 8 (high).
 */
void get_dsp_ptrs(void)
	{
	int i, j, k;

	dsp_cmd('F', 0X44, 0, 0, 4);	/* Make FFT free running */
	delay(50);
	dsp_cmd('R', 32, 2, 3, 4);	/* Immediate Fresh Start */
	/* Need a delay here, since the serial receiver is not working
	 * until the reset is finished.
	 */
	delay(500);

	ASYNCCLEAR();			/* Clear the serial buffer */

	/* bit 1=1 stops the serial data stream and changes to ESC 'B' */
	bits_8_16 |= 0X0002;
	dsp_cmd(9, bits_8_16, 0, 0, 0);

	/* First, get the address of the ptr table:   */
	if(!get_status_bytes('B', 10.0))	/* Fill in dsp_status */
	    {
	    printf("\nError in obtaining DSP pointer table. Filter design\n");
	    printf("will not be available. Hit Enter to continue\n");
	    table_adr=0;
	    getch();
	    return;
	    }
/*
	printf("status 0 2 7 8 13 17= %.2X %.2X %.2X %.2X %.2X %.2X\n",
		dsp_status[0], dsp_status[2],  dsp_status[7],
		dsp_status[8], dsp_status[13], dsp_status[17]);
 */
	/* Form the address from the 2 bytes: */
	if (dsp_status[17]==2)
	    {
	    table_adr=256*dsp_status[8]+dsp_status[7];
	    printf("\nTable address = %.4X, Getting pointers...\n", table_adr);
	    }
	else
	    {
	    printf("\nError in obtaining DSP table address. Filter design\n");
	    printf("will not be available. Hit Enter to continue\n");
	    table_adr=0;
	    getch();
	    return;
	    }

	/* We now have a pointer to the dsp pointer table. Get them:

	/* Set up the FFT for software trigger (that comes from 5 cmd)
	 * and no data collection to speed things up.
	 */
	fft_fcn &= 0XB5FF;		/* Clear bits 9, 11, 14 */
	fft_fcn |= 0X4200;		/* Software trigger, No FFT data */
	dsp_cmd('F', fft_fcn, fft_fcn>>8, 3, 4);

	/* The process is to request an address, get an address, ...
	 * continuing on 38 times: */
	for(j=0; j<38; j++)
	    {
	    k=-1;
	    do
		{
		dsp_cmd(5, table_adr+j, (table_adr+j)>>8, j+1, 4);
	        delay(20);
		if (!get_status_bytes('B', 1.0))  /* Fill in dsp_status */
		  {
		  if (k>=0)
		    {
		    printf("\nRetry in obtaining DSP pointer table entry.\n");
		    printf("Attempt %d\n", k);
		    }
		  break;
		  }
		dsp_ptr[j]= 256*dsp_status[8]+dsp_status[7];
		}
	    while (dsp_status[13] != (j+1) && (k++)<2);
	    }

	if (dsp_ptr[37] != 0X1234 || dsp_status[18] != 0X56)
	    {
	    printf("\nError in obtaining adr dsp_ptr[31]. Filter design\n");
	    printf("will not be available. Hit Enter to continue\n");
	    table_adr=0;
	    getch();
	    return;
	    }

	bits_8_16 &= 0X0001;		/* Restore FFT data */
	dsp_cmd(9, bits_8_16, 0, 0, 0);

	/* Restore FFT functions to normal free run */
	fft_fcn &= 0XB5FF;		/* Clear bits 9, 11, 14 */
	dsp_cmd('F', fft_fcn, fft_fcn>>8, 3, 4);
	}

/* get_status_bytes()  -  Reads serial data stream until status bytes
 * are found. Fills these to dsp_status[].
 * The DSP inserts a byte after the ESC character to indicate the
 * nature of the following data. Here it is called type and
 *    type=='A' status and spectral data
 *    type=='B' status data only
 */
int get_status_bytes(int type, double timeout)
	{
	int new_char, last_char, i, j, esc_count;
	time_t t0;

	time(&t0);
/* dstat, below can go?? */

	/* We are now looking for a sequence of bytes:
	 *   ESC type b0 b1 b2 ... b18
	 * where bi are DSP status_buf bytes.
	 */
	esc_count=0;
pp:	for(;;)					/* Wait for 1st character */
		{
		if (difftime(time(NULL), t0) > timeout) {beep(500); return 0;}
		if ( ASYNCINSTAT() > 0 )	/* Have ch, see if ESC_CHAR */
			{new_char = ASYNCIN();
			 if (new_char==ESC_CHAR)  esc_count++;
			 break;			/* Break the for(;;) */
			}
		}
	for(;;)			/* Wait here for ESC_CHAR B to come */
		{last_char=new_char;
		 if (difftime(time(NULL), t0) > timeout)  return 0;
		 if(ASYNCINSTAT() > 0)
			{new_char = ASYNCIN();
		 	 if (new_char==ESC_CHAR)  esc_count++;	
			 if (esc_count==2)	/* Waiting for ESC_CHAR,type */
				{esc_count=0;	/*    but got 2xESC_CHAR     */
				 goto pp;	/* So try again              */
				}
			 if ((last_char==ESC_CHAR) && (new_char==type))
				break;
			}
		}
	/* Now get dsp_status bytes */
	i=0; esc_count=0;
	for(;;)
		{
		if (difftime(time(NULL), t0) > timeout)  return 0;
		if( ASYNCINSTAT() > 0 )
			{new_char = ASYNCIN();
			 if (new_char==ESC_CHAR) esc_count++;
			 if (esc_count==2)
				{esc_count=0;
				 goto dstat1;
				}
			 dsp_status[i] = new_char;
			 /* If status is ESC, needs to be entered to table */
			 if (++i > 18)  break;
 dstat1:			}
		}
	return 1;
	}

/* get_dsp_mem()  -  The DM() or PM() item that is in the dsp_ptr[] table
 * can be retrieved. Offset is the number of addresses above the base.
 *   what==0  Request of DSP to get DM()
 *   what==1  Request of DSP to get PM()
 *   what==2  Acquire DM() from status bytes
 *   what==3  Acquire DM() from status bytes
 *
 * The status of this process is carried in the global int dsp_data_index.
 *   dsp_data_index==0   No data being acquired
 *   dsp_data_index>0    This index is currently requested from DSP
 *   dsp_data_index==-1  Request was cancelled by timeout
 *
 * The process can fail due to data being lost and the request can be cancelled
 * by timeouts. This generates an error.  dsp_data_index==-1 indicates that a
 * re-request is needed.
 *
 * In using this function, keep in mind that it is inherently very slow. When
 * FFT data is being acquired, it allows one data item every 0.6 sec. Normally,
 * almost no data is needed. NOT FINISHED
 */
void get_dsp_mem(int what, int* tbl, int item, int offset)
	{
#if 0
	static int last_index=1;
	static int* ptr_last_table;
	static int last_item;
	static int last_offset;
#endif
	if(what==2570) tbl[1]=item+offset; /*JUNK TO STOP WARNINGS */
	if (dsp_data_index==-1) tbl[1]=5432;/*MORE SUCH JUNK*/
	}

/* set_dsp_mem()  -  Downloads either DM(), for what==0, or DM()
 * for what==1. DM is 16 bit and PM is 24 bit.  NOT FINISHED
 */
void set_dsp_mem(int what, int adr, long int data)
	{

 if (what==adr+data +2345) return;/* Dummy stuff */
	if (what==0)			/* Download data memory */
	    {
	    dsp_cmd('3', adr, adr>>8, (int)data, (int)(data>>8));
	    }
	else if (what==1)		/* Download program memory */
	    {
	    }
	}

/**** UPDATE_MOON() *************************************************
 *
 * do_now=0    Update if minute has changed and print on screen
 * do_now=1    Update immediately, and print data on screen
 * do_now=2    Update immediately, but don't print data on screen
 *
 * returns the roundtrip doppler dop00+dop11 (double)
 */
double update_moon(struct tm *ptm, int do_now)
    {
    int j;
    static int last_min;
    static int last_day;
    struct latlon ll;
    struct radec  rd;
    struct hadec  hd, hd1;
    struct azel   ae;			/* Az-El coordinates of Moon */
    double rmoon;			/* Range loc[0] to Moon center */
    double gst;
    double pm;
    double mrange;
    double um_freq;			/* Freq used in doppler calc */
    double dop00, dop11;

    if (! input_if_af)		/* If RF processor use um_freq */
	um_freq = (double)tfreqL - (double)tvtr[band].sb * tvtr[band].lo_freq;

    if ((ptm->tm_yday != last_day) || (do_now==1) )
	{
	last_day=ptm->tm_yday;	/* Won't do this for another day */
	cur_env.day=(double) ptm->tm_mday;
	cur_env.month=1+ptm->tm_mon;
	cur_env.year=1900+ptm->tm_year;
	julday(&cur_env);		/* Update Julian day */
	nutat(&cur_env);		/* Once a day find nutation */
	obliq(&cur_env);		/*  and obliquity of ecliptic */
	}
    /* Calc the moon apparent position and Doppler, if appropriate:  */
    if ((last_min != ptm->tm_min) || (do_now>0) )	/* Appropriate? */
	{
	last_min=ptm->tm_min;	/* Don't do this for another min */


#if 0
	gst=cnvlt2gst( (double)ptm->tm_hour +
	    ((((double)ptm->tm_min) + ((double)ptm->tm_sec/60.0))/60.0),
	    &cur_env, &loc[0], &ut);
	moon(cur_env.dj, &ll.lon, &ll.lat, &pm);
	cnvll2radec(&ll, &rd, &cur_env);
#endif

	/* We calculate doppler twice, once for home loc[0] and
	 * then for loc[cur_loc] to get dx location doppler */

	j=cur_loc;				/* Now for DX location */


#if 0
	hd.ha=cnvra2ha(rd.ra, gst, loc[j].lon);
	hd.dec=rd.dec;
	parallax(loc[j].lat, &hd, pm, &hd1);
	cnvhadec2azel(&hd1, &loc[j], &ae);
#endif


	if (input_if_af)		/* Audio processor, use f_eme */
	    dop11=doppler(ptm, &ae, &rmoon, f_eme, &loc[j]);
	else				/* Full radio, use tfreqL */
	    dop11=doppler(ptm, &ae, &rmoon, um_freq, &loc[j]);


	j=0;
#if 0
	hd.ha=cnvra2ha(rd.ra, gst, loc[j].lon);
	hd.dec=rd.dec;
	parallax(loc[j].lat, &hd, pm, &hd1);
	cnvhadec2azel(&hd1, &loc[j], &ae);
#endif

#if UDUMP
 sprintf(cc, "Lat=%lf lon=%lf", r2d(loc[0].lat), r2d(loc[0].lon));
 setviewport(160, 390, 630, 400, NO_CLIP);
 clearviewport();
 settextjustify(CENTER_TEXT, TOP_TEXT);
 moveto(235, 1);
 outtext(cc);
#endif

	if (input_if_af)		/* Audio processor, use f_eme */
	    dop00=doppler(ptm, &ae, &rmoon, f_eme, &loc[j]);
	else				/* Full radio, use tfreqL */
	    dop00=doppler(ptm, &ae, &rmoon, um_freq, &loc[j]);

	if (do_now<2)
	    {
#if UDUMP
 sprintf(cc, "Version 2.01  Range O-M=%lf  Mean=%lf", rmoon, REM_MEAN);
 setviewport(160, 380, 630, 390, NO_CLIP);
 clearviewport();
 settextjustify(CENTER_TEXT, TOP_TEXT);
 moveto(235, 1);
 outtext(cc);
#endif
	    setviewport(129, 470, 579, 479, DO_CLIP); /* Clear line */
	    clearviewport();
	    settextjustify(0,2);
	    sprintf(cc, "Moon: %s-%s AZ=%.2lf EL=%.2lf %dHz %.2lfdB",
	        loc[0].nm, loc[cur_loc].nm, r2d(ae.az), r2d(ae.el),
	        near_int(dop00+dop11), -40.0*log10(rmoon/REM_MEAN));
	    moveto(0, 0);
	    outtext(cc);
	    }
	}
    return  dop00+dop11;
    }

/* update_sun()  -  If the diag_mode==4 this will print the az and el of the
 * Sun on the bottom diagnostic line.  Earth location is loc[0] always.
 * Also slight variation in Sun intensity due to distance != 1 AU.
 */
void update_sun(struct tm *ptm, int do_now)
    {
    static int last_min;
    static int last_day;
    struct latlon ll;
    struct radec  rd;
    struct hadec  hd, hd1;
    struct azel   ae;
    double gst;
    double sun_dist;

    if ((ptm->tm_yday != last_day) || do_now)
	{
	last_day=ptm->tm_yday;	/* Won't do this for another day */
	cur_env.day=(double) ptm->tm_mday;
	cur_env.month=1+ptm->tm_mon;
	cur_env.year=1900+ptm->tm_year;
	julday(&cur_env);		/* Update Julian day */
	nutat(&cur_env);		/* Once a day find nutation */
	obliq(&cur_env);		/*  and obliquity of ecliptic */
	}
    if ((last_min != ptm->tm_min) || do_now)
	{
	last_min=ptm->tm_min;	/* Don't do this for another min */
	gst=cnvlt2gst( (double)ptm->tm_hour +
	    ((((double)ptm->tm_min) + ((double)ptm->tm_sec/60.0))/60.0),
	    &cur_env, &loc[cur_loc], &ut);
	sun(cur_env.dj, ut, &ll.lon, &sun_dist);
	ll.lat=0.0;			/* Easy for the Sun */
	/* Now a couple of minor corrections.  These are not generally
	 * needed.  They maintain an accuracy of better than 1 sec in
	 * angle.  First for nutation from cur_env.dp [from nutat() ].
	 * Then a fixed correction of 0.00569 deg that is the change in Sun
	 * position during the time that the light was travelling to the Earth,
	 * called aberation.
	 */
	ll.lon += (cur_env.dp-d2r(0.00569));
	cnvll2radec(&ll, &rd, &cur_env);	/* Get RA and Dec */
	/* Convert RA to hour angle: */
	hd.ha=cnvra2ha(rd.ra, gst, loc[cur_loc].lon);
	hd.dec=rd.dec;
	/* A tiny correction for parallax (we are not at Earth center):  */
	parallax(loc[cur_loc].lat, &hd, d2r(0.0024428)/sun_dist, &hd1);
	cnvhadec2azel(&hd1, &loc[cur_loc], &ae);
	    
	setviewport(129, 470, 579, 479, DO_CLIP); /* Clear line */
	clearviewport();
	settextjustify(0,2);
	sprintf(cc, "Sun: %s AZ=%.2lf EL=%.2lf %.2lfdB",
	    loc[cur_loc].nm, r2d(ae.az), r2d(ae.el), -20.0*log10(sun_dist));
	moveto(0, 0);
	outtext(cc);
	}
    }

/* set_cur_loc()  -  Not all loc[i] need exist.  This moves through  to
 * those that do.  Note the data is initialized so that loc[0] always
 * exists so this routine can't hang. loc[i].time_zone is not used in
 * the program now except to keep track of the active loc[i].
 * If updn==-1 there is no change in loc[i], but just update screen.
 */
void set_cur_loc(int updn)
    {
    time_t t0;

    if (updn==1)				/* Go to next */
	{
	if (++cur_loc > 9)  cur_loc=0;
	while (loc[cur_loc].time_zone == 999)
	    {
	    if (++cur_loc > 9)  cur_loc=0;
	    }
	}
    else if (updn==0)				/* Go to previous */
	{
	if (--cur_loc < 0)  cur_loc=9;
	while (loc[cur_loc].time_zone == 999)
	    {
	    if (--cur_loc < 0)  cur_loc=9;
	    }
	}
    /* Now update bottom line of screen */
    if (ut_available)
	{
	if (ms_available)
	    t0=update_clock(1)+dsp_clock_base;
	else
	    (void)time(&t0);
	cnvut2tdt(&t0, &tdt);

	if (diag_mode==3)
	    (void)update_moon(&tdt, 1);	/* 1 for 'do_now'  */
	else if (diag_mode==4)
	    update_sun(&tdt, 1);
	}
    if(!ut_available && (diag_mode==3 || diag_mode==4))
	{
	setviewport(129, 470, 579, 479, DO_CLIP); /* Clear line */
	clearviewport();
	settextjustify(0,2);
	moveto(0, 0);
	outtext("Moon/Sun data unavailable; need TZ in DOS enviornment");
	}
    }

/* Hardware specific code to load both the LMX1501 synthesizer as well
 * as the 16 bit control shift register.  lsynth is a 19 bit data
 * word to program the synthesizer.  It comes in two forms, one to
 * set up the ref divider and another to set the frequency.  Both are treated
 * identically here.
 *
 * The 19+16=35 bits are sent out as three 16 bit commands with the first 13
 * bits being passed through the shift registers:
 *	word_first	MS 3 bits of synthesizer data
 *	word_mid	LS 16 bits of synthesizer data
 * Only 16 bits are sent per command to ensure that the shift reg is
 * loaded before the next command is sent asking for more.
 */
void send_shift_data(unsigned long lsynthx)
	{int i;
	 unsigned int  sr;
	 unsigned int  word_first, word_mid;

	 word_first = (unsigned int) (0X7UL & (lsynthx>>16));  /* 3 bits */
	 word_mid   = (unsigned int) (0XFFFFUL & lsynthx);     /* 16 bits */
	 /* 'Word_last' comes from shift_reg (low 8) and
	    rfg[i].gainbyte (high 8)                  */

	 ASYNCOUT(ESC_CHAR);  ASYNCOUT(8);  /* Clock in first 16 bits */
	 ASYNCOUT(word_first & 0XFF); ASYNCOUT((word_first >> 8) & 0X00FFU);
	 ASYNCOUT(3); ASYNCOUT(4);	/* Filler bytes */
/*  printf("                 SHIFT1: %4X ", word_first); */
	 ASYNCOUT(ESC_CHAR);  ASYNCOUT(8);  /* Clock in middle 16 bits */
	 ASYNCOUT( word_mid & 0XFF ); ASYNCOUT( (word_mid >> 8) & 0X00FFU);
	 ASYNCOUT(3); ASYNCOUT(4);
/*  printf("%4X ", word_mid ); */
	 ASYNCOUT(ESC_CHAR);  ASYNCOUT(8);  /* Clock in last 16 bits */
	 ASYNCOUT( shift_reg & 0XFF );
	 ASYNCOUT( rfg[rfgain].gainbyte & 0XFF );
	 ASYNCOUT(3); ASYNCOUT(4);
/*  printf("%4X\n", shift_reg + 256 * rfg[rfgain].gainbyte ); */

	 ASYNCOUT(ESC_CHAR);  ASYNCOUT(7);  /* The 7 cmd sets the enable gate*/
	 ASYNCOUT(1); ASYNCOUT(2);	    /*   to load the Shift Reg */
	 ASYNCOUT(dsp_deltap_lo); ASYNCOUT(dsp_deltap_hi); /* and dsp freq */

	 ASYNCOUT(ESC_CHAR);  ASYNCOUT(7);  /* Close the enable gate to */
	 ASYNCOUT(0); ASYNCOUT(2);	    /*   prepare for next time */
	 ASYNCOUT(dsp_deltap_lo); ASYNCOUT(dsp_deltap_hi);
	}

/* do_t_error(err, tau)  -  Displays error err for about 15 seconds or
 * until replaced by next error.  Sounds beep for tau millisecgonds.
 * err==err+1000 clears err if the current error is err.
 * err==9999 clears all errors.
 * Similar to do_error, but prints text, and no return.
 */
void do_t_error(int err, char *ce, int tau)
	{
	static int current_err=0;

	if (beep_on_error && tau>0)  beep(tau);
	if ((err>0) && (err<1000))
	    display_error_count=25;	/* About 15 seconds */

	if ((err==9999) || ((err>1000) && (current_err==(err-1000))))
	    current_err=0;
	if (err<1000)
	    current_err=err;
	setviewport(0,457,127,467,DO_CLIP);
	clearviewport();
	setcolor(c_text_note);
	settextstyle(0, 0, 1);
	settextjustify(1, 2);
	moveto(63,1);
	if (current_err>0)  outtext(ce);
	setcolor(c_text_bold);
	}

/* near_int(x)  -  Returns the integer nearest x.
 */
int near_int(double x)
	{
	if (x <= -32768.0)  return -32768;
	if (x >= 32767.0)  return 32767;
	if (x >= 0.0)  return (int)(x+0.5);
	return (int)(x-0.5);
	}

/*****************************************************************************
 *
 * double upow10(double x)  -  Returns 10^x.
 * Wraps around the library pow(10.0, x) to not create an error message.
 * The D_VBIG response is intended as the best one can
 * do for a real-time process.  This function is not intended for
 * math computation, in general.
 * RSL 2 March 00
 * Still problems. Reduced max |x| to 20  RSL 16 Jan 00
 */
double upow10(double x)
    {
    /* This seems to still lead to overflow in pow(10.0, x)  */
    /* if (x > DBL_MAX_10_EXP-12) return D_VBIG;  */

    if      (x>20.0)  x=20.0;
    else if (x<-20.0) x=-20.0;

    return pow(10.0, x);
    }
    
/*****************************************************************************
 *
 * double p2db(double p)  -  Returns 10*log10(p), accurate to 7 places.
 * Replaces library log10() as it does not create error interrupt.
 * About the same speed as 10*log10(p), depending on the value of p,
 * but less accurate.
 * Data from Abramowitz and Stegun, scaled from ln(x).
 * RSL 2 March 00
 */
double p2db(double p)
    {
    double t1, t2, tpow;
    double db_scale;                    /* dB due to scaling into range */
    double d;                           /* Power in dB */
    int i;

    /* 10log(0) d= -2956.5  A response for 0 or negative */
    if (p < D_TINY) return -2956.5;
    db_scale=0.0;
    while (p<0.316227766)
        {p *= 10.0;
         db_scale -= 10.0;
        }

    while (p>3.16227766)
        {p *= 0.1;
         db_scale += 10.0;
        }

    /* p is in range .316 to 3.16 */
    t1=(p-1.0)/(p+1.0);
    t2=t1*t1;
    d=8.68591718*t1;
    t1 *= t2;
    d  += 2.89335524*t1;
    t1 *= t2;
    d  += 1.77522071*t1;
    t1 *= t2;
    d  += 0.94376476*t1;
    t1 *= t2;
    d  += (db_scale + 1.91337714*t1);

    return d;
    }

/* get_kbd_int()  -  Returns an integer from keyboard entry.  Assumes text
 * mode (not graphics.)  Ends on 'Enter' and responds to Backspace
 */
int  get_kbd_int(void)
	{
	char  c, *cptr;
	cptr=cc;
	for(;;)
	    {
	    if (kbhit()) 
		{
		c=(char)(_bios_keybrd(_KEYBRD_READ) & 0XFF);
		if ( (c==8) && (cptr>cc) )
		    {
		    putch(8);
		    putch(32);
		    putch(8);
		    cptr--;
		    }
		else if (c==13)
		    {
		    *cptr=0;
		    putch(13);
		    putch(10);
		    break;
		    }
		else if (c>=32 && c<127)
		    {
		    putch((int)c);
		    *cptr++=c;
		    }
		if ( (cptr-cc)>49 )
		    {
		    *cptr=0;
		    break;
		    }
		}
	    }
	return atoi(cc);
	}	

/* keyident() - Utility that is available before graphics is started.
 * Gives information on key that has been hit, for use in setting up
 * header file for keyboard definition.
 */
int keyident(void)
	{
	unsigned int stat, s, ichr;
	long lkey;

	stat = _bios_keybrd(_KEYBRD_SHIFTSTATUS);
	s = stat;
	/* Make left and right shift keys equals (bits 0 and 1) and
	 * mask out Caps Lock, Num Lock and Insert (bits 5, 6, and 7).     */
	if (stat&2)   stat |= 1;
	stat &= 0X1D;
	if ((stat&1) > 0) printf("SHIFT, ");
	if ((stat&4) > 0) printf("CTRL, ");
	if ((stat&8) > 0) printf("ALT, ");
	if ((stat&16) > 0) printf("ScrlLk, ");
	ichr = _bios_keybrd(_KEYBRD_READ);
	lkey = (long)ichr + (((long)stat)<<16);
	printf(" ichr=%X  stat=%X.  Hex value for header file: 0X%lX\n",
		ichr, s, lkey);
	if (stat==0 && (ichr&0XFF)==0X0D)
	    return 1;
	else
	    {
	    printf("\nHit 'Enter' to proceed (others give key ident)\n");
	    return 0;
	    }
	}

void get_mem(void)			/* Get memory for screen scroll */
	{
	int hc;

	/* This is part of sign-on and set-up! */
	if ( (big_area_ptr1 = (char *)malloc((size_t)60000) ) == NULL)
		{printf("Unable to get memory for bit image (top). ");
		 printf("'Enter' to continue\n");
		 getch();
		 exit(0);
		}
	if ( (big_area_ptr2 = (char *)malloc((size_t)50000) ) == NULL)
		{printf("Unable to get memory for bit image (bottom). ");
		 printf("'Enter' to continue\n");
		 getch();
		 exit(0);
		}
	if ( (power10 = (double *)calloc(256, sizeof(double)) ) == NULL)
		{printf("Unable to get memory for power10[]. ");
		 printf("'Enter' to continue\n");
		 getch();
		 exit(0);
		}
	if ( (pwrdat = (double *)calloc(515, sizeof(double)) ) == NULL)
		{printf("Unable to get memory for pwrdat[]. ");
		 printf("'Enter' to continue\n");
		 getch();
		 exit(0);
		}
	if ( (dnorm = (double *)calloc(515, sizeof(double)) ) == NULL)
		{printf("Unable to get memory for dnorm[]. ");
		 printf("'Enter' to continue\n");
		 getch();
		 exit(0);
		}
	if ( (lt_dnorm = (double *)calloc(256, sizeof(double)) ) == NULL)
		{printf("Unable to get memory for lt_dnorm[]. ");
		 printf("'Enter' to continue\n");
		 getch();
		 exit(0);
		}
	if ( (lt_pwrdat = (double *)calloc(256, sizeof(double)) ) == NULL)
		{printf("Unable to get memory for lt_pwrdat[]. ");
		 printf("'Enter' to continue\n");
		 getch();
		 exit(0);
		}
	if ( (pd16 = (double *)calloc(256, sizeof(double)) ) == NULL)
		{printf("Unable to get memory for pd16[]. ");
		 printf("'Enter' to continue\n");
		 getch();
		 exit(0);
		}

	if (startup_detail)
	    {
	    printf("big_area_ptr1 = %p\n", big_area_ptr1);
	    printf("big_area_ptr2 = %p\n", big_area_ptr2);
	    printf("power10[]     = %p\n", power10);
	    printf("pwrdat[]      = %p\n", pwrdat);
	    printf("dnorm[]       = %p\n", dnorm);
	    printf("lt_dnorm[]    = %p\n", lt_dnorm);
	    printf("lt_pwrdat[]   = %p\n", lt_pwrdat);
	    printf("pd16[]        = %p\n", pd16);
	    }
	hc=heapcheck();
	if (hc==_HEAPCORRUPT)
	    {
	    printf("Heap has been corrupted\n");
	    exit(0);
	    }
	if (hc==_HEAPEMPTY)
	    {
	    printf("No heap\n");
	    exit(0);
	    }
	if (startup_detail && hc==_HEAPOK)      printf("Heap is verified\n");
	}
